// =================================================================================================
// Copyright 2002-2005 Adobe Systems Incorporated
// All Rights Reserved.
//
// NOTICE:  Adobe permits you to use, modify, and distribute this file in accordance with the terms
// of the Adobe license agreement accompanying it.
//
// Adobe patent application tracking #P435, entitled 'Unique markers to simplify embedding data of
// one format in a file with a different format', inventors: Sean Parent, Greg Gilley.
// =================================================================================================

#include "XMP_Environment.h"	// ! This must be the first include!
#include "XMP_Impl.hpp"

#include "XMPMeta.hpp"
#include "XMPIterator.hpp"
#include "XMPUtils.hpp"

#include "XMP_Version.h"
#include "UnicodeInlines.incl_cpp"
#include "UnicodeConversions.hpp"
#include "ExpatAdapter.hpp"

#if XMP_DebugBuild
	#include <iostream>
#endif

using namespace std;

#if XMP_WinBuild
	#pragma warning ( disable : 4533 )	// initialization of '...' is skipped by 'goto ...'
	#pragma warning ( disable : 4702 )	// unreachable code
	#pragma warning ( disable : 4800 )	// forcing value to bool 'true' or 'false' (performance warning)
#endif


// *** Use the XMP_PropIsXyz (Schema, Simple, Struct, Array, ...) macros
// *** Add debug codegen checks, e.g. that typical masking operations really work
// *** Change all uses of strcmp and strncmp to XMP_LitMatch and XMP_LitNMatch


// =================================================================================================
// Local Types and Constants
// =========================

#ifndef DumpXMLParseTree
	#define DumpXMLParseTree 0
#endif

typedef unsigned char XMP_CLTMatch;

enum {	// Values for XMP_CLTMatch.
	kXMP_CLT_NoValues,
	kXMP_CLT_SpecificMatch,
	kXMP_CLT_SingleGeneric,
	kXMP_CLT_MultipleGeneric,
	kXMP_CLT_XDefault,
	kXMP_CLT_FirstItem
};

static const char * kPacketHeader  = "<?xpacket begin=\"\xEF\xBB\xBF\" id=\"W5M0MpCehiHzreSzNTczkc9d\"?>";
static const char * kPacketTrailer = "<?xpacket end=\"w\"?>";	// ! The w/r is at [size-4].

static const char * kPXMP_PacketStart = "<pxmp:XMP_Packet";
static const char * kPXMP_PacketEnd   = "</pxmp:XMP_Packet>";

static const char * kPXMP_SchemaGroup = "XMP_SchemaGroup";

static const char * kRDF_XMPMetaStart = "<x:xmpmeta xmlns:x=\"adobe:ns:meta/\" x:xmptk=\"";
static const char * kRDF_XMPMetaEnd   = "</x:xmpmeta>";

static const char * kRDF_RDFStart     = "<rdf:RDF xmlns:rdf=\"http://www.w3.org/1999/02/22-rdf-syntax-ns#\">";
static const char * kRDF_RDFEnd       = "</rdf:RDF>";

static const char * kRDF_SchemaStart  = "<rdf:Description rdf:about=";
static const char * kRDF_SchemaEnd    = "</rdf:Description>";

static const char * kRDF_StructStart  = "<rdf:Description>";
static const char * kRDF_StructEnd    = "</rdf:Description>";

static const char * kRDF_BagStart     = "<rdf:Bag>";
static const char * kRDF_BagEnd       = "</rdf:Bag>";

static const char * kRDF_SeqStart     = "<rdf:Seq>";
static const char * kRDF_SeqEnd       = "</rdf:Seq>";

static const char * kRDF_AltStart     = "<rdf:Alt>";
static const char * kRDF_AltEnd       = "</rdf:Alt>";

static const char * kRDF_ItemStart    = "<rdf:li>";
static const char * kRDF_ItemEnd      = "</rdf:li>";

static const char * kRDF_ValueStart   = "<rdf:value>";
static const char * kRDF_ValueEnd     = "</rdf:value>";

static const char *	kHexDigits    = "0123456789ABCDEF";


// =================================================================================================
// Static Variables
// ================

XMP_VarString * xdefaultName = 0;

XMP_AssertWrapperProc sAssertWrapper = 0;
XMP_AssertNotifyProc  sAssertNotify  = 0;
void *                sAssertRefCon  = 0;

#if XMP_DebugBuild
	#define kXMP_DebugFlag 1
#else
	#define kXMP_DebugFlag 0
#endif

#define kXMP_VersionNumber	( (kXMP_DebugFlag << 31)    |		\
                              (XMP_API_VERSION_MAJOR << 24) |	\
						      (XMP_API_VERSION_MINOR << 16) |	\
						      (XMP_API_VERSION_MICRO << 8) )


#define kToolkitName "Public XMP Toolkit"
#define kXMP_VersionMessage	kToolkitName " " XMP_API_VERSION_STRING

// These are embedded version strings.

static const char * kXMP_EmbeddedVersion   = kXMP_VersionMessage;
static const char * kXMP_EmbeddedCopyright = kToolkitName " copyright (c) 2002-2005, Adobe Systems Incorporated";

// =================================================================================================
// Local Utilities
// ===============


#define IsHexDigit(ch)		( (('0' <= (ch)) && ((ch) <= '9')) || (('A' <= (ch)) && ((ch) <= 'F')) )
#define HexDigitValue(ch)	( (((ch) - '0') < 10) ? ((ch) - '0') : ((ch) - 'A' + 10) )

static const char * kTenSpaces = "          ";
#define OutProcPadding(pad)	{ size_t padLen = (pad); 												\
							  for ( ; padLen >= 10; padLen -= 10 ) OutProcNChars ( kTenSpaces, 10 );	\
							  for ( ; padLen > 0; padLen -= 1 ) OutProcNChars ( " ", 1 ); }


#define OutProcNewline()	{ status = (*outProc) ( refCon, "\n", 1 );  if ( status != 0 ) goto EXIT; }

#define OutProcNChars(p,n)	{ status = (*outProc) ( refCon, (p), (n) );  if ( status != 0 ) goto EXIT; }

#define OutProcLiteral(lit)	{ status = (*outProc) ( refCon, (lit), strlen(lit) );  if ( status != 0 ) goto EXIT; }

#define OutProcString(str)	{ status = (*outProc) ( refCon, (str).c_str(), (str).size() );  if ( status != 0 ) goto EXIT; }

#define OutProcDecInt(num)	{ snprintf ( buffer, sizeof(buffer), "%d", (num) ); /* AUDIT: Using sizeof for snprintf length is safe */	\
							  status = (*outProc) ( refCon, buffer, strlen(buffer) );  if ( status != 0 ) goto EXIT; }

#define OutProcHexInt(num)	{ snprintf ( buffer, sizeof(buffer), "%X", (num) ); /* AUDIT: Using sizeof for snprintf length is safe */	\
							  status = (*outProc) ( refCon, buffer, strlen(buffer) );  if ( status != 0 ) goto EXIT; }

static const char * kIndent = "   ";
#define OutProcIndent(lev)	{ for ( size_t i = 0; i < (lev); ++i ) OutProcNChars ( kIndent, 3 ); }


// -------------------------------------------------------------------------------------------------
// DumpStringMap
// -------------

static XMP_Status
DumpStringMap ( const XMP_StringMap & map, XMP_StringPtr label, XMP_TextOutputProc outProc, void * refCon )
{
	XMP_Status			status;
	XMP_cStringMapPos	currPos;
	XMP_cStringMapPos	endPos = map.end();
	
	size_t maxLen = 0;
	for ( currPos = map.begin(); currPos != endPos; ++currPos ) {
		size_t currLen = currPos->first.size();
		if ( currLen > maxLen ) maxLen = currLen;
	}
	
	OutProcNewline();
	OutProcLiteral ( label );
	OutProcNewline();
	
	for ( currPos = map.begin(); currPos != endPos; ++currPos ) {
		OutProcNChars ( "  ", 2 );
		OutProcString ( currPos->first );
		OutProcPadding ( maxLen - currPos->first.size() );
		OutProcNChars ( " => ", 4 );
		OutProcString ( currPos->second );
		OutProcNewline();
	}
	
EXIT:
	return status;
	
}	// DumpStringMap


// -------------------------------------------------------------------------------------------------
// DumpNodeOptions
// ---------------

static XMP_Status
DumpNodeOptions	( XMP_OptionBits	 options,
				  XMP_TextOutputProc outProc,
				  void *			 refCon )
{
	XMP_Status	status;
	char		buffer [32];	// Decimal of a 64 bit int is at most about 20 digits.

	static const char * optNames[] = { " schema",		// 0x8000_0000
									   " ?30",
									   " ?29",
									   " -COMMAS-",
									   " ?27",			// 0x0800_0000
									   " ?26",
									   " ?25",
									   " ?24",
									   " ?23",			// 0x0080_0000
									   " isStale",
									   " isDerived",
									   " isStable",
									   " ?19",			// 0x0008_0000
									   " isInternal",
									   " hasAliases",
									   " isAlias",
									   " -AFTER-",		// 0x0000_8000
									   " -BEFORE-",
									   " isCompact",
									   " isLangAlt",
									   " isAlt",		// 0x0000_0800
									   " isOrdered",
									   " isArray",
									   " isStruct",
									   " hasType",		// 0x0000_0080
									   " hasLang",
									   " isQual",
									   " hasQual",
									   " ?3",			// 0x0000_0008
									   " ?2",
									   " URI",
									   " ?0" };

	if ( options == 0 ) {
	
		OutProcNChars ( "(0x0)", 5 );
	
	} else {
	
		OutProcNChars ( "(0x", 3 );
		OutProcHexInt ( options );
		OutProcNChars ( " :", 2 );

		XMP_OptionBits mask = 0x80000000;
		for ( int b = 0; b < 32; ++b ) {
			if ( options & mask ) OutProcLiteral ( optNames[b] );
			mask = mask >> 1;
		}
		OutProcNChars ( ")", 1 );
	
	}
	
EXIT:
	return status;

}	// DumpNodeOptions


// -------------------------------------------------------------------------------------------------
// DumpPropertyTree
// ----------------

// *** Extract the validation code into a separate routine to call on exit in debug builds.

static XMP_Status
DumpPropertyTree ( const XMP_Node *	  currNode,
				   int				  indent,
				   size_t			  itemIndex,
				   XMP_TextOutputProc outProc,
				   void *			  refCon )
{
	XMP_Status	status;
	char		buffer [32];	// Decimal of a 64 bit int is at most about 20 digits.

	OutProcIndent ( (size_t)indent );
	if ( itemIndex == 0 ) {
		if ( currNode->options & kXMP_PropIsQualifier ) OutProcNChars ( "? ", 2 );
		OutProcString ( currNode->name );
	} else {
		OutProcNChars ( "[", 1 );
		OutProcDecInt ( itemIndex );
		OutProcNChars ( "]", 1 );
	}

	if ( ! (currNode->options & kXMP_PropCompositeMask) ) {
		OutProcNChars ( " = \"", 4 );
		OutProcString ( currNode->value );
		OutProcNChars ( "\"", 1 );
	}

	if ( currNode->options != 0 ) {
		OutProcNChars ( "  ", 2 );
		status = DumpNodeOptions ( currNode->options, outProc, refCon );
		if ( status != 0 ) goto EXIT;
	}
	
	if ( currNode->options & kXMP_PropHasLang ) {
		if ( currNode->qualifiers.empty() || (currNode->qualifiers[0]->name != "xml:lang") ) {
			OutProcLiteral ( "  ** bad lang flag **" );
		}
	}
	// *** Check rdf:type also.

	if ( ! (currNode->options & kXMP_PropCompositeMask) ) {
		if ( ! currNode->children.empty() ) OutProcLiteral ( "  ** bad children **" );
	} else if ( currNode->options & kXMP_PropValueIsArray ) {
		if ( currNode->options & kXMP_PropValueIsStruct ) OutProcLiteral ( "  ** bad comp flags **" );
	} else if ( (currNode->options & kXMP_PropCompositeMask) != kXMP_PropValueIsStruct ) {
		OutProcLiteral ( "  ** bad comp flags **" );
	}

	#if 0	// *** XMP_DebugBuild
		if ( (currNode->_namePtr != currNode->name.c_str()) ||
			 (currNode->_valuePtr != currNode->value.c_str()) ) OutProcLiteral ( "  ** bad debug string **" );
	#endif
	
	OutProcNewline();
	
	for ( size_t qualNum = 0, qualLim = currNode->qualifiers.size(); qualNum < qualLim; ++qualNum ) {

		const XMP_Node * currQual = currNode->qualifiers[qualNum];

		if ( currQual->parent != currNode ) OutProcLiteral ( "** bad parent link => " );
		if ( currQual->name == kXMP_ArrayItemName ) OutProcLiteral ( "** bad qual name => " );
		if ( ! (currQual->options & kXMP_PropIsQualifier) ) OutProcLiteral ( "** bad qual flag => " );
		if ( currQual->name == "xml:lang" ) {
			if ( (qualNum != 0) || (! (currNode->options & kXMP_PropHasLang)) ) OutProcLiteral ( "** bad lang qual => " );
		}

		status = DumpPropertyTree ( currQual, indent+2, 0, outProc, refCon );
		if ( status != 0 ) goto EXIT;

	}

	for ( size_t childNum = 0, childLim = currNode->children.size(); childNum < childLim; ++childNum ) {

		const XMP_Node * currChild = currNode->children[childNum];

		if ( currChild->parent != currNode ) OutProcLiteral ( "** bad parent link => " );
		if ( currChild->options & kXMP_PropIsQualifier ) OutProcLiteral ( "** bad qual flag => " );
	
		if ( currNode->options & kXMP_PropValueIsArray ) {
			itemIndex = childNum+1;
			if ( currChild->name != kXMP_ArrayItemName ) OutProcLiteral ( "** bad item name => " );
		} else {
			itemIndex = 0;
			if ( currChild->name == kXMP_ArrayItemName ) OutProcLiteral ( "** bad field name => " );
		}

		status = DumpPropertyTree ( currChild, indent+1, itemIndex, outProc, refCon );
		if ( status != 0 ) goto EXIT;

	}
	
EXIT:
	return status;

}	// DumpPropertyTree


// -------------------------------------------------------------------------------------------------
// DumpXMLTree
// -----------

#if DumpXMLParseTree

static inline void PutHexByte ( FILE * log, unsigned char ch )
{
	static const char * kHexDigits = "0123456789ABCDEF";

	fprintf ( log, "\\x" );
	if ( ch < 0x10 ) {
		fprintf ( log, "%c", kHexDigits[ch] );
	} else {
		fprintf ( log, "%c%c", kHexDigits[ch>>4], kHexDigits[ch&0xF] );
	}
	
}	// PutHexByte

// -------------------------------------------------------------------------------------------------

static void PutClearString ( FILE * log, const std::string & str )
{

	for ( size_t i = 0; i != str.size(); ++i ) {
		unsigned char ch = str[i];
		if ( (0x20 <= ch) && (ch <= 0x7F) ) {
			fprintf ( log, "%c", ch );
		} else {
			PutHexByte ( log, ch );
		}
	}

}	// PutClearString

// -------------------------------------------------------------------------------------------------

static void DumpXMLTree ( FILE * log, const XML_Node & node, int indent )
{
	size_t i;
	
	#if 0	// *** XMP_DebugBuild
		if ( (node._namePtr != node.name.c_str()) ||
			 (node._valuePtr != node.value.c_str()) ) fprintf ( log, "*** bad debug string ***\n" );
	#endif
	
	for ( i = 0; i != (size_t)indent; ++i ) fprintf ( log, "  " );
	
	switch ( node.kind ) {
	
		case kRootNode :
			fprintf ( log, "\nStart of XML tree dump\n\n" );
			if ( (indent != 0) || (! node.attrs.empty()) ||
				 (! node.ns.empty()) || (! node.name.empty()) || (!node.value.empty()) ) fprintf ( log, " ** invalid root ** \n" );
			for ( i = 0; i < node.children.size(); ++i ) {
				XMP_Uns8 kind = node.children[i]->kind;
				if ( (kind == kRootNode) || (kind == kAttrNode) ) fprintf ( log, " ** invalid child ** \n" );
				DumpXMLTree ( log, *node.children[i], indent+1 );
			}
			fprintf ( log, "\nEnd of XML tree dump\n" );
			break;
	
		case kElemNode :
			fprintf ( log, "Elem %s", node.name.c_str() );
			if ( indent == 0 ) fprintf ( log, " ** invalid elem ** " );
			if ( ! node.ns.empty() ) fprintf ( log, " @ %s", node.ns.c_str() );
			fprintf ( log, "\n" );
			for ( i = 0; i < node.attrs.size(); ++i ) {
				XMP_Uns8 kind = node.attrs[i]->kind;
				if ( kind != kAttrNode ) fprintf ( log, " ** invalid attr ** \n" );
				DumpXMLTree ( log, *node.attrs[i], indent+2 );
			}
			for ( i = 0; i < node.children.size(); ++i ) {
				XMP_Uns8 kind = node.children[i]->kind;
				if ( (kind == kRootNode) || (kind == kAttrNode) ) fprintf ( log, " ** invalid child ** \n" );
				DumpXMLTree ( log, *node.children[i], indent+1 );
			}
			break;
	
		case kAttrNode :
			fprintf ( log, "Attr %s", node.name.c_str() );
			if ( (indent == 0) || node.name.empty() || (! node.attrs.empty()) || (! node.children.empty()) ) fprintf ( log, " ** invalid attr ** " );
			fprintf ( log, " = \"" );
			PutClearString ( log, node.value );
			fprintf ( log, "\"" );
			if ( ! node.ns.empty() ) fprintf ( log, " @ %s", node.ns.c_str() );
			fprintf ( log, "\n" );
			break;
	
		case kCDataNode :
			if ( (indent == 0) || (! node.ns.empty()) || (! node.name.empty()) ||
				 (! node.attrs.empty()) || (! node.children.empty()) ) fprintf ( log, " ** invalid cdata ** \n" );
			fprintf ( log, "\"" );
			PutClearString ( log, node.value );
			fprintf ( log, "\"\n" );
			break;
	
		case kPINode :
			fprintf ( log, "PI %s", node.name.c_str() );
			if ( (indent == 0) || node.name.empty() || (! node.children.empty()) ) fprintf ( log, " ** invalid pi ** \n" );
			if ( ! node.value.empty() ) {
				fprintf ( log, " <? " );
				PutClearString ( log, node.value );
				fprintf ( log, " ?>" );
			}
			fprintf ( log, "\n" );
			break;
	
	}

}	// DumpXMLTree

#endif	// DumpXMLParseTree


// -------------------------------------------------------------------------------------------------
// SetNodeValue
// ------------

static inline void
SetNodeValue ( XMP_Node * node, XMP_StringPtr value )
{

	#if XMP_DebugBuild	// ! Hack to force an assert.
		if ( (node->name == "xap:TestAssertNotify") && XMP_LitMatch ( value, "DoIt!" ) ) {
			XMP_Assert ( node->name != "xap:TestAssertNotify" );
		}
	#endif
	
	VerifyUTF8 ( value );	// Throws if the string is not legit UTF-8.
	node->value = value;
	if ( XMP_PropIsQualifier(node->options) && (node->name == "xml:lang") ) NormalizeLangValue ( &node->value );
	#if 0	// *** XMP_DebugBuild
		node->_valuePtr = node->value.c_str();
	#endif
	
}	// SetNodeValue


// -------------------------------------------------------------------------------------------------
// SetNode
// -------
//
// The internals for SetProperty and related calls, used after the node is found or created.

static void
SetNode	( XMP_Node * node, XMP_StringPtr value, XMP_OptionBits options )
{
	if ( options & kXMP_DeleteExisting ) {
		XMP_ClearOption ( options, kXMP_DeleteExisting );
		node->options = options;
		node->value.erase();
		node->RemoveChildren();
		node->RemoveQualifiers();
	}
	
	node->options |= options;	// Keep options set by FindNode when creating a new node.

	if ( value != 0 ) {
	
		// This is setting the value of a leaf node.
		if ( node->options & kXMP_PropCompositeMask ) XMP_Throw ( "Composite nodes can't have values", kXMPErr_BadXPath );
		XMP_Assert ( node->children.empty() );
		SetNodeValue ( node, value );
	
	} else {
	
		// This is setting up an array or struct.
		if ( ! node->value.empty() ) XMP_Throw ( "Composite nodes can't have values", kXMPErr_BadXPath );
		if ( node->options & kXMP_PropCompositeMask ) {	// Can't change an array to a struct, or vice versa.
			if ( (options & kXMP_PropCompositeMask) != (node->options & kXMP_PropCompositeMask) ) {
				XMP_Throw ( "Requested and existing composite form mismatch", kXMPErr_BadXPath );
			}
		}
		node->RemoveChildren();
	
	}
	
}	// SetNode


// -------------------------------------------------------------------------------------------------
// DoSetArrayItem
// --------------

static void
DoSetArrayItem ( XMP_Node *		arrayNode,
				 XMP_Index		itemIndex,
				 XMP_StringPtr	itemValue,
				 XMP_OptionBits options )
{
	XMP_OptionBits itemLoc = options & kXMP_PropArrayLocationMask;
	XMP_Index      arraySize = arrayNode->children.size();
	
	options &= ~kXMP_PropArrayLocationMask;
	options = VerifySetOptions ( options, itemValue );
	
	// Now locate or create the item node and set the value. Note the index parameter is one-based!
	// The index can be in the range [0..size+1] or "last", normalize it and check the insert flags.
	// The order of the normalization checks is important. If the array is empty we end up with an
	// index and location to set item size+1.
	
	XMP_Node * itemNode = 0;
	
	if ( itemIndex == kXMP_ArrayLastItem ) itemIndex = arraySize;
	if ( (itemIndex == 0) && (itemLoc == kXMP_InsertAfterItem) ) {
		itemIndex = 1;
		itemLoc = kXMP_InsertBeforeItem;
	}
	if ( (itemIndex == arraySize) && (itemLoc == kXMP_InsertAfterItem) ) {
		itemIndex += 1;
		itemLoc = 0;
	}
	if ( (itemIndex == arraySize+1) && (itemLoc == kXMP_InsertBeforeItem) ) itemLoc = 0;
	
	if ( itemIndex == arraySize+1 ) {

		if ( itemLoc != 0 ) XMP_Throw ( "Can't insert before or after implicit new item", kXMPErr_BadIndex );
		itemNode = new XMP_Node ( arrayNode, kXMP_ArrayItemName, 0 );
		arrayNode->children.push_back ( itemNode );

	} else {

		if ( (itemIndex < 1) || (itemIndex > arraySize) ) XMP_Throw ( "Array index out of bounds", kXMPErr_BadIndex );
		--itemIndex;	// ! Convert the index to a C zero-based value!
		if ( itemLoc == 0 ) {
			itemNode = arrayNode->children[itemIndex];
		} else {
			XMP_NodePtrPos itemPos = arrayNode->children.begin() + itemIndex;
			if ( itemLoc == kXMP_InsertAfterItem ) ++itemPos;
			itemNode = new XMP_Node ( arrayNode, kXMP_ArrayItemName, 0 );
			itemPos = arrayNode->children.insert ( itemPos, itemNode );
		}

	}
	
	SetNode ( itemNode, itemValue, options );
	
}	// DoSetArrayItem


// -------------------------------------------------------------------------------------------------
// ChooseLocalizedText
// -------------------
//
// 1. Look for an exact match with the specific language.
// 2. If a generic language is given, look for partial matches.
// 3. Look for an "x-default" item.
// 4. Choose the first item.

static XMP_CLTMatch
ChooseLocalizedText ( const XMP_Node *	 arrayNode,
					  XMP_StringPtr		 genericLang,
					  XMP_StringPtr		 specificLang,
					  const XMP_Node * * itemNode )
{
	const XMP_Node * currItem = 0;
	const size_t itemLim = arrayNode->children.size();
	size_t itemNum;
	
	// See if the array has the right form. Allow empty alt arrays, that is what parsing returns.
	// *** Should check alt-text bit when that is reliably maintained.

	if ( ! ( XMP_ArrayIsAltText(arrayNode->options) ||
	         (arrayNode->children.empty() && XMP_ArrayIsAlternate(arrayNode->options)) ) ) {
		XMP_Throw ( "Localized text array is not alt-text", kXMPErr_BadXPath );
	}
	if ( arrayNode->children.empty() ) {
		*itemNode = 0;
		return kXMP_CLT_NoValues;
	}

	for ( itemNum = 0; itemNum < itemLim; ++itemNum ) {
		currItem = arrayNode->children[itemNum];
		if ( currItem->options & kXMP_PropCompositeMask ) {
			XMP_Throw ( "Alt-text array item is not simple", kXMPErr_BadXPath );
		}
		if ( currItem->qualifiers.empty() || (currItem->qualifiers[0]->name != "xml:lang") ) {
			XMP_Throw ( "Alt-text array item has no language qualifier", kXMPErr_BadXPath );
		}
	}

	// Look for an exact match with the specific language.
	for ( itemNum = 0; itemNum < itemLim; ++itemNum ) {
		currItem = arrayNode->children[itemNum];
		if ( currItem->qualifiers[0]->value == specificLang ) {
			*itemNode = currItem;
			return kXMP_CLT_SpecificMatch;
		}
	}
	
	if ( *genericLang != 0 ) {

		// Look for the first partial match with the generic language.
		const size_t genericLen = strlen ( genericLang );
		for ( itemNum = 0; itemNum < itemLim; ++itemNum ) {
			currItem = arrayNode->children[itemNum];
			XMP_StringPtr currLang = currItem->qualifiers[0]->value.c_str();
			const size_t currLangSize = currItem->qualifiers[0]->value.size();
			if ( (currLangSize >= genericLen) &&
				 XMP_LitNMatch ( currLang, genericLang, genericLen ) &&
				 ((currLangSize == genericLen) || (currLang[genericLen] == '-')) ) {
				*itemNode = currItem;
				break;	// ! Don't return, need to look for other matches.
			}
		}

		if ( itemNum < itemLim ) {
			
			// Look for a second partial match with the generic language.
			for ( ++itemNum; itemNum < itemLim; ++itemNum ) {
				currItem = arrayNode->children[itemNum];
				XMP_StringPtr currLang = currItem->qualifiers[0]->value.c_str();
				const size_t currLangSize = currItem->qualifiers[0]->value.size();
				if ( (currLangSize >= genericLen) &&
					 XMP_LitNMatch ( currLang, genericLang, genericLen ) &&
					 ((currLangSize == genericLen) || (currLang[genericLen] == '-')) ) {
					return kXMP_CLT_MultipleGeneric;	// ! Leave itemNode with the first partial match.
				}
			}
			return kXMP_CLT_SingleGeneric;	// No second partial match was found.

		}
		
	}
	
	// Look for an 'x-default' item.
	for ( itemNum = 0; itemNum < itemLim; ++itemNum ) {
		currItem = arrayNode->children[itemNum];
		if ( currItem->qualifiers[0]->value == "x-default" ) {
			*itemNode = currItem;
			return kXMP_CLT_XDefault;
		}
	}
	
	// Everything failed, choose the first item.
	*itemNode = arrayNode->children[0];
	return kXMP_CLT_FirstItem;
	
}	// ChooseLocalizedText


// -------------------------------------------------------------------------------------------------
// AppendLangItem
// --------------

static void
AppendLangItem ( XMP_Node * arrayNode, XMP_StringPtr itemLang, XMP_StringPtr itemValue )
{
	XMP_Node * newItem  = new XMP_Node ( arrayNode, kXMP_ArrayItemName, itemValue, (kXMP_PropHasQualifiers | kXMP_PropHasLang) );
	XMP_Node * langQual = new XMP_Node ( newItem, "xml:lang", itemLang, kXMP_PropIsQualifier );
	newItem->qualifiers.push_back ( langQual );

	if ( (arrayNode->children.empty()) || (langQual->value != "x-default") ) {
		arrayNode->children.push_back ( newItem );
	} else {
		arrayNode->children.insert ( arrayNode->children.begin(), newItem );
	}

}	// AppendLangItem

// -------------------------------------------------------------------------------------------------
// PickBestRoot
// ------------
//
// Pick the first pxmp:XMP_Packet or x:xmpmeta among multiple root candidates. If there aren't any,
// pick the first bare rdf:RDF if that is allowed. The returned root is the rdf:RDF child if an
// x:xmpmeta element was chosen.  The search is breadth first, so a higher level candiate is chosen
// over a lower level one that was textually earlier in the serialized XML.

static const XML_Node * PickBestRoot ( const XML_Node & xmlParent, XMP_OptionBits options )
{
	// Look among this parent's content for pxmp:XMP_Packet or x:xmpmeta. The recursion for
	// x:xmpmeta is broader than the strictly defined choice, but gives us smaller code.
	for ( size_t childNum = 0, childLim = xmlParent.content.size(); childNum < childLim; ++childNum ) {
		const XML_Node * childNode = xmlParent.content[childNum];
		if ( childNode->kind != kElemNode ) continue;
		if ( childNode->name == "pxmp:XMP_Packet" ) return childNode;
		if ( (childNode->name == "x:xmpmeta") || (childNode->name == "x:xapmeta") ) return PickBestRoot ( *childNode, 0 );
	}
	
	// Look among this parent's content for a bare rdf:RDF if that is allowed.
	if ( ! (options & kXMP_RequireXMPMeta) ) {
		for ( size_t childNum = 0, childLim = xmlParent.content.size(); childNum < childLim; ++childNum ) {
			const XML_Node * childNode = xmlParent.content[childNum];
			if ( childNode->kind != kElemNode ) continue;
			if ( childNode->name == "rdf:RDF" ) return childNode;
		}
	}
	
	// Recurse into the content.
	for ( size_t childNum = 0, childLim = xmlParent.content.size(); childNum < childLim; ++childNum ) {
		const XML_Node * foundRoot = PickBestRoot ( *xmlParent.content[childNum], options );
		if ( foundRoot != 0 ) return foundRoot;
	}
	
	return 0;

}	// PickBestRoot

// -------------------------------------------------------------------------------------------------
// FindRootNode
// ------------
//
// Find the XML node that is the root of the XMP data tree. Generally this will be an outer node,
// but it could be anywhere if a general XML document is parsed (e.g. SVG). The XML parser counted
// all rdf:RDF and pxmp:XMP_Packet nodes, and kept a pointer to the last one. If there is more than
// one possible root use PickBestRoot to choose among them.
//
// If there is a root node, try to extract the version of the previous XMP toolkit.

static const XML_Node * FindRootNode ( XMPMeta * thiz, const XMLParserAdapter & xmlParser, XMP_OptionBits options )
{
	const XML_Node * rootNode = xmlParser.rootNode;
	
	if ( xmlParser.rootCount > 1 ) rootNode = PickBestRoot ( xmlParser.tree, options );
	if ( rootNode == 0 ) return 0;
	
	// We have a root node. Try to extract previous toolkit version number.
	
	XMP_StringPtr verStr = "";
	
	if ( rootNode->name == "pxmp:XMP_Packet" ) {

		for ( size_t attrNum = 0, attrLim = rootNode->attrs.size(); attrNum < attrLim; ++attrNum ) {
			const XML_Node * currAttr =rootNode->attrs[attrNum];
			if ( currAttr->name == "pxmp:xmptk" ) {
				verStr = currAttr->value.c_str();
				break;
			}
		}

	} else {
	
		XMP_Assert ( rootNode->name == "rdf:RDF" );
	
		if ( (options & kXMP_RequireXMPMeta) &&
		     ((rootNode->parent == 0) ||
		      ((rootNode->parent->name != "x:xmpmeta") && (rootNode->parent->name != "x:xapmeta"))) ) return 0;

		for ( size_t attrNum = 0, attrLim = rootNode->parent->attrs.size(); attrNum < attrLim; ++attrNum ) {
			const XML_Node * currAttr =rootNode->parent->attrs[attrNum];
			if ( (currAttr->name == "x:xmptk") || (currAttr->name == "x:xaptk") ) {
				verStr = currAttr->value.c_str();
				break;
			}
		}
		
	}
	
	// Decode the version number into MMmmuubbb digits. If any part is too big, peg it at 99 or 999.
	
	unsigned long part;
	while ( (*verStr != 0) && ((*verStr < '0') || (*verStr > '9')) ) ++verStr;
	
	part = 0;
	while ( (*verStr != 0) && ('0' <= *verStr) && (*verStr <= '9') ) {
		part = (part * 10) + (*verStr - '0');
		++verStr;
	}
	if ( part > 99 ) part = 99;
	thiz->prevTkVer = part * 100*100*1000;
	
	part = 0;
	if ( *verStr == '.' ) ++verStr;
	while ( (*verStr != 0) && ('0' <= *verStr) && (*verStr <= '9') ) {
		part = (part * 10) + (*verStr - '0');
		++verStr;
	}
	if ( part > 99 ) part = 99;
	thiz->prevTkVer += part * 100*1000;
	
	part = 0;
	if ( *verStr == '.' ) ++verStr;
	while ( (*verStr != 0) && ('0' <= *verStr) && (*verStr <= '9') ) {
		part = (part * 10) + (*verStr - '0');
		++verStr;
	}
	if ( part > 99 ) part = 99;
	thiz->prevTkVer += part * 1000;
	
	part = 0;
	if ( *verStr == '-' ) ++verStr;
	while ( (*verStr != 0) && ('0' <= *verStr) && (*verStr <= '9') ) {
		part = (part * 10) + (*verStr - '0');
		++verStr;
	}
	if ( part > 999 ) part = 999;
	thiz->prevTkVer += part;
	
	return rootNode;
	
}	// FindRootNode

// -------------------------------------------------------------------------------------------------
// NormalizeDCArrays
// -----------------
//
// Undo the denormalization performed by the XMP used in Acrobat 5. If a Dublin Core array had only
// one item, it was serialized as a simple property. The xml:lang attribute was dropped from an
// alt-text item if the language was x-default.

// *** This depends on the dc: namespace prefix.

static void
NormalizeDCArrays ( XMP_Node * xmpTree )
{
	XMP_Node * dcSchema = FindSchemaNode ( xmpTree, kXMP_NS_DC, kXMP_ExistingOnly );
	if ( dcSchema == 0 ) return;
	
	for ( size_t propNum = 0, propLimit = dcSchema->children.size(); propNum < propLimit; ++propNum ) {
		XMP_Node *     currProp  = dcSchema->children[propNum];
		XMP_OptionBits arrayForm = 0;
		
		if ( ! XMP_PropIsSimple ( currProp->options ) ) continue;	// Nothing to do if not simple.
		
		if ( (currProp->name == "dc:creator" )     ||	// See if it is supposed to be an array.
		     (currProp->name == "dc:date" ) ) {			// *** Think about an array of char* and a loop.
			arrayForm = kXMP_PropArrayIsOrdered;
		} else if (
		     (currProp->name == "dc:description" ) ||
		     (currProp->name == "dc:rights" )      ||
		     (currProp->name == "dc:title" ) ) {
			arrayForm = kXMP_PropArrayIsAltText;
		} else if (
		     (currProp->name == "dc:contributor" ) ||
		     (currProp->name == "dc:language" )    ||
		     (currProp->name == "dc:publisher" )   ||
		     (currProp->name == "dc:relation" )    ||
		     (currProp->name == "dc:subject" )     ||
		     (currProp->name == "dc:type" ) ) {
			arrayForm = kXMP_PropValueIsArray;
		}
		if ( arrayForm == 0 ) continue;	// Nothing to do if it isn't supposed to be an array.
		
		arrayForm = VerifySetOptions ( arrayForm, 0 );	// Set the implicit array bits.
		XMP_Node * newArray = new XMP_Node ( dcSchema, currProp->name.c_str(), arrayForm );
		dcSchema->children[propNum] = newArray;
		newArray->children.push_back ( currProp );
		currProp->parent = newArray;
		currProp->name = kXMP_ArrayItemName;
		
		if ( XMP_ArrayIsAltText ( arrayForm ) && (! (currProp->options & kXMP_PropHasLang)) ) {
			XMP_Node * newLang = new XMP_Node ( currProp, "xml:lang", "x-default", kXMP_PropIsQualifier );
			currProp->options |= (kXMP_PropHasQualifiers | kXMP_PropHasLang);
			if ( currProp->qualifiers.empty() ) {	// *** Need a util?
				currProp->qualifiers.push_back ( newLang );
			} else {
				currProp->qualifiers.insert ( currProp->qualifiers.begin(), newLang );
			}
		}

	}
	
}	// NormalizeDCArrays


// -------------------------------------------------------------------------------------------------
// CompareAliasedSubtrees
// ----------------------

// *** Change to do some alias-specific setup, then use CompareSubtrees. One special case for
// *** aliases is a simple to x-default alias, the options and qualifiers obviously differ.

static void
CompareAliasedSubtrees ( XMP_Node * aliasNode, XMP_Node * baseNode, bool outerCall = true )
{
	// ! The outermost call is special. The names almost certainly differ. The qualifiers (and
	// ! hence options) will differ for an alias to the x-default item of a langAlt array.
	if ( (aliasNode->value != baseNode->value) ||
	     (aliasNode->children.size() != baseNode->children.size()) ) {
		XMP_Throw ( "Mismatch between alias and base nodes", kXMPErr_BadXMP );
	}
	if ( ! outerCall ) {
		if ( (aliasNode->name != baseNode->name) ||
		     (aliasNode->options != baseNode->options) ||
		     (aliasNode->qualifiers.size() != baseNode->qualifiers.size()) ) {
			XMP_Throw ( "Mismatch between alias and base nodes", kXMPErr_BadXMP );
		}
	}
	
	for ( size_t childNum = 0, childLim = aliasNode->children.size(); childNum < childLim; ++childNum ) {
		XMP_Node * aliasChild = aliasNode->children[childNum];
		XMP_Node * baseChild  = baseNode->children[childNum];
		CompareAliasedSubtrees ( aliasChild, baseChild, false );
	}
	
	for ( size_t qualNum = 0, qualLim = aliasNode->qualifiers.size(); qualNum < qualLim; ++qualNum ) {
		XMP_Node * aliasQual = aliasNode->qualifiers[qualNum];
		XMP_Node * baseQual  = baseNode->qualifiers[qualNum];
		CompareAliasedSubtrees ( aliasQual, baseQual, false );
	}
	
}	// CompareAliasedSubtrees


// -------------------------------------------------------------------------------------------------
// TransplantArrayItemAlias
// ------------------------

static void
TransplantArrayItemAlias ( XMP_Node * oldParent, size_t oldNum, XMP_Node * newParent )
{
	XMP_Node * childNode = oldParent->children[oldNum];

	if ( newParent->options & kXMP_PropArrayIsAltText ) {
		if ( childNode->options & kXMP_PropHasLang ) {
			XMP_Throw ( "Alias to x-default already has a language qualifier", kXMPErr_BadXMP );	// *** Allow x-default.
		}
		childNode->options |= (kXMP_PropHasQualifiers | kXMP_PropHasLang);
		XMP_Node * langQual = new XMP_Node ( childNode, "xml:lang", "x-default", kXMP_PropIsQualifier );	// *** AddLangQual util?
		if ( childNode->qualifiers.empty() ) {
			childNode->qualifiers.push_back ( langQual );
		} else {
			childNode->qualifiers.insert ( childNode->qualifiers.begin(), langQual );
		}
	}

	oldParent->children.erase ( oldParent->children.begin() + oldNum );
	childNode->name = kXMP_ArrayItemName;
	childNode->parent = newParent;
	if ( newParent->children.empty() ) {
		newParent->children.push_back ( childNode );
	} else {
		newParent->children.insert ( newParent->children.begin(), childNode );
	}

}	// TransplantArrayItemAlias


// -------------------------------------------------------------------------------------------------
// TransplantNamedAlias
// --------------------

static void
TransplantNamedAlias ( XMP_Node * oldParent, size_t oldNum, XMP_Node * newParent, XMP_VarString & newName )
{
	XMP_Node * childNode = oldParent->children[oldNum];

	oldParent->children.erase ( oldParent->children.begin() + oldNum );
	childNode->name = newName;
	childNode->parent = newParent;
	newParent->children.push_back ( childNode );

}	// TransplantNamedAlias


// -------------------------------------------------------------------------------------------------
// CleanupParsedAliases
// --------------------

static void
CleanupParsedAliases ( XMP_Node * tree, XMP_OptionBits parseOptions )
{
	tree->options ^= kXMP_PropHasAliases;
	const bool strictAliasing = ((parseOptions & kXMP_StrictAliasing) != 0);
	
	// Visit all of the top level nodes looking for aliases. If there is no base, transplant the
	// alias subtree. If there is a base and strict aliasing is on, make sure the alias and base 
	// subtrees match.
	
	// ! Use "while" loops not "for" loops since both the schema and property loops can remove the
	// ! current item from the vector being traversed. And don't increment the counter for a delete.
	
	size_t schemaNum = 0;
	while ( schemaNum < tree->children.size() ) {
		XMP_Node * currSchema = tree->children[schemaNum];
		
		size_t propNum = 0;
		while ( propNum < currSchema->children.size() ) {
			XMP_Node * currProp = currSchema->children[propNum];
			if ( ! (currProp->options & kXMP_PropIsAlias) ) {
				++propNum;
				continue;
			}
			currProp->options ^= kXMP_PropIsAlias;

			// Find the base path, look for the base schema and root node.

			XMP_AliasMapPos aliasPos = sRegisteredAliasMap->find ( currProp->name );
			XMP_Assert ( aliasPos != sRegisteredAliasMap->end() );
			XMP_ExpandedXPath & basePath = aliasPos->second;
			XMP_OptionBits arrayOptions = (basePath[kRootPropStep].options & kXMP_PropArrayFormMask);

			XMP_Node * baseSchema = FindSchemaNode ( tree, basePath[kSchemaStep].step.c_str(), kXMP_CreateNodes );
			if ( baseSchema->options & kXMP_NewImplicitNode ) baseSchema->options ^= kXMP_NewImplicitNode;
			XMP_Node * baseNode = FindChildNode ( baseSchema, basePath[kRootPropStep].step.c_str(), kXMP_ExistingOnly );

			if ( baseNode == 0 ) {
			
				if ( basePath.size() == 2 ) {
					// A top-to-top alias, transplant the property.
					TransplantNamedAlias ( currSchema, propNum, baseSchema, basePath[kRootPropStep].step );
				} else {
					// An alias to an array item, create the array and transplant the property.
					baseNode = new XMP_Node ( baseSchema, basePath[kRootPropStep].step.c_str(), arrayOptions );
					baseSchema->children.push_back ( baseNode );
					TransplantArrayItemAlias ( currSchema, propNum, baseNode );
				}
			
			} else if ( basePath.size() == 2 ) {
			
				// The base node does exist and this is a top-to-top alias. Check for conflicts if
				// strict aliasing is on. Remove and delete the alias subtree.
				if ( strictAliasing ) CompareAliasedSubtrees ( currProp, baseNode );
				currSchema->children.erase ( currSchema->children.begin() + propNum );
				delete currProp;
			
			} else {
			
				// This is an alias to an array item and the array exists. Look for the aliased item.
				// Then transplant or check & delete as appropriate.
				
				XMP_Node * itemNode = 0;
				if ( arrayOptions & kXMP_PropArrayIsAltText ) {
					XMP_Index xdIndex = LookupLangItem ( baseNode, *xdefaultName );
					if ( xdIndex != -1 ) itemNode = baseNode->children[xdIndex];
				} else if ( ! baseNode->children.empty() ) {
					itemNode = baseNode->children[0];
				}
				
				if ( itemNode == 0 ) {
					TransplantArrayItemAlias ( currSchema, propNum, baseNode );
				} else {
					if ( strictAliasing ) CompareAliasedSubtrees ( currProp, itemNode );
					currSchema->children.erase ( currSchema->children.begin() + propNum );
					delete currProp;
				}

			}

		}	// Property loop
		
		// Increment the counter or remove an empty schema node.
		if ( currSchema->children.size() > 0 ) {
			++schemaNum;
		} else {
			delete tree->children[schemaNum];	// ! Delete the schema node itself.
			tree->children.erase ( tree->children.begin() + schemaNum );
		}
		
	}	// Schema loop
	
}	// CleanupParsedAliases


// -------------------------------------------------------------------------------------------------
// FixASCIIControls
// ----------------
//
// Undo the "doubled escaping" of ASCII control characters. Look for "&#xnn;", where "nn" is one or
// two hex digits. Where found, replace with the single byte whose value is nn.
//
// It is a bit sad to walk the entire tree to do this, but doing it in the parsing code it risky.
// Doing it in the XML parsing might improperly process text that is not actually doubled escapes.
// Doing it in the RDF recognizer is complex, hard to make sure a value is checked exactly once.

static void FixASCIIControls ( XMP_VarString & strValue )
{
	if ( strValue.size() < 5 ) return;
	XMP_StringPtr cStr = strValue.c_str();	// ! Need the terminating null to ensure cStr[5] is valid.

	for ( size_t i = 0; i+5 <= strValue.size(); ++i ) {	// ! Need dynamic limit, strValue.size() can change.

		if ( (cStr[i] == '&') && (cStr[i+1] == '#') && (cStr[i+2] == 'x') && IsHexDigit(cStr[i+3]) ) {

			int excessLen = 4;
			unsigned char chValue = HexDigitValue ( cStr[i+3] );

			if ( IsHexDigit(cStr[i+4]) ) {
				++excessLen;
				chValue = (chValue << 4) + HexDigitValue(cStr[i+4]);
			}
			if ( cStr[i+excessLen] != ';' ) continue;	// Not a complete escape sequence.

			strValue[i] = chValue;
			strValue.erase ( i+1, excessLen );
			cStr = strValue.c_str();	// Probably doesn't change, but let's be safe.

		}

	}
	
}	// FixASCIIControls


// -------------------------------------------------------------------------------------------------
// TouchUpParsedNodes
// ------------------

static void
TouchUpParsedNodes ( XMP_Node * currNode )
{

	if ( currNode->parent == 0 ) {
	
		// This is the root node, recurse on the schema nodes.
		for ( size_t schemaNum = 0, schemaLim = currNode->children.size(); schemaNum < schemaLim; ++schemaNum ) {
			XMP_Node * currSchema = currNode->children[schemaNum];
			TouchUpParsedNodes ( currSchema );
		}
		
		// Tweak old XMP: Move an instance ID from rdf:about to the xmpMM:InstanceID property.
		// An old instance ID usually looks like "uuid:bac965c4-9d87-11d9-9a30-000d936b79c4",
		// plus InDesign 3.0 wrote them like "bac965c4-9d87-11d9-9a30-000d936b79c4". If the
		// name looks like a UUID simply move it to xmpMM:InstanceID, don't worry about any
		// existing xmpMM:InstanceID. Both will only be present when a newer file with the
		// xmpMM:InstanceID property is updated by an old app that uses rdf:about.
		
		if ( ! currNode->name.empty() ) {

			bool nameIsUUID = false;
			XMP_StringPtr nameStr = currNode->name.c_str();

			if ( XMP_LitNMatch ( currNode->name.c_str(), "uuid:", 5 ) ) {

				nameIsUUID = true;

			} else if ( currNode->name.size() == 36 ) {

				nameIsUUID = true;	// ! Assume true, we'll set it to false below if not.
				for ( int i = 0;  i < 36; ++i ) {
					if ( nameStr[i] == '-' ) {
						if ( (i != 8) && (i != 13) && (i != 18) && (i != 23) ) {
							nameIsUUID = false;
							break;
						}
					} else if ( ! ((('0' <= nameStr[i]) && (nameStr[i] <= '9')) || (('a' <= nameStr[i]) && (nameStr[i] <= 'z'))) ) {
						nameIsUUID = false;
						break;
					}
				}

			}
			
			if ( nameIsUUID ) {

				XMP_ExpandedXPath expPath;
				ExpandXPath ( kXMP_NS_XMP_MM, "InstanceID", &expPath );
				XMP_Node * idNode = FindNode ( currNode, expPath, kXMP_CreateNodes, 0 );
				if ( idNode == 0 ) XMP_Throw ( "Failure creating xmpMM:InstanceID", kXMPErr_InternalFailure );

				idNode->options = 0;	// Clobber any existing xmpMM:InstanceID.
				idNode->value = currNode->name;
				idNode->RemoveChildren();
				idNode->RemoveQualifiers();

				currNode->name.erase();

			}

		}
	
	} else if ( XMP_NodeIsSchema ( currNode->options ) ) {
	
		// This is a schema node, recurse on the properties.
		for ( size_t propNum = 0, propLim = currNode->children.size(); propNum < propLim; ++propNum ) {
			XMP_Node * currProp = currNode->children[propNum];
			TouchUpParsedNodes ( currProp );
		}
	
	} else {

		// This is a property or qualifier. Recurse on the children of structs and arrays, and the
		// qualifiers for all nodes. Fix up the ASCII controls in the values of simple leaf nodes.
		
		// *** We should only do this for RDF serialized by 3.1-053 or later. This awaits pulling
		// *** version info from the x:xmpmeta element. Doing it all the time will change user data
		// *** of "&#x9;" into a tab because old toolkits didn't double escape when serializing.

		if ( XMP_PropIsSimple ( currNode->options ) ) {
			FixASCIIControls ( currNode->value );
		} else {
			for ( size_t childNum = 0, childLim = currNode->children.size(); childNum < childLim; ++childNum ) {
				XMP_Node * currChild = currNode->children[childNum];
				TouchUpParsedNodes ( currChild );
			}
		}
		
		for ( size_t qualNum = 0, qualLim = currNode->qualifiers.size(); qualNum < qualLim; ++qualNum ) {
			XMP_Node * currQual = currNode->qualifiers[qualNum];
			TouchUpParsedNodes ( currQual );
		}
		
	}

}	// TouchUpParsedNodes


// -------------------------------------------------------------------------------------------------
// FixGPSTimeStamp
// ---------------
//
// Do some special case cleanup of the parsed tree.

static void
FixGPSTimeStamp ( XMP_Node * tree )
{
	
	XMP_Node * exifSchema  = FindSchemaNode ( tree, kXMP_NS_EXIF, kXMP_ExistingOnly );
	if ( exifSchema == 0 ) return;
	
	XMP_Node * gpsDateTime = FindChildNode ( exifSchema, "exif:GPSTimeStamp", kXMP_ExistingOnly );
	if ( gpsDateTime == 0 ) return;
	
	XMP_DateTime binGPSStamp;
	try {
		XMPUtils::ConvertToDate ( gpsDateTime->value.c_str(), &binGPSStamp );
	} catch ( ... ) {
		return;	// Don't let a bad date stop other things.
	}
	if ( (binGPSStamp.year != 0) || (binGPSStamp.month != 0) || (binGPSStamp.day != 0) ) return;
	
	XMP_Node * otherDate = FindChildNode ( exifSchema, "exif:DateTimeOriginal", kXMP_ExistingOnly );
	if ( otherDate == 0 ) otherDate = FindChildNode ( exifSchema, "exif:DateTimeDigitized", kXMP_ExistingOnly );
	if ( otherDate == 0 ) return;

	XMP_DateTime binOtherDate;
	try {
		XMPUtils::ConvertToDate ( otherDate->value.c_str(), &binOtherDate );
	} catch ( ... ) {
		return;	// Don't let a bad date stop other things.
	}
	
	binGPSStamp.year  = binOtherDate.year;
	binGPSStamp.month = binOtherDate.month;
	binGPSStamp.day   = binOtherDate.day;

	XMP_StringPtr goodStr;
	XMP_StringLen goodLen;
	XMPUtils::ConvertFromDate ( binGPSStamp, &goodStr, &goodLen );
	
	gpsDateTime->value.assign ( goodStr, goodLen );

}	// FixGPSTimeStamp

// -------------------------------------------------------------------------------------------------
// EstimateRDFSize
// ---------------

//  *** Pull the strlen(kXyz) calls into constants.

static size_t
EstimateRDFSize ( const XMP_Node * currNode, XMP_Index indent, size_t indentLen )
{
	size_t outputLen = 2 * (indent*indentLen + currNode->name.size() + 4);	// The property element tags.
	
	if ( ! currNode->qualifiers.empty() ) {
		// This node has qualifiers, assume it is written using rdf:value and estimate the qualifiers.

		indent += 2;	// Everything else is indented inside the rdf:Description element.
		outputLen += 2 * ((indent-1)*indentLen + strlen(kRDF_StructStart) + 2);	// The rdf:Description tags.
		outputLen += 2 * (indent*indentLen + strlen(kRDF_ValueStart) + 2);		// The rdf:value tags.

		for ( size_t qualNum = 0, qualLim = currNode->qualifiers.size(); qualNum < qualLim; ++qualNum ) {
			const XMP_Node * currQual = currNode->qualifiers[qualNum];
			outputLen += EstimateRDFSize ( currQual, indent, indentLen );
		}

	}
	
	if ( currNode->options & kXMP_PropValueIsStruct ) {
		indent += 1;
		outputLen += 2 * (indent*indentLen + strlen(kRDF_StructStart) + 2);	// The rdf:Description tags.
	} else if ( currNode->options & kXMP_PropValueIsArray ) {
		indent += 2;
		outputLen += 2 * ((indent-1)*indentLen + strlen(kRDF_BagStart) + 2);		// The rdf:Bag/Seq/Alt tags.
		outputLen += 2 * currNode->children.size() * (strlen(kRDF_ItemStart) + 2);	// The rdf:li tags, indent counted in children.
	} else if ( ! (currNode->options & kXMP_SchemaNode) ) {
		outputLen += currNode->value.size();	// This is a leaf value node.
	}

	for ( size_t childNum = 0, childLim = currNode->children.size(); childNum < childLim; ++childNum ) {
		const XMP_Node * currChild = currNode->children[childNum];
		outputLen += EstimateRDFSize ( currChild, indent+1, indentLen );
	}

	return outputLen;
	
}	// EstimateRDFSize


// -------------------------------------------------------------------------------------------------
// EstimatePlainXMPSize
// --------------------

//  *** Pull the strlen(kXyz) calls into constants.

static size_t
EstimatePlainXMPSize ( const XMP_Node * currNode, XMP_Index indent, size_t indentLen )
{
	size_t outputLen = 2 * (indent*indentLen + currNode->name.size() + 4);	// The property element tags.
	
	if ( XMP_PropIsSimple ( currNode->options ) ) {
		outputLen += currNode->value.size() + 10;	// For the value="..." portion.
	} else {
		outputLen += 15;	// For the kind="..." portion.
		for ( size_t childNum = 0, childLim = currNode->children.size(); childNum < childLim; ++childNum ) {
			const XMP_Node * currChild = currNode->children[childNum];
			outputLen += EstimatePlainXMPSize ( currChild, indent+1, indentLen );
		}
	}
	
	if ( ! currNode->qualifiers.empty() ) {	// This node has qualifiers, estimate their size.
		for ( size_t qualNum = 0, qualLim = currNode->qualifiers.size(); qualNum < qualLim; ++qualNum ) {
			const XMP_Node * currQual = currNode->qualifiers[qualNum];
			if ( XMP_PropIsSimple ( currQual->options ) ) {
				outputLen += currNode->name.size() + currNode->value.size() + 4;
			} else {
				outputLen += EstimatePlainXMPSize ( currQual, indent+1, indentLen );
			}
		}
	}

	return outputLen;
	
}	// EstimatePlainXMPSize


// -------------------------------------------------------------------------------------------------
// DeclareOneNamespace
// -------------------

static void
DeclareOneNamespace	( const XMP_VarString &	nsPrefix,
					  const XMP_VarString &	nsURI,
					  XMP_VarString	&		usedNS,		// ! A catenation of the prefixes with colons.
					  XMP_VarString &		outputStr,
					  XMP_StringPtr			newline,
					  XMP_StringPtr			indentStr,
					  XMP_Index				indent )
{
	size_t nsPos = usedNS.find ( nsPrefix );

	if ( nsPos == XMP_VarString::npos ) {
		
		outputStr += newline;
		for ( ; indent > 0; --indent ) outputStr += indentStr;
		outputStr += "xmlns:";
		outputStr += nsPrefix;
		outputStr[outputStr.size()-1] = '=';	// Change the colon to =.
		outputStr += '"';
		outputStr += nsURI;
		outputStr += '"';

		usedNS += nsPrefix;

	}

}	// DeclareOneNamespace


// -------------------------------------------------------------------------------------------------
// DeclareElemNamespace
// --------------------

static void
DeclareElemNamespace ( const XMP_VarString & elemName,
					   XMP_VarString &		 usedNS,
					   XMP_VarString &		 outputStr,
					   XMP_StringPtr		 newline,
					   XMP_StringPtr		 indentStr,
					   XMP_Index			 indent )
{
	size_t colonPos = elemName.find ( ':' );

	if ( colonPos != XMP_VarString::npos ) {
		XMP_VarString nsPrefix ( elemName.substr ( 0, colonPos+1 ) );
		XMP_StringMapPos prefixPos = sNamespacePrefixToURIMap->find ( nsPrefix );
		XMP_Enforce ( prefixPos != sNamespacePrefixToURIMap->end() );
		DeclareOneNamespace ( nsPrefix, prefixPos->second, usedNS, outputStr, newline, indentStr, indent );
	}

}	// DeclareElemNamespace


// -------------------------------------------------------------------------------------------------
// DeclareUsedNamespaces
// ---------------------

// ??? Should iterators be passed by reference to avoid temp copies?

static void
DeclareUsedNamespaces ( const XMP_Node * currNode,
						XMP_VarString &  usedNS,
						XMP_VarString &	 outputStr,
						XMP_StringPtr	 newline,
						XMP_StringPtr	 indentStr,
						XMP_Index		 indent )
{

	if ( currNode->options & kXMP_SchemaNode ) {
		// The schema node name is the URI, the value is the prefix.
		DeclareOneNamespace ( currNode->value, currNode->name, usedNS, outputStr, newline, indentStr, indent );
	} else if ( currNode->options & kXMP_PropValueIsStruct ) {
		for ( size_t fieldNum = 0, fieldLim = currNode->children.size(); fieldNum < fieldLim; ++fieldNum ) {
			const XMP_Node * currField = currNode->children[fieldNum];
			DeclareElemNamespace ( currField->name, usedNS, outputStr, newline, indentStr, indent );
		}
	}

	for ( size_t childNum = 0, childLim = currNode->children.size(); childNum < childLim; ++childNum ) {
		const XMP_Node * currChild = currNode->children[childNum];
		DeclareUsedNamespaces ( currChild, usedNS, outputStr, newline, indentStr, indent );
	}

	for ( size_t qualNum = 0, qualLim = currNode->qualifiers.size(); qualNum < qualLim; ++qualNum ) {
		const XMP_Node * currQual = currNode->qualifiers[qualNum];
		DeclareElemNamespace ( currQual->name, usedNS, outputStr, newline, indentStr, indent );
		DeclareUsedNamespaces ( currQual, usedNS, outputStr, newline, indentStr, indent );
	}
	
}	// DeclareUsedNamespaces

// -------------------------------------------------------------------------------------------------
// EmitRDFArrayTag
// ---------------

// ??? Should iterators be passed by reference to avoid temp copies?

enum {
	kIsStartTag = true,
	kIsEndTag   = false
};

static void
EmitRDFArrayTag	( XMP_OptionBits  arrayForm,
				  XMP_VarString & outputStr,
				  XMP_StringPtr	  newline,
				  XMP_StringPtr	  indentStr,
				  XMP_Index		  indent,
				  XMP_Index		  arraySize,
				  bool			  isStartTag )
{
	if ( (! isStartTag) && (arraySize == 0) ) return;
	
	for ( XMP_Index level = indent; level > 0; --level ) outputStr += indentStr;
	if ( isStartTag ) {
		outputStr += "<rdf:";
	} else {
		outputStr += "</rdf:";
	}

	if ( arrayForm & kXMP_PropArrayIsAlternate ) {
		outputStr += "Alt";
	} else if ( arrayForm & kXMP_PropArrayIsOrdered ) {
		outputStr += "Seq";
	} else {
		outputStr += "Bag";
	}
	
	if ( isStartTag && (arraySize == 0) ) outputStr += '/';
	outputStr += '>';
	outputStr += newline;

}	// EmitRDFArrayTag


// -------------------------------------------------------------------------------------------------
// AppendNodeValue
// ---------------
//
// Append a property or qualifier value to the output with appropriate XML escaping. The escaped
// characters for elements and attributes are '&', '<', '>', and ASCII controls. In addition, '"' is
// escaped for attributes. For efficiency, this is done in a double loop. The outer loop makes sure
// the whole value is processed. The inner loop does a contiguous unescaped run followed by one
// escaped character (if we're not at the end).
//
// The escaping for ASCII controls other than tab, lf, and cr is "doubled". Apparently XML doesn't
// even allow the other controls as numeric character references. At any rate the expat that we have
// doesn't. So we'll serialize them as "&amp;#xnn;", which expat will turn into "&#xnn;", and which
// we will turn back into the control.
//
// An actual occurence of "&#x" in the value will be serialized as "&amp;#x26;#x", which expat will
// turn into "&#x26;#x", and which we will turn back into "&#x". This is only done if there are at
// least 6 bytes left in the string, the parsing conversion looks for "&#xnn;".

enum {
	kForAttribute = true,
	kForElement   = false
};

static void
AppendNodeValue ( XMP_VarString & outputStr, const XMP_VarString & value, bool forAttribute )
{

	unsigned char * runStart = (unsigned char *) value.c_str();
	unsigned char * runLimit  = runStart + value.size();
	unsigned char * runEnd;
	unsigned char   ch;
	
	while ( runStart < runLimit ) {
	
		for ( runEnd = runStart; runEnd < runLimit; ++runEnd ) {
			ch = *runEnd;
			if ( forAttribute && (ch == '"') ) break;
			if ( (ch < 0x20) || (ch == '&') || (ch == '<') || (ch == '>') ) break;
		}
		
		outputStr.append ( (char *) runStart, (runEnd - runStart) );
		
		if ( runEnd < runLimit ) {

			if ( ch >= 0x20 ) {

				if ( ch == '"' ) {
					outputStr += "&quot;";
				} else if ( ch == '<' ) {
					outputStr += "&lt;";
				} else if ( ch == '>' ) {
					outputStr += "&gt;";
				} else {
					XMP_Assert ( ch == '&' );
					if ( ((runLimit - runEnd) < 6) || (*(runEnd+1) != '#') || (*(runEnd+2) != 'x') ) {
						outputStr += "&amp;";
					} else {
						outputStr += "&amp;#x26;";
					}
				}

			} else {

				char hexBuf[16];
				XMP_Assert ( (strlen("&#xn;") == 5) && (strlen("&amp;#xnn;") == 10) );
				
				if ( (ch == kTab) || (ch == kLF) || (ch == kCR) ) {
					memcpy ( hexBuf, "&#xn;", 5 );	// AUDIT: Length of "&#xn;" is 5, hexBuf size is 16.
					hexBuf[3] = kHexDigits[ch];
					outputStr.append ( hexBuf, 5 );
				} else {
					memcpy ( hexBuf, "&amp;#xnn;", 10 );	// AUDIT: Length of "&amp;#xnn;" is 10, hexBuf size is 16.
					hexBuf[7] = kHexDigits[ch>>4];
					hexBuf[8] = kHexDigits[ch&0xF];
					outputStr.append ( hexBuf, 10 );
				}

			}

			++runEnd;

		}
		
		runStart = runEnd;
	
	}

}	// AppendNodeValue


// -------------------------------------------------------------------------------------------------
// CanBeRDFAttrProp
// ----------------

static bool
CanBeRDFAttrProp ( const XMP_Node * propNode )
{
	
	if ( propNode->name[0] == '[' ) return false;
	if ( ! propNode->qualifiers.empty() ) return false;
	if ( propNode->options & kXMP_PropValueIsURI ) return false;
	if ( propNode->options & kXMP_PropCompositeMask ) return false;
	
	return true;
	
}	// CanBeRDFAttrProp


// -------------------------------------------------------------------------------------------------
// IsRDFAttrQualifier
// ------------------

static XMP_StringPtr sAttrQualifiers[] = { "xml:lang", "rdf:resource", "rdf:ID", "rdf:bagID", "rdf:nodeID", "" };

static bool
IsRDFAttrQualifier ( XMP_VarString qualName )
{

	for ( size_t i = 0; *sAttrQualifiers[i] != 0; ++i ) {
		if ( qualName == sAttrQualifiers[i] ) return true;
	}

	return false;
	
}	// IsRDFAttrQualifier


// -------------------------------------------------------------------------------------------------
// SerializePrettyRDFProperty
// --------------------------
//
// Recursively handles the "value" for a node. It does not matter if it is a top level property, a
// field of a struct, or an item of an array.  The indent is that for the property element. An
// xml:lang qualifier is written as an attribute of the property start tag, not by itself forcing
// the qualified property form. The patterns below mostly ignore attribute qualifiers like xml:lang.
// Except for the one struct case, attribute qualifiers don't affect the output form.
//
//	<ns:UnqualifiedSimpleProperty>value</ns:UnqualifiedSimpleProperty>
//
//	<ns:UnqualifiedStructProperty rdf:parseType="Resource">	(If no rdf:resource qualifier)
//		... Fields, same forms as top level properties
//	</ns:UnqualifiedStructProperty>
//
//	<ns:ResourceStructProperty rdf:resource="URI"
//		... Fields as attributes
//	>
//
//	<ns:UnqualifiedArrayProperty>
//		<rdf:Bag> or Seq or Alt
//			... Array items as rdf:li elements, same forms as top level properties
//		</rdf:Bag>
//	</ns:UnqualifiedArrayProperty>
//
//	<ns:QualifiedProperty rdf:parseType="Resource">
//		<rdf:value> ... Property "value" following the unqualified forms ... </rdf:value>
//		... Qualifiers looking like named struct fields
//	</ns:QualifiedProperty>

static void
SerializePrettyRDFProperty ( const XMP_Node * propNode,
							 XMP_VarString &  outputStr,
							 XMP_StringPtr	  newline,
							 XMP_StringPtr	  indentStr,
							 XMP_Index		  indent,
							 bool			  emitAsRDFValue = false )
{
	XMP_Index level;
	bool emitEndTag   = true;
	bool indentEndTag = true;

	XMP_OptionBits propForm = propNode->options & kXMP_PropCompositeMask;

	// ------------------------------------------------------------------------------------------
	// Determine the XML element name. Open the start tag with the name and attribute qualifiers.
	
	XMP_StringPtr elemName = propNode->name.c_str();
	if ( emitAsRDFValue ) {
		elemName= "rdf:value";
	} else if ( *elemName == '[' ) {
		elemName = "rdf:li";
	}

	for ( level = indent; level > 0; --level ) outputStr += indentStr;
	outputStr += '<';
	outputStr += elemName;
	
	#define isCompact	false
	bool hasGeneralQualifiers = isCompact;	// Might also become true later.
	bool hasRDFResourceQual   = false;
	
	for ( size_t qualNum = 0, qualLim = propNode->qualifiers.size(); qualNum < qualLim; ++qualNum ) {
		const XMP_Node * currQual = propNode->qualifiers[qualNum];
		if ( ! IsRDFAttrQualifier ( currQual->name ) ) {
			hasGeneralQualifiers = true;
		} else {
			if ( currQual->name == "rdf:resource" ) hasRDFResourceQual = true;
			if ( ! emitAsRDFValue ) {
				outputStr += ' ';
				outputStr += currQual->name;
				outputStr += "=\"";
				AppendNodeValue ( outputStr, currQual->value, kForAttribute );
				outputStr += '"';
			}
		}
	}
	
	// --------------------------------------------------------
	// Process the property according to the standard patterns.
	
	if ( hasGeneralQualifiers && (! emitAsRDFValue) ) {
	
		// -----------------------------------------------------------------------------------------
		// This node has general, non-attribute, qualifiers. Emit using the qualified property form.
		// ! The value is output by a recursive call ON THE SAME NODE with emitAsRDFValue set.

		if ( hasRDFResourceQual ) {
			XMP_Throw ( "Can't mix rdf:resource and general qualifiers", kXMPErr_BadRDF );
		}
		
		outputStr += " rdf:parseType=\"Resource\">";
		outputStr += newline;

		SerializePrettyRDFProperty ( propNode, outputStr, newline, indentStr, indent+1, true );
		
		for ( size_t qualNum = 0, qualLim = propNode->qualifiers.size(); qualNum < qualLim; ++qualNum ) {
			const XMP_Node * currQual = propNode->qualifiers[qualNum];
			if ( IsRDFAttrQualifier ( currQual->name ) ) continue;
			SerializePrettyRDFProperty ( currQual, outputStr, newline, indentStr, indent+1 );
		}
		
	} else {

		// --------------------------------------------------------------------
		// This node has no general qualifiers. Emit using an unqualified form.
		
		if ( propForm == 0 ) {
		
			// --------------------------
			// This is a simple property.
			
			if ( propNode->options & kXMP_PropValueIsURI ) {
				outputStr += " rdf:resource=\"";
				AppendNodeValue ( outputStr, propNode->value, kForAttribute );
				outputStr += "\"/>";
				outputStr += newline;
				emitEndTag = false;
			} else if ( propNode->value.empty() ) {
				outputStr += "/>";
				outputStr += newline;
				emitEndTag = false;
			} else {
				outputStr += '>';
				AppendNodeValue ( outputStr, propNode->value, kForElement );
				indentEndTag = false;
			}
			
		} else if ( propForm & kXMP_PropValueIsArray ) {
		
			// This is an array.
			outputStr += '>';
			outputStr += newline;
			EmitRDFArrayTag ( propForm, outputStr, newline, indentStr, indent+1, propNode->children.size(), kIsStartTag );
			if ( XMP_ArrayIsAltText(propNode->options) ) NormalizeLangArray ( (XMP_Node*)propNode );
			for ( size_t childNum = 0, childLim = propNode->children.size(); childNum < childLim; ++childNum ) {
				const XMP_Node * currChild = propNode->children[childNum];
				SerializePrettyRDFProperty ( currChild, outputStr, newline, indentStr, indent+2 );
			}
			EmitRDFArrayTag ( propForm, outputStr, newline, indentStr, indent+1, propNode->children.size(), kIsEndTag );
		
		
		} else if ( ! hasRDFResourceQual ) {
		
			// This is a "normal" struct, use the rdf:parseType="Resource" form.
			XMP_Assert ( propForm & kXMP_PropValueIsStruct );
			if ( propNode->children.size() == 0 ) {
				outputStr += " rdf:parseType=\"Resource\"/>";
				outputStr += newline;
				emitEndTag = false;
			} else {
				outputStr += " rdf:parseType=\"Resource\">";
				outputStr += newline;
				for ( size_t childNum = 0, childLim = propNode->children.size(); childNum < childLim; ++childNum ) {
					const XMP_Node * currChild = propNode->children[childNum];
					SerializePrettyRDFProperty ( currChild, outputStr, newline, indentStr, indent+1 );
				}
			}
		
		} else {
		
			// This is a struct with an rdf:resource attribute, use the "empty property element" form.
			XMP_Assert ( propForm & kXMP_PropValueIsStruct );
			for ( size_t childNum = 0, childLim = propNode->children.size(); childNum < childLim; ++childNum ) {
				const XMP_Node * currChild = propNode->children[childNum];
				if ( ! CanBeRDFAttrProp ( currChild ) ) {
					XMP_Throw ( "Can't mix rdf:resource and complex fields", kXMPErr_BadRDF );
				}
				outputStr += newline;
				for ( level = indent+1; level > 0; --level ) outputStr += indentStr;
				outputStr += ' ';
				outputStr += currChild->name;
				outputStr += "=\"";
				outputStr += currChild->value;
				outputStr += '"';
			}
			outputStr += "/>";
			outputStr += newline;
			emitEndTag = false;
		
		}
		
	}
	
	// ----------------------------------
	// Emit the property element end tag.
	
	if ( emitEndTag ) {
		if ( indentEndTag ) for ( level = indent; level > 0; --level ) outputStr += indentStr;
		outputStr += "</";
		outputStr += elemName;
		outputStr += '>';
		outputStr += newline;
	}
	
}	// SerializePrettyRDFProperty


// -------------------------------------------------------------------------------------------------
// SerializePrettyRDFSchema
// ------------------------
//
// Each schema's properties are written in a separate rdf:Description element. All of the necessary
// namespaces are declared in the rdf:Description element. The baseIndent is the base level for the
// entire serialization, that of the x:xmpmeta element. An xml:lang qualifier is written as an
// attribute of the property start tag, not by itself forcing the qualified property form.
//
//	<rdf:Description rdf:about="TreeName"
//			xmlns:ns="URI" ... >
//
//		... The actual properties of the schema, see SerializePrettyRDFProperty
//
//		<!-- ns1:Alias is aliased to ns2:Actual -->  ... If alias comments are wanted
//
//	</rdf:Description>

static void
SerializePrettyRDFSchema ( const XMP_VarString & treeName,
						   const XMP_Node *		 schemaNode,
						   XMP_VarString &		 outputStr,
						   XMP_OptionBits		 options,
						   XMP_StringPtr		 newline,
						   XMP_StringPtr		 indentStr,
						   XMP_Index			 baseIndent )
{
	XMP_Assert ( schemaNode->options & kXMP_SchemaNode );
	XMP_Assert ( schemaNode->qualifiers.empty() );
	
	// Write the rdf:Description start tag with the namespace declarations.
	
	XMP_Index level;
	for ( level = baseIndent+2; level > 0; --level ) outputStr += indentStr;
	outputStr += kRDF_SchemaStart;
	outputStr += '"';
	outputStr += treeName;
	outputStr += '"';

	size_t totalLen = 8;	// Start at 8 for "xml:rdf:".
	XMP_cStringMapPos currPos = sNamespacePrefixToURIMap->begin();
	XMP_cStringMapPos endPos  = sNamespacePrefixToURIMap->end();
	for ( ; currPos != endPos; ++currPos ) totalLen += currPos->first.size();

	XMP_VarString usedNS;
	usedNS.reserve ( totalLen );
	usedNS = "xml:rdf:";
	DeclareUsedNamespaces ( schemaNode, usedNS, outputStr, newline, indentStr, baseIndent+4 );

	outputStr += ">";
	outputStr += newline;
	
	// Write alias comments, if wanted.

	if ( options & kXMP_WriteAliasComments ) {	// *** Hoist into a routine, used for Plain XMP also.

		#if 0	// *** Buggy, disable for now.
		
		XMP_cAliasMapPos aliasPos = sRegisteredAliasMap->begin();
		XMP_cAliasMapPos aliasEnd = sRegisteredAliasMap->end();
		
		for ( ; aliasPos != aliasEnd; ++aliasPos ) {

			size_t nsPos = aliasPos->first.find ( schemaNode->value );
			if ( nsPos == XMP_VarString::npos ) continue;
			XMP_Assert ( nsPos == 0 );

			for ( level = baseIndent+3; level > 0; --level ) outputStr += indentStr;

			outputStr += "<!-- ";
			outputStr += aliasPos->first;
			outputStr += " is aliased to ";
			for ( size_t step = 1, stepLim = aliasPos->second.size(); step != stepLim; ++step ) {
				outputStr += aliasPos->second[step].step;
			}
			outputStr += " -->";
			outputStr += newline;

		}
		
		#endif

	}
	
	// Write each of the schema's actual properties.
	for ( size_t propNum = 0, propLim = schemaNode->children.size(); propNum < propLim; ++propNum ) {
		const XMP_Node * currProp = schemaNode->children[propNum];
		SerializePrettyRDFProperty ( currProp, outputStr, newline, indentStr, baseIndent+3 );
	}
	
	// Write the rdf:Description end tag.
	for ( level = baseIndent+2; level > 0; --level ) outputStr += indentStr;
	outputStr += kRDF_SchemaEnd;
	outputStr += newline;

}	// SerializePrettyRDFSchema


// -------------------------------------------------------------------------------------------------
// SerializeCompactRDFAttrProps
// ----------------------------
//
// Write each of the parent's simple unqualified properties as an attribute. Returns true if all
// of the properties are written as attributes.

static bool
SerializeCompactRDFAttrProps ( const XMP_Node *	parentNode,
							   XMP_VarString &	outputStr,
							   XMP_StringPtr	newline,
							   XMP_StringPtr	indentStr,
							   XMP_Index		indent )
{
	size_t prop, propLim;
	bool allAreAttrs = true;

	for ( prop = 0, propLim = parentNode->children.size(); prop != propLim; ++prop ) {

		const XMP_Node * currProp = parentNode->children[prop];
		if ( ! CanBeRDFAttrProp ( currProp ) ) {
			allAreAttrs = false;
			continue;
		}
		
		outputStr += newline;
		for ( XMP_Index level = indent; level > 0; --level ) outputStr += indentStr;
		outputStr += currProp->name;
		outputStr += "=\"";
		AppendNodeValue ( outputStr, currProp->value, kForAttribute );
		outputStr += '"';

	}
	
	return allAreAttrs;

}	// SerializeCompactRDFAttrProps


// -------------------------------------------------------------------------------------------------
// SerializeCompactRDFElemProps
// ----------------------------
//
// Recursively handles the "value" for a node that must be written as an RDF property element. It
// does not matter if it is a top level property, a field of a struct, or an item of an array.  The
// indent is that for the property element. The patterns bwlow ignore attribute qualifiers such as
// xml:lang, they don't affect the output form.
//
//	<ns:UnqualifiedStructProperty-1
//		... The fields as attributes, if all are simple and unqualified
//	/>
//
//	<ns:UnqualifiedStructProperty-2 rdf:parseType="Resource">
//		... The fields as elements, if none are simple and unqualified
//	</ns:UnqualifiedStructProperty-2>
//
//	<ns:UnqualifiedStructProperty-3>
//		<rdf:Description
//			... The simple and unqualified fields as attributes
//		>
//			... The compound or qualified fields as elements
//		</rdf:Description>
//	</ns:UnqualifiedStructProperty-3>
//
//	<ns:UnqualifiedArrayProperty>
//		<rdf:Bag> or Seq or Alt
//			... Array items as rdf:li elements, same forms as top level properties
//		</rdf:Bag>
//	</ns:UnqualifiedArrayProperty>
//
//	<ns:QualifiedProperty rdf:parseType="Resource">
//		<rdf:value> ... Property "value" following the unqualified forms ... </rdf:value>
//		... Qualifiers looking like named struct fields
//	</ns:QualifiedProperty>

// *** Consider numbered array items, but has compatibility problems.
// *** Consider qualified form with rdf:Description and attributes.

static void
SerializeCompactRDFElemProps ( const XMP_Node *	parentNode,
							   XMP_VarString &	outputStr,
							   XMP_StringPtr	newline,
							   XMP_StringPtr	indentStr,
							   XMP_Index		indent )
{
	XMP_Index level;

	for ( size_t prop = 0, propLim = parentNode->children.size(); prop != propLim; ++prop ) {

		const XMP_Node * propNode = parentNode->children[prop];
		if ( CanBeRDFAttrProp ( propNode ) ) continue;

		bool emitEndTag = true;
		bool indentEndTag = true;

		XMP_OptionBits propForm = propNode->options & kXMP_PropCompositeMask;

		// -----------------------------------------------------------------------------------
		// Determine the XML element name, write the name part of the start tag. Look over the
		// qualifiers to decide on "normal" versus "rdf:value" form. Emit the attribute
		// qualifiers at the same time.
		
		XMP_StringPtr elemName = propNode->name.c_str();
		if ( *elemName == '[' ) elemName = "rdf:li";

		for ( level = indent; level > 0; --level ) outputStr += indentStr;
		outputStr += '<';
		outputStr += elemName;

		#define isCompact	false
		bool hasGeneralQualifiers = isCompact;	// Might also become true later.
		bool hasRDFResourceQual   = false;

		for ( size_t qualNum = 0, qualLim = propNode->qualifiers.size(); qualNum < qualLim; ++qualNum ) {
			const XMP_Node * currQual = propNode->qualifiers[qualNum];
			if ( ! IsRDFAttrQualifier ( currQual->name ) ) {
				hasGeneralQualifiers = true;
			} else {
				if ( currQual->name == "rdf:resource" ) hasRDFResourceQual = true;
				outputStr += ' ';
				outputStr += currQual->name;
				outputStr += "=\"";
				AppendNodeValue ( outputStr, currQual->value, kForAttribute );
				outputStr += '"';
			}
		}
		
		// --------------------------------------------------------
		// Process the property according to the standard patterns.
	
		if ( hasGeneralQualifiers ) {
		
			// -------------------------------------------------------------------------------------
			// The node has general qualifiers, ones that can't be attributes on a property element.
			// Emit using the qualified property pseudo-struct form. The value is output by a call
			// to SerializePrettyRDFProperty with emitAsRDFValue set.
			
			// *** We're losing compactness in the calls to SerializePrettyRDFProperty.
			// *** Should refactor to have SerializeCompactRDFProperty that does one node.

			outputStr += " rdf:parseType=\"Resource\">";
			outputStr += newline;

			SerializePrettyRDFProperty ( propNode, outputStr, newline, indentStr, indent+1, true );
		
			size_t qualNum = 0;
			size_t qualLim = propNode->qualifiers.size();
			if ( propNode->options & kXMP_PropHasLang ) ++qualNum;
			
			for ( ; qualNum < qualLim; ++qualNum ) {
				const XMP_Node * currQual = propNode->qualifiers[qualNum];
				SerializePrettyRDFProperty ( currQual, outputStr, newline, indentStr, indent+1 );
			}
			
		} else {

			// --------------------------------------------------------------------
			// This node has only attribute qualifiers. Emit as a property element.
			
			if ( propForm == 0 ) {
			
				// --------------------------
				// This is a simple property.
				
				if ( propNode->options & kXMP_PropValueIsURI ) {
					outputStr += " rdf:resource=\"";
					AppendNodeValue ( outputStr, propNode->value, kForAttribute );
					outputStr += "\"/>";
					outputStr += newline;
					emitEndTag = false;
				} else if ( propNode->value.empty() ) {
					outputStr += "/>";
					outputStr += newline;
					emitEndTag = false;
				} else {
					outputStr += '>';
					AppendNodeValue ( outputStr, propNode->value, kForElement );
					indentEndTag = false;
				}
				
			} else if ( propForm & kXMP_PropValueIsArray ) {

				// -----------------
				// This is an array.
				
				outputStr += '>';
				outputStr += newline;
				EmitRDFArrayTag ( propForm, outputStr, newline, indentStr, indent+1, propNode->children.size(), kIsStartTag );
			
				if ( XMP_ArrayIsAltText(propNode->options) ) NormalizeLangArray ( (XMP_Node*)propNode );
				SerializeCompactRDFElemProps ( propNode, outputStr, newline, indentStr, indent+2 );

				EmitRDFArrayTag ( propForm, outputStr, newline, indentStr, indent+1, propNode->children.size(), kIsEndTag );

			} else {

				// ----------------------
				// This must be a struct.
				
				XMP_Assert ( propForm & kXMP_PropValueIsStruct );

				bool hasAttrFields = false;
				bool hasElemFields = false;
			
				size_t field, fieldLim;
				for ( field = 0, fieldLim = propNode->children.size(); field != fieldLim; ++field ) {
					XMP_Node * currField = propNode->children[field];
					if ( CanBeRDFAttrProp ( currField ) ) {
						hasAttrFields = true;
						if ( hasElemFields ) break;	// No sense looking further.
					} else {
						hasElemFields = true;
						if ( hasAttrFields ) break;	// No sense looking further.
					}
				}
				
				if ( hasRDFResourceQual && hasElemFields ) {
					XMP_Throw ( "Can't mix rdf:resource qualifier and element fields", kXMPErr_BadRDF );
				}
				
				if ( propNode->children.size() == 0 ) {
				
					// Catch an empty struct as a special case. The case below would emit an empty
					// XML element, which gets reparsed as a simple property with an empty value.
					outputStr += " rdf:parseType=\"Resource\"/>";
					outputStr += newline;
					emitEndTag = false;
				
				} else if ( ! hasElemFields ) {

					// All fields can be attributes, use the emptyPropertyElt form.
					SerializeCompactRDFAttrProps ( propNode, outputStr, newline, indentStr, indent+1 );
					outputStr += "/>";
					outputStr += newline;
					emitEndTag = false;

				} else if ( ! hasAttrFields ) {
				
					// All fields must be elements, use the parseTypeResourcePropertyElt form.
					outputStr += " rdf:parseType=\"Resource\">";
					outputStr += newline;
					SerializeCompactRDFElemProps ( propNode, outputStr, newline, indentStr, indent+1 );
				
				} else {
				
					// Have a mix of attributes and elements, use an inner rdf:Description.
					outputStr += '>';
					outputStr += newline;
					for ( level = indent+1; level > 0; --level ) outputStr += indentStr;
					outputStr += "<rdf:Description";
					SerializeCompactRDFAttrProps ( propNode, outputStr, newline, indentStr, indent+2 );
					outputStr += ">";
					outputStr += newline;
					SerializeCompactRDFElemProps ( propNode, outputStr, newline, indentStr, indent+1 );
					for ( level = indent+1; level > 0; --level ) outputStr += indentStr;
					outputStr += kRDF_StructEnd;
					outputStr += newline;

				}

			}

		}
		
		// ----------------------------------
		// Emit the property element end tag.
		
		if ( emitEndTag ) {
			if ( indentEndTag ) for ( level = indent; level > 0; --level ) outputStr += indentStr;
			outputStr += "</";
			outputStr += elemName;
			outputStr += '>';
			outputStr += newline;
		}

	}
	
}	// SerializeCompactRDFElemProps


// -------------------------------------------------------------------------------------------------
// SerializeCompactRDFSchemas
// --------------------------
//
// All properties from all schema are written in a single rdf:Description element, as are all of the
// necessary namespace declarations. The baseIndent is the base level for the entire serialization,
// that of the x:xmpmeta element. The x:xmpmeta and rdf:RDF elements have already been written.
//
// Top level simple unqualified properties are written as attributes of the (only) rdf:Description
// element. Structs, arrays, and qualified properties are written by SerializeCompactRDFElemProp. An
// xml:lang qualifier on a simple property prevents the attribute form.
//
//	<rdf:Description rdf:about="TreeName"
//			xmlns:ns="URI" ...
//		ns:UnqualifiedSimpleProperty="value" ... >
//		... The remaining properties of the schema, see SerializeCompactRDFElemProps
//	</rdf:Description>

static void
SerializeCompactRDFSchemas ( const XMP_Node & xmpTree,
							 XMP_VarString &  outputStr,
							 XMP_StringPtr	  newline,
							 XMP_StringPtr	  indentStr,
							 XMP_Index		  baseIndent )
{
	XMP_Index level;
	size_t schema, schemaLim;
	
	// Begin the rdf:Description start tag.
	for ( level = baseIndent+2; level > 0; --level ) outputStr += indentStr;
	outputStr += kRDF_SchemaStart;
	outputStr += '"';
	outputStr += xmpTree.name;
	outputStr += '"';
	
	// Write all necessary xmlns attributes.
	
	size_t totalLen = 8;	// Start at 8 for "xml:rdf:".
	XMP_cStringMapPos currPos = sNamespacePrefixToURIMap->begin();
	XMP_cStringMapPos endPos  = sNamespacePrefixToURIMap->end();
	for ( ; currPos != endPos; ++currPos ) totalLen += currPos->first.size();

	XMP_VarString usedNS;
	usedNS.reserve ( totalLen );
	usedNS = "xml:rdf:";

	for ( schema = 0, schemaLim = xmpTree.children.size(); schema != schemaLim; ++schema ) {
		const XMP_Node * currSchema = xmpTree.children[schema];
		DeclareUsedNamespaces ( currSchema, usedNS, outputStr, newline, indentStr, baseIndent+4 );
	}
	
	// Write the top level "attrProps" and close the rdf:Description start tag.
	bool allAreAttrs = true;
	for ( schema = 0, schemaLim = xmpTree.children.size(); schema != schemaLim; ++schema ) {
		const XMP_Node * currSchema = xmpTree.children[schema];
		allAreAttrs &= SerializeCompactRDFAttrProps ( currSchema, outputStr, newline, indentStr, baseIndent+3 );
	}
	if ( ! allAreAttrs ) {
		outputStr += ">";
		outputStr += newline;
	} else {
		outputStr += "/>";
		outputStr += newline;
		return;	// ! Done if all properties in all schema are written as attributes.
	}

	// Write the remaining properties for each schema.
	for ( schema = 0, schemaLim = xmpTree.children.size(); schema != schemaLim; ++schema ) {
		const XMP_Node * currSchema = xmpTree.children[schema];
		SerializeCompactRDFElemProps ( currSchema, outputStr, newline, indentStr, baseIndent+3 );
	}
	
	// Write the rdf:Description end tag.
	// *** Elide the end tag if everything (all props in all schema) is an attr.
	for ( level = baseIndent+2; level > 0; --level ) outputStr += indentStr;
	outputStr += kRDF_SchemaEnd;
	outputStr += newline;

}	// SerializeCompactRDFSchemas


// -------------------------------------------------------------------------------------------------
// SerializeAsRDF
// --------------
//
//	<?xpacket begin... ?>
//	<x:xmpmeta xmlns:x=... >
//		<rdf:RDF xmlns:rdf=... >
//
//			... The properties, see SerializePrettyRDFSchema or SerializeCompactRDFSchemas
//
//		</rdf:RDF>
//	</x:xmpmeta>
//	<?xpacket end... ?>

// *** Need to strip empty arrays?
// *** Option to strip/keep empty structs?
// *** Need to verify handling of rdf:type qualifiers in pretty and compact.
// *** Need to verify round tripping of rdf:ID and similar qualifiers, see RDF 7.2.21.
// *** Check cases of rdf:resource plus explicit attr qualifiers (like xml:lang).

static void
SerializeAsRDF ( const XMPMeta & xmpObj,
				 XMP_VarString & headStr,	// Everything up to the padding.
				 XMP_VarString & tailStr,	// Everything after the padding.
				 XMP_OptionBits	 options,
				 XMP_StringPtr	 newline,
				 XMP_StringPtr	 indentStr,
				 XMP_Index		 baseIndent )
{
	const size_t treeNameLen = xmpObj.tree.name.size();
	const size_t indentLen   = strlen ( indentStr );

	// First estimate the worst case space and reserve room in the output string. This optimization
	// avoids reallocating and copying the output as it grows. The initial count does not look at
	// the values of properties, so it does not account for character entities, e.g. &#xA; for newline.
	// Since there can be a lot of these in things like the base 64 encoding of a large thumbnail,
	// inflate the count by 1/4 (easy to do) to accommodate.
	
	// *** Need to include estimate for alias comments.
	
	size_t outputLen = 2 * (strlen(kPacketHeader) + strlen(kRDF_XMPMetaStart) + strlen(kRDF_RDFStart) + 3*baseIndent*indentLen);
	
	for ( size_t schemaNum = 0, schemaLim = xmpObj.tree.children.size(); schemaNum < schemaLim; ++schemaNum ) {
		const XMP_Node * currSchema = xmpObj.tree.children[schemaNum];
		outputLen += 2*(baseIndent+2)*indentLen + strlen(kRDF_SchemaStart) + treeNameLen + strlen(kRDF_SchemaEnd) + 2;
		outputLen += EstimateRDFSize ( currSchema, baseIndent+2, indentLen );
	}
	
	outputLen += (outputLen >> 2);	// Inflate by 1/4, an empirical fudge factor.
	
	// Now generate the RDF into the head string as UTF-8.
	
	XMP_Index level;
	
	headStr.erase();
	headStr.reserve ( outputLen );
	
	// Write the packet header PI.
	if ( ! (options & kXMP_OmitPacketWrapper) ) {
		for ( level = baseIndent; level > 0; --level ) headStr += indentStr;
		headStr += kPacketHeader;
		headStr += newline;
	}

	// Write the xmpmeta element's start tag.
	for ( level = baseIndent; level > 0; --level ) headStr += indentStr;
	headStr += kRDF_XMPMetaStart;
	headStr += kXMP_VersionMessage "\">";
	headStr += newline;

	// Write the rdf:RDF start tag.
	for ( level = baseIndent+1; level > 0; --level ) headStr += indentStr;
	headStr += kRDF_RDFStart;
	headStr += newline;
	
	// Write all of the properties.
	if ( options & kXMP_UseCompactFormat ) {
		SerializeCompactRDFSchemas ( xmpObj.tree, headStr, newline, indentStr, baseIndent );
	} else {
		if ( xmpObj.tree.children.size() > 0 ) {
			for ( size_t schemaNum = 0, schemaLim = xmpObj.tree.children.size(); schemaNum < schemaLim; ++schemaNum ) {
				const XMP_Node * currSchema = xmpObj.tree.children[schemaNum];
				SerializePrettyRDFSchema ( xmpObj.tree.name, currSchema, headStr, options, newline, indentStr, baseIndent );
			}
		} else {
			for ( XMP_Index level = baseIndent+2; level > 0; --level ) headStr += indentStr;
			headStr += kRDF_SchemaStart;	// Special case an empty XMP object.
			headStr += '"';
			headStr += xmpObj.tree.name;
			headStr += "\"/>";
			headStr += newline;
		}
	}

	// Write the rdf:RDF end tag.
	for ( level = baseIndent+1; level > 0; --level ) headStr += indentStr;
	headStr += kRDF_RDFEnd;
	headStr += newline;

	// Write the xmpmeta end tag.
	for ( level = baseIndent; level > 0; --level ) headStr += indentStr;
	headStr += kRDF_XMPMetaEnd;
	headStr += newline;
	
	// Write the packet trailer PI into the tail string as UTF-8.
	tailStr.erase();
	if ( ! (options & kXMP_OmitPacketWrapper) ) {
		tailStr.reserve ( strlen(kPacketTrailer) + (strlen(indentStr) * baseIndent) );
		for ( level = baseIndent; level > 0; --level ) tailStr += indentStr;
		tailStr += kPacketTrailer;
		if ( options & kXMP_ReadOnlyPacket ) tailStr[tailStr.size()-4] = 'r';
	}
	
	// ! This assert is just a performance check, to see if the reserve was enough.
	XMP_Assert ( headStr.size() <= outputLen );
	
}	// SerializeAsRDF

// =================================================================================================
// Constructors
// ============


XMPMeta::XMPMeta() : tree(XMP_Node(0,"",0)), clientRefs(0), prevTkVer(0), xmlParser(0)
{
	// Nothing more to do, clientRefs is incremented in wrapper.
	#if XMP_TraceCTorDTor
		printf ( "Default construct XMPMeta @ %.8X\n", this );
	#endif
}	// XMPMeta

// -------------------------------------------------------------------------------------------------

XMPMeta::~XMPMeta() RELEASE_NO_THROW
{
	#if XMP_TraceCTorDTor
		printf ( "Destruct XMPMeta @ %.8X\n", this );
	#endif

	XMP_Assert ( this->clientRefs <= 0 );
	if ( xmlParser != 0 ) delete ( xmlParser );
	xmlParser = 0;

}	// ~XMPMeta


// =================================================================================================
// Class Static Functions
// ======================
//
//
// =================================================================================================

// -------------------------------------------------------------------------------------------------
// GetVersionInfo
// --------------

/* class-static */ void
XMPMeta::GetVersionInfo ( XMP_VersionInfo * info )
{

	memset ( info, 0, sizeof(XMP_VersionInfo) );
	
	info->major   = XMP_API_VERSION_MAJOR;
	info->minor   = XMP_API_VERSION_MINOR;
	info->micro   = XMP_API_VERSION_MICRO;
	info->isDebug = kXMP_DebugFlag;
	info->flags   = 0;	// ! None defined yet.
	info->message = kXMP_VersionMessage;
	
}	// GetVersionInfo

// -------------------------------------------------------------------------------------------------
// Initialize
// ----------

/* class-static */ bool
XMPMeta::Initialize()
{
	// Allocate and initialize static objects.
	
	++sXMP_InitCount;
	if ( sXMP_InitCount > 1 ) return true;
	
	#if TraceXMPLocking
		xmpOut = fopen ( "xmp.out", "w" );
		fprintf ( xmpOut, "XMP initializing\n" ); fflush ( xmpOut );
	#endif
	
	sExceptionMessage = new XMP_VarString();
	XMP_InitMutex ( &sXMPMetaLock );
    sOutputStr   = new XMP_VarString;

	xdefaultName = new XMP_VarString ( "x-default" );
	
	sNamespaceURIToPrefixMap	= new XMP_StringMap;
	sNamespacePrefixToURIMap	= new XMP_StringMap;
	sRegisteredAliasMap			= new XMP_AliasMap;
	
	InitializeUnicodeConversions();
	
	// Register standard namespaces and aliases.
	
	XMP_StringPtr	voidPtr;
	XMP_StringLen	voidLen;
	
	(void) RegisterNamespace ( kXMP_NS_XML, "xml", &voidPtr, &voidLen );
	(void) RegisterNamespace ( kXMP_NS_RDF, "rdf", &voidPtr, &voidLen );
	(void) RegisterNamespace ( kXMP_NS_DC, "dc", &voidPtr, &voidLen );
	(void) RegisterNamespace ( kXMP_NS_IPTCCore, "Iptc4xmpCore", &voidPtr, &voidLen );

	(void) RegisterNamespace ( kXMP_NS_XMP, "xap", &voidPtr, &voidLen );
	(void) RegisterNamespace ( kXMP_NS_PDF, "pdf", &voidPtr, &voidLen );
	(void) RegisterNamespace ( kXMP_NS_Photoshop, "photoshop", &voidPtr, &voidLen );
	(void) RegisterNamespace ( kXMP_NS_PSAlbum, "album", &voidPtr, &voidLen );
	(void) RegisterNamespace ( kXMP_NS_EXIF, "exif", &voidPtr, &voidLen );
	(void) RegisterNamespace ( kXMP_NS_EXIF_Aux, "aux", &voidPtr, &voidLen );
	(void) RegisterNamespace ( kXMP_NS_TIFF, "tiff", &voidPtr, &voidLen );
	(void) RegisterNamespace ( kXMP_NS_PNG, "png", &voidPtr, &voidLen );
	(void) RegisterNamespace ( kXMP_NS_JPEG, "jpeg", &voidPtr, &voidLen );
	(void) RegisterNamespace ( kXMP_NS_JP2K, "jp2k", &voidPtr, &voidLen );
	(void) RegisterNamespace ( kXMP_NS_CameraRaw, "crs", &voidPtr, &voidLen );

	(void) RegisterNamespace ( kXMP_NS_AdobeStockPhoto, "bmsp", &voidPtr, &voidLen );

	(void) RegisterNamespace ( kXMP_NS_XMP_Rights, "xapRights", &voidPtr, &voidLen );
	(void) RegisterNamespace ( kXMP_NS_XMP_MM, "xapMM", &voidPtr, &voidLen );
	(void) RegisterNamespace ( kXMP_NS_XMP_BJ, "xapBJ", &voidPtr, &voidLen );

	(void) RegisterNamespace ( kXMP_NS_XMP_Text, "xapT", &voidPtr, &voidLen );
	(void) RegisterNamespace ( kXMP_NS_XMP_PagedFile, "xapTPg", &voidPtr, &voidLen );
	(void) RegisterNamespace ( kXMP_NS_XMP_Graphics, "xapG", &voidPtr, &voidLen );
	(void) RegisterNamespace ( kXMP_NS_XMP_Image, "xapGImg", &voidPtr, &voidLen );
	(void) RegisterNamespace ( kXMP_NS_XMP_Font, "stFNT", &voidPtr, &voidLen );
	(void) RegisterNamespace ( kXMP_NS_XMP_Dimensions, "stDim", &voidPtr, &voidLen );
	(void) RegisterNamespace ( kXMP_NS_XMP_ResourceEvent, "stEvt", &voidPtr, &voidLen );
	(void) RegisterNamespace ( kXMP_NS_XMP_ResourceRef, "stRef", &voidPtr, &voidLen );
	(void) RegisterNamespace ( kXMP_NS_XMP_ST_Version, "stVer", &voidPtr, &voidLen );
	(void) RegisterNamespace ( kXMP_NS_XMP_ST_Job, "stJob", &voidPtr, &voidLen );
	(void) RegisterNamespace ( kXMP_NS_XMP_ManifestItem, "stMfs", &voidPtr, &voidLen );
	(void) RegisterNamespace ( kXMP_NS_XMP_IdentifierQual, "xmpidq", &voidPtr, &voidLen );
	
	(void) RegisterNamespace ( "adobe:ns:meta/", "x", &voidPtr, &voidLen );
	(void) RegisterNamespace ( "http://ns.adobe.com/iX/1.0/", "iX", &voidPtr, &voidLen );
	
	XMPMeta::RegisterStandardAliases ( "" );
	
	// Initialize the other core classes.
	
	if ( ! XMPIterator::Initialize() ) XMP_Throw ( "Failure from XMPIterator::Initialize", kXMPErr_InternalFailure );
	if ( ! XMPUtils::Initialize() ) XMP_Throw ( "Failure from XMPUtils::Initialize", kXMPErr_InternalFailure );

	// Do miscelaneous semantic checks of types and arithmetic.
	
	XMP_Assert ( sizeof(XMP_Int8) == 1 );
	XMP_Assert ( sizeof(XMP_Int16) == 2 );
	XMP_Assert ( sizeof(XMP_Int32) == 4 );
	XMP_Assert ( sizeof(XMP_Int64) == 8 );
	XMP_Assert ( sizeof(XMP_Uns8) == 1 );
	XMP_Assert ( sizeof(XMP_Uns16) == 2 );
	XMP_Assert ( sizeof(XMP_Uns32) == 4 );
	XMP_Assert ( sizeof(XMP_Uns64) == 8 );
	
	XMP_Assert ( sizeof(XMP_OptionBits) == 4 );	// Check that option masking work on all 32 bits.
	XMP_OptionBits flag = ~0UL;
	XMP_Assert ( flag == (XMP_OptionBits)(-1L) );
	XMP_Assert ( (flag ^ kXMP_PropHasLang) == 0xFFFFFFBFUL );
	XMP_Assert ( (flag & ~kXMP_PropHasLang) == 0xFFFFFFBFUL );
	
	XMP_OptionBits opt1 = 0;	// Check the general option bit macros.
	XMP_OptionBits opt2 = ~0UL;
	XMP_SetOption ( opt1, kXMP_PropValueIsArray );
	XMP_ClearOption ( opt2, kXMP_PropValueIsArray );
	XMP_Assert ( opt1 == ~opt2 );
	XMP_Assert ( XMP_TestOption ( opt1, kXMP_PropValueIsArray ) );
	XMP_Assert ( ! XMP_TestOption ( opt2, kXMP_PropValueIsArray ) );
	
	XMP_Assert ( XMP_PropIsSimple ( ~kXMP_PropCompositeMask ) );	// Check the special option bit macros.
	XMP_Assert ( ! XMP_PropIsSimple ( kXMP_PropValueIsStruct ) );
	XMP_Assert ( ! XMP_PropIsSimple ( kXMP_PropValueIsArray ) );

	XMP_Assert ( XMP_PropIsStruct ( kXMP_PropValueIsStruct ) );
	XMP_Assert ( XMP_PropIsArray  ( kXMP_PropValueIsArray ) );
	XMP_Assert ( ! XMP_PropIsStruct ( ~kXMP_PropValueIsStruct ) );
	XMP_Assert ( ! XMP_PropIsArray  ( ~kXMP_PropValueIsArray ) );
	
	XMP_Assert ( XMP_ArrayIsUnordered ( ~kXMP_PropArrayIsOrdered ) );
	XMP_Assert ( XMP_ArrayIsOrdered   ( kXMP_PropArrayIsOrdered ) );
	XMP_Assert ( XMP_ArrayIsAlternate ( kXMP_PropArrayIsAlternate ) );
	XMP_Assert ( XMP_ArrayIsAltText   ( kXMP_PropArrayIsAltText ) );
	XMP_Assert ( ! XMP_ArrayIsUnordered ( kXMP_PropArrayIsOrdered ) );
	XMP_Assert ( ! XMP_ArrayIsOrdered   ( ~kXMP_PropArrayIsOrdered ) );
	XMP_Assert ( ! XMP_ArrayIsAlternate ( ~kXMP_PropArrayIsAlternate ) );
	XMP_Assert ( ! XMP_ArrayIsAltText   ( ~kXMP_PropArrayIsAltText ) );
	
	XMP_Assert ( XMP_PropHasQualifiers ( kXMP_PropHasQualifiers ) );
	XMP_Assert ( XMP_PropIsQualifier   ( kXMP_PropIsQualifier ) );
	XMP_Assert ( XMP_PropHasLang       ( kXMP_PropHasLang ) );
	XMP_Assert ( ! XMP_PropHasQualifiers ( ~kXMP_PropHasQualifiers ) );
	XMP_Assert ( ! XMP_PropIsQualifier   ( ~kXMP_PropIsQualifier ) );
	XMP_Assert ( ! XMP_PropHasLang       ( ~kXMP_PropHasLang ) );
	
	XMP_Assert ( XMP_NodeIsSchema ( kXMP_SchemaNode ) );
	XMP_Assert ( XMP_PropIsAlias  ( kXMP_PropIsAlias ) );
	XMP_Assert ( ! XMP_NodeIsSchema ( ~kXMP_SchemaNode ) );
	XMP_Assert ( ! XMP_PropIsAlias  ( ~kXMP_PropIsAlias ) );
	
	#if 0	// Generally off, enable to hand check generated code.
		extern XMP_OptionBits opt3, opt4;
		if ( XMP_TestOption ( opt3, kXMP_PropValueIsArray ) ) opt4 = opt3;
		if ( ! XMP_TestOption ( opt3, kXMP_PropValueIsStruct ) ) opt4 = opt3;
		static bool ok1 = XMP_TestOption ( opt4, kXMP_PropValueIsArray );
		static bool ok2 = ! XMP_TestOption ( opt4, kXMP_PropValueIsStruct );
	#endif
	
	return true;

}	// Initialize


// -------------------------------------------------------------------------------------------------
// Terminate
// ---------

/* class-static */ void
XMPMeta::Terminate() RELEASE_NO_THROW
{
	--sXMP_InitCount;
	if ( sXMP_InitCount > 0 ) return;

	#if TraceXMPLocking
		fprintf ( xmpOut, "XMP terminating\n" ); fflush ( xmpOut );
		fclose ( xmpOut );
	#endif
	
	XMPIterator::Terminate();
	XMPUtils::Terminate();

	delete sNamespaceURIToPrefixMap;
	delete sNamespacePrefixToURIMap;
	delete sRegisteredAliasMap;
	
	sNamespaceURIToPrefixMap	= 0;
	sNamespacePrefixToURIMap	= 0;
	sRegisteredAliasMap			= 0;
    
    delete ( xdefaultName );
    xdefaultName = 0;
    
    delete ( sOutputStr );
    sOutputStr = 0;

	XMP_TermMutex ( sXMPMetaLock );

	delete ( sExceptionMessage );
	sExceptionMessage = 0;
	
}	// Terminate


// -------------------------------------------------------------------------------------------------
// Unlock
// ------

/* class-static */ void
XMPMeta::Unlock ( XMP_OptionBits options )
{
	options = options;	// Avoid unused parameter warning.	// *** Need IgnoreParam macro.

    #if TraceXMPLocking
    	fprintf ( xmpOut, "  Unlocking XMP toolkit, count = %d\n", sLockCount ); fflush ( xmpOut );
	#endif
    --sLockCount;
    XMP_Assert ( sLockCount == 0 );
	XMP_ExitCriticalRegion ( sXMPMetaLock );

}	// Unlock


// -------------------------------------------------------------------------------------------------
// UnlockObject
// ------------

void
XMPMeta::UnlockObject ( XMP_OptionBits options ) const
{
	options = options;	// Avoid unused parameter warning.

	XMPMeta::Unlock ( 0 );

}	// UnlockObject


// -------------------------------------------------------------------------------------------------
// RegisterAssertNotify
// --------------------

/* class-static */ void
XMPMeta::RegisterAssertNotify ( XMP_AssertWrapperProc wrapperProc, 
                                XMP_AssertNotifyProc  notifyProc,
                                void *                refCon )
{

	if ( sAssertNotify == 0 ) {
		sAssertWrapper = wrapperProc;
		sAssertNotify  = notifyProc;
		sAssertRefCon  = refCon;
	}

}	// RegisterAssertNotify


// -------------------------------------------------------------------------------------------------
// UnregisterAssertNotify
// ----------------------

/* class-static */ void
XMPMeta::UnregisterAssertNotify	( XMP_AssertNotifyProc notifyProc )
{

	if ( (notifyProc == 0) || (notifyProc == sAssertNotify) ) {
		sAssertWrapper = 0;
		sAssertNotify  = 0;
		sAssertRefCon  = 0;
	}

}	// UnregisterAssertNotify


// -------------------------------------------------------------------------------------------------
// SendAssertNotify
// ----------------

/* class-static */ void
XMPMeta::SendAssertNotify ( XMP_StringPtr message )
{

	if ( sAssertNotify == 0 ) {
		assert ( false );	// *** Would be nice to have a way to expose the message.
	} else {
		sAssertWrapper ( sAssertNotify, sAssertRefCon, message );
	}

}	// SendAssertNotify


// -------------------------------------------------------------------------------------------------
// DumpNamespaces
// --------------
//
// Dump the prefix to URI map (easier to read) and verify that both are consistent and legit.

// *** Should put checks in a separate routine for regular calling in debug builds.

/* class-static */ XMP_Status
XMPMeta::DumpNamespaces ( XMP_TextOutputProc outProc,
                          void *             refCon )
{
	XMP_Assert ( outProc != 0 );	// ! Enforced by wrapper.
	XMP_Status status;
	
	XMP_StringMapPos p2uEnd = sNamespacePrefixToURIMap->end();	// ! Move up to avoid gcc complaints.
	XMP_StringMapPos u2pEnd = sNamespaceURIToPrefixMap->end();
	
	status = DumpStringMap ( *sNamespacePrefixToURIMap, "Dumping namespace prefix to URI map", outProc, refCon );
	if ( status != 0 ) goto EXIT;
	
	if ( sNamespacePrefixToURIMap->size() != sNamespaceURIToPrefixMap->size() ) {
		OutProcLiteral ( "** bad namespace map sizes **" );
		XMP_Throw ( "Fatal namespace map problem", kXMPErr_InternalFailure );
	}
	
	for ( XMP_StringMapPos nsLeft = sNamespacePrefixToURIMap->begin(); nsLeft != p2uEnd; ++nsLeft ) {

		XMP_StringMapPos nsOther = sNamespaceURIToPrefixMap->find ( nsLeft->second );
		if ( (nsOther == u2pEnd) || (nsLeft != sNamespacePrefixToURIMap->find ( nsOther->second )) ) {
			OutProcLiteral ( "  ** bad namespace URI **  " );
			OutProcString ( nsLeft->second );
			goto FAILURE;
		}
		
		for ( XMP_StringMapPos nsRight = nsLeft; nsRight != p2uEnd; ++nsRight ) {
			if ( nsRight == nsLeft ) continue;	// ! Can't start at nsLeft+1, no operator+!
			if ( nsLeft->second == nsRight->second ) {
				OutProcLiteral ( "  ** duplicate namespace URI **  " );
				OutProcString ( nsLeft->second );
				goto FAILURE;
			}
		}

	}
	
	for ( XMP_StringMapPos nsLeft = sNamespaceURIToPrefixMap->begin(); nsLeft != u2pEnd; ++nsLeft ) {

		XMP_StringMapPos nsOther = sNamespacePrefixToURIMap->find ( nsLeft->second );
		if ( (nsOther == u2pEnd) || (nsLeft != sNamespaceURIToPrefixMap->find ( nsOther->second )) ) {
			OutProcLiteral ( "  ** bad namespace prefix **  " );
			OutProcString ( nsLeft->second );
			goto FAILURE;
		}

		for ( XMP_StringMapPos nsRight = nsLeft; nsRight != u2pEnd; ++nsRight ) {
			if ( nsRight == nsLeft ) continue;	// ! Can't start at nsLeft+1, no operator+!
			if ( nsLeft->second == nsRight->second ) {
				OutProcLiteral ( "  ** duplicate namespace prefix **  " );
				OutProcString ( nsLeft->second );
				goto FAILURE;
			}
		}

	}

EXIT:
	return status;

FAILURE:
	OutProcNewline();
	(void) DumpStringMap ( *sNamespaceURIToPrefixMap, "Dumping namespace URI to prefix map", outProc, refCon );
	XMP_Throw ( "Fatal namespace map problem", kXMPErr_InternalFailure );
	return 0;
	
}	// DumpNamespaces


// -------------------------------------------------------------------------------------------------
// DumpAliases
// -----------

/* class-static */ XMP_Status
XMPMeta::DumpAliases ( XMP_TextOutputProc outProc,
                       void *             refCon )
{
	XMP_Assert ( outProc != 0 );	// ! Enforced by wrapper.
	XMP_Status status;

	XMP_Assert ( sRegisteredAliasMap != 0 );

	XMP_cAliasMapPos aliasPos;
	XMP_cAliasMapPos aliasEnd = sRegisteredAliasMap->end();
	
	size_t maxLen = 0;
	for ( aliasPos = sRegisteredAliasMap->begin(); aliasPos != aliasEnd; ++aliasPos ) {
		size_t currLen = aliasPos->first.size();
		if ( currLen > maxLen ) maxLen = currLen;
	}
	
	OutProcLiteral ( "Dumping alias name to actual path map" );
	OutProcNewline();
		
	for ( aliasPos = sRegisteredAliasMap->begin(); aliasPos != aliasEnd; ++aliasPos ) {

		OutProcNChars ( "   ", 3 );
		OutProcString ( aliasPos->first );
		OutProcPadding ( maxLen - aliasPos->first.size() );
		OutProcNChars ( " => ", 4 );

		size_t actualPathSize = aliasPos->second.size();
		for ( size_t stepNum = 1; stepNum < actualPathSize; ++stepNum ) OutProcString ( aliasPos->second[stepNum].step );

		XMP_OptionBits arrayForm = aliasPos->second[1].options & kXMP_PropArrayFormMask;

		if ( arrayForm == 0 ) {
			if ( actualPathSize != 2 ) OutProcLiteral ( "  ** bad actual path **" );
		} else {
			OutProcNChars ( "  ", 2 );
			DumpNodeOptions ( arrayForm, outProc, refCon );
			if ( ! (arrayForm & kXMP_PropValueIsArray) ) OutProcLiteral ( "  ** bad array form **" );
			if ( actualPathSize != 3 ) OutProcLiteral ( "  ** bad actual path **" );
		}

		if ( aliasPos->second[0].options != kXMP_SchemaNode ) OutProcLiteral ( "  ** bad schema form **" );

		OutProcNewline();

	}
	
EXIT:
	return status;
	
}	// DumpAliases


// -------------------------------------------------------------------------------------------------
// DumpPropertyTraits
// ------------------

/* class-static */ XMP_Status
XMPMeta::DumpPropertyTraits ( XMP_TextOutputProc outProc,
                              void *             refCon )
{
	XMP_Assert ( outProc != 0 );	// ! Enforced by wrapper.

	XMP_Throw ( "Unimplemented method XMPMeta::DumpPropertyTraits", kXMPErr_Unimplemented );
		void * p; p = &outProc; p = &refCon;	// Avoid unused param warnings.
		return -1;	// Avoid no-result warning.

}	// DumpPropertyTraits


// -------------------------------------------------------------------------------------------------
// GetGlobalOptions
// ----------------

/* class-static */ XMP_OptionBits
XMPMeta::GetGlobalOptions()
{
	XMP_OptionBits	options = 0;
	
	return options;
	
}	// GetGlobalOptions


// -------------------------------------------------------------------------------------------------
// SetGlobalOptions
// ----------------

/* class-static */ void
XMPMeta::SetGlobalOptions ( XMP_OptionBits options )
{

	XMP_Throw ( "Unimplemented method XMPMeta::SetGlobalOptions", kXMPErr_Unimplemented );
		void * p; p = &options;	// Avoid unused param warnings.

}	// SetGlobalOptions


// -------------------------------------------------------------------------------------------------
// RegisterNamespace
// -----------------

/* class-static */ bool
XMPMeta::RegisterNamespace ( XMP_StringPtr	 namespaceURI,
							 XMP_StringPtr	 suggestedPrefix,
							 XMP_StringPtr * registeredPrefix,
							 XMP_StringLen * prefixSize )
{
	bool	prefixMatches	= false;
	
	XMP_Assert ( (registeredPrefix != 0) && (prefixSize != 0) );		// ! Enforced by wrapper.
	if ( (*namespaceURI == 0) || (*suggestedPrefix == 0) ) {
		XMP_Throw ( "Empty namespace URI or prefix", kXMPErr_BadParam );
	}
	
	XMP_VarString	nsURI ( namespaceURI );
	XMP_VarString	suggPrefix ( suggestedPrefix );
	if ( suggPrefix[suggPrefix.size()-1] != ':' ) suggPrefix += ':';
	VerifySimpleXMLName ( suggestedPrefix, suggestedPrefix+suggPrefix.size()-1 );	// Exclude the colon.
	
	XMP_StringMapPos	uriPos	= sNamespaceURIToPrefixMap->find ( nsURI );
	
	if ( uriPos == sNamespaceURIToPrefixMap->end() ) {
		
		// The URI is not yet registered, make sure we use a unique prefix.
		
		XMP_VarString	uniqPrefix ( suggPrefix );
		int		suffix = 0;
		char	buffer [32];

		while ( true ) {
			if ( sNamespacePrefixToURIMap->find ( uniqPrefix ) == sNamespacePrefixToURIMap->end() ) break;
			++suffix;
			snprintf ( buffer, sizeof(buffer), "_%d_:", suffix );	// AUDIT: Using sizeof for snprintf length is safe.
			uniqPrefix = suggPrefix;
			uniqPrefix.erase ( uniqPrefix.size()-1 );	// ! Remove the trailing ':'.
			uniqPrefix += buffer;
		}
		
		// Add the new namespace to both maps.
		
		XMP_StringPair	newNS ( nsURI, uniqPrefix );
		uriPos = sNamespaceURIToPrefixMap->insert ( sNamespaceURIToPrefixMap->end(), newNS );
		
		newNS.first.swap ( newNS.second );
		(void) sNamespacePrefixToURIMap->insert ( sNamespacePrefixToURIMap->end(), newNS );

	}
	
	// Return the actual prefix and see if it matches the suggested prefix.
	
	*registeredPrefix = uriPos->second.c_str();
	*prefixSize = uriPos->second.size();
	
	prefixMatches = ( uriPos->second == suggPrefix );
	return prefixMatches;
	
}	// RegisterNamespace


// -------------------------------------------------------------------------------------------------
// GetNamespacePrefix
// ------------------

/* class-static */ bool
XMPMeta::GetNamespacePrefix	( XMP_StringPtr   namespaceURI,
							  XMP_StringPtr * namespacePrefix,
							  XMP_StringLen * prefixSize )
{
	bool	found	= false;
	
	XMP_Assert ( *namespaceURI != 0 ); 	// ! Enforced by wrapper.
	XMP_Assert ( (namespacePrefix != 0) && (prefixSize != 0) );	// ! Enforced by wrapper.

	XMP_VarString    nsURI ( namespaceURI );
	XMP_StringMapPos uriPos	= sNamespaceURIToPrefixMap->find ( nsURI );
	
	if ( uriPos != sNamespaceURIToPrefixMap->end() ) {
		*namespacePrefix = uriPos->second.c_str();
		*prefixSize = uriPos->second.size();
		found = true;
	}
	
	return found;
	
}	// GetNamespacePrefix


// -------------------------------------------------------------------------------------------------
// GetNamespaceURI
// ---------------

/* class-static */ bool
XMPMeta::GetNamespaceURI ( XMP_StringPtr   namespacePrefix,
						   XMP_StringPtr * namespaceURI,
						   XMP_StringLen * uriSize )
{
	bool	found	= false;
	
	XMP_Assert ( *namespacePrefix != 0 ); 	// ! Enforced by wrapper.
	XMP_Assert ( (namespacePrefix != 0) && (namespaceURI != 0) );	// ! Enforced by wrapper.

	XMP_VarString nsPrefix ( namespacePrefix );
	if ( nsPrefix[nsPrefix.size()-1] != ':' ) nsPrefix += ':';
	
	XMP_StringMapPos prefixPos = sNamespacePrefixToURIMap->find ( nsPrefix );
	
	if ( prefixPos != sNamespacePrefixToURIMap->end() ) {
		*namespaceURI = prefixPos->second.c_str();
		*uriSize = prefixPos->second.size();
		found = true;
	}
	
	return found;
	
}	// GetNamespaceURI


// -------------------------------------------------------------------------------------------------
// DeleteNamespace
// ---------------

// *** Don't allow standard namespaces to be deleted.

/* class-static */ void
XMPMeta::DeleteNamespace ( XMP_StringPtr namespaceURI )
{
	namespaceURI = namespaceURI;	// Avoid unused parameter warning.
	XMP_Assert ( *namespaceURI != 0 ); 	// ! Enforced by wrapper.
	XMP_Throw ( "Unimplemented method XMPMeta::DeleteNamespace", kXMPErr_Unimplemented );   // *** #error "write me"

}	// DeleteNamespace


// -------------------------------------------------------------------------------------------------
// RegisterAlias
// -------------
//
// Allow 3 kinds of alias: 
//	TopProp => TopProp
//	TopProp => TopArray[1]
//	TopProp => TopArray[@xml:lang='x-default']
//
// A new alias can be made to something that is already aliased, as long as the net result is one of
// the legitimate forms. The new alias can already have aliases to it, also as long as result of
// adjusting all of the exiting aliases leaves them legal.
//
// ! The caller assumes all risk that new aliases do not invalidate existing XMPMeta objects. Any
// ! conflicts will result in later references throwing bad XPath exceptions.

/* class-static */ void
XMPMeta::RegisterAlias ( XMP_StringPtr 	aliasNS,
						 XMP_StringPtr	aliasProp,
						 XMP_StringPtr	actualNS,
						 XMP_StringPtr	actualProp,
						 XMP_OptionBits arrayForm )
{
	XMP_ExpandedXPath	expAlias, expActual;
	XMP_AliasMapPos		mapPos;
	XMP_ExpandedXPath *	regActual = 0;

	XMP_Assert ( (aliasNS != 0) && (aliasProp != 0) && (actualNS != 0) && (actualProp != 0) );	// Enforced by wrapper.
		
	// Expand the alias and actual names, make sure they are one of the basic 3 forms. When counting
	// the expanded XPath size remember that the schema URI is the first component. We don't have to
	// compare the schema URIs though, the (unique) prefix is part of the top property name.
	
	ExpandXPath ( aliasNS, aliasProp, &expAlias );
	ExpandXPath ( actualNS, actualProp, &expActual );
	if ( (expAlias.size() != 2) || (expActual.size() != 2) ) {
		XMP_Throw ( "Alias and actual property names must be simple", kXMPErr_BadXPath );
	}
	
	arrayForm = VerifySetOptions ( arrayForm, 0 );
	if ( arrayForm != 0 ) {
		if ( (arrayForm & ~kXMP_PropArrayFormMask) != 0 ) XMP_Throw ( "Only array form flags are allowed", kXMPErr_BadOptions );
		expActual[1].options |= arrayForm;	// Set the array form for the top level step.
		if ( ! (arrayForm & kXMP_PropArrayIsAltText) ) {
			expActual.push_back ( XPathStepInfo ( "[1]", kXMP_ArrayIndexStep ) );
		} else {
			expActual.push_back ( XPathStepInfo ( "[?xml:lang=\"x-default\"]", kXMP_QualSelectorStep ) );
		}
	}

	// See if there are any conflicts with existing aliases. A couple of the checks are easy. If the
	// alias is already aliased it is only OK to reregister an identical alias. If the actual is
	// already aliased to something else and the new chain is legal, just swap in the old base.

	mapPos = sRegisteredAliasMap->find ( expAlias[kRootPropStep].step );
	if ( mapPos != sRegisteredAliasMap->end() ) {

		// This alias is already registered to something, make sure it is the same something.

		regActual = &mapPos->second;
		if ( arrayForm != (mapPos->second[1].options & kXMP_PropArrayFormMask) ) {
			XMP_Throw ( "Mismatch with existing alias array form", kXMPErr_BadParam );
		}
		if ( expActual.size() != regActual->size() ) {
			XMP_Throw ( "Mismatch with existing actual path", kXMPErr_BadParam );
		}
		if ( expActual[kRootPropStep].step != (*regActual)[kRootPropStep].step ) {
			XMP_Throw ( "Mismatch with existing actual name", kXMPErr_BadParam );
		}
		if ( (expActual.size() == 3) && (expActual[kAliasIndexStep].step != (*regActual)[kAliasIndexStep].step) ) {
			XMP_Throw ( "Mismatch with existing actual array item", kXMPErr_BadParam );
		}
		return;

	}

	mapPos = sRegisteredAliasMap->find ( expActual[kRootPropStep].step );
	if ( mapPos != sRegisteredAliasMap->end() ) {
	
		// The actual is already aliased to something else.
		
		regActual = &mapPos->second;
		if ( expActual.size() == 2 ) {
			expActual = *regActual;	// TopProp => TopProp => anything : substitute the entire old base.
		} else if ( regActual->size() != 2 ) {
			XMP_Throw ( "Can't alias an array item to an array item", kXMPErr_BadParam );	// TopProp => TopArray[] => TopArray[] : nope.
		} else {
			expActual[kSchemaStep].step = (*regActual)[kSchemaStep].step;		// TopProp => TopArray[] => TopProp :
			expActual[kRootPropStep].step   = (*regActual)[kRootPropStep].step;	//	substitute the old base name.
		}
	
	}
	
	// Checking for existing aliases to this one is touchier. This involves updating the alias map,
	// which must not be done unless all of the changes are legal. So we need 2 loops, one to verify
	// that everything is OK, and one to make the changes. The bad case is:
	//	TopProp => TopArray[] => TopArray[]
	// In the valid cases we back substitute the new base.
	
	for ( mapPos = sRegisteredAliasMap->begin(); mapPos != sRegisteredAliasMap->end(); ++mapPos ) {
		regActual = &mapPos->second;
		if ( expAlias[kRootPropStep].step == (*regActual)[kRootPropStep].step ) {
			if ( (regActual->size() == 2) && (expAlias.size() == 2) ) {
				XMP_Throw ( "Can't alias an array item to an array item", kXMPErr_BadParam );
			}
		}
	}
	
	for ( mapPos = sRegisteredAliasMap->begin(); mapPos != sRegisteredAliasMap->end(); ++mapPos ) {
		regActual = &mapPos->second;
		if ( expAlias[kRootPropStep].step == (*regActual)[kRootPropStep].step ) {

			if ( regActual->size() == 1 ) {
				*regActual = expActual;	// TopProp => TopProp => anything : substitute the entire new base.
			} else {
				(*regActual)[kSchemaStep].step = expActual[kSchemaStep].step;	// TopProp => TopArray[] => TopProp :
				(*regActual)[kRootPropStep].step   = expActual[kRootPropStep].step;		//	substitute the new base name.
			}

		}
	}
	
	// Finally, all is OK to register the new alias.
	
	(void) sRegisteredAliasMap->insert ( XMP_AliasMap::value_type ( expAlias[kRootPropStep].step, expActual ) );

}	// RegisterAlias


// -------------------------------------------------------------------------------------------------
// ResolveAlias
// ------------

/* class-static */ bool
XMPMeta::ResolveAlias ( XMP_StringPtr	 aliasNS,
						XMP_StringPtr	 aliasProp,
						XMP_StringPtr *	 actualNS,
						XMP_StringLen *	 nsSize,
						XMP_StringPtr *	 actualProp,
						XMP_StringLen *	 propSize,
						XMP_OptionBits * arrayForm )
{
	XMP_Assert ( (aliasNS != 0) && (aliasProp != 0) );	// Enforced by wrapper.
	XMP_Assert ( (actualNS != 0) && (nsSize != 0) && (actualProp != 0) && (propSize != 0) && (arrayForm != 0) );	// Enforced by wrapper.
		
	XMP_ExpandedXPath expPath;
	ExpandXPath ( aliasNS, aliasProp, &expPath );
	if ( expPath.size() != 2 ) XMP_Throw ( "Alias property name must be simple", kXMPErr_BadXPath );	// The alias must be a top level name.

	XMP_AliasMapPos mapPos = sRegisteredAliasMap->find ( expPath[kRootPropStep].step );
	if ( mapPos == sRegisteredAliasMap->end() ) return false;
	
	*actualNS = mapPos->second[kSchemaStep].step.c_str();
	*nsSize   = mapPos->second[kSchemaStep].step.size();
	
	if ( mapPos->second.size() == 2 ) {
		*actualProp = mapPos->second[kRootPropStep].step.c_str();	// This is a "direct" alias.
		*propSize   = mapPos->second[kRootPropStep].step.size();
	} else {
		*sOutputStr  = mapPos->second[kRootPropStep].step;		// This is an alias to an array item.
		*sOutputStr  += mapPos->second[kAliasIndexStep].step;
		*actualProp = sOutputStr->c_str();
		*propSize   = sOutputStr->size();
	}

	*arrayForm  = mapPos->second[kRootPropStep].options & kXMP_PropArrayFormMask;
	
	return true;
	
}	// ResolveAlias


// -------------------------------------------------------------------------------------------------
// DeleteAlias
// -----------

/* class-static */ void
XMPMeta::DeleteAlias ( XMP_StringPtr aliasNS,
					   XMP_StringPtr aliasProp )
{

	XMP_Assert ( (aliasNS != 0) && (aliasProp != 0) );	// Enforced by wrapper.
	XMP_Throw ( "Unimplemented method XMPMeta::DeleteAlias", kXMPErr_Unimplemented );   // *** #error "write me"
		void * p; p = &aliasNS; p = &aliasProp;	// Avoid unused param warnings.

}	// DeleteAlias


// -------------------------------------------------------------------------------------------------
// RegisterStandardAliases
// -----------------------

/* class-static */ void
XMPMeta::RegisterStandardAliases ( XMP_StringPtr schemaNS )
{
	XMP_Assert ( schemaNS != 0 );	// Enforced by wrapper.

	const bool doAll = (*schemaNS == 0);
	
	if ( doAll || XMP_LitMatch ( schemaNS, kXMP_NS_XMP ) ) {
		// Aliases from XMP to DC.
	    XMPMeta::RegisterAlias ( kXMP_NS_XMP, "Author",       kXMP_NS_DC, "creator", kXMP_PropArrayIsOrdered );
	    XMPMeta::RegisterAlias ( kXMP_NS_XMP, "Authors",      kXMP_NS_DC, "creator", 0 );
	    XMPMeta::RegisterAlias ( kXMP_NS_XMP, "Description",  kXMP_NS_DC, "description", 0 );
	    XMPMeta::RegisterAlias ( kXMP_NS_XMP, "Format",       kXMP_NS_DC, "format", 0 );
	    XMPMeta::RegisterAlias ( kXMP_NS_XMP, "Keywords",     kXMP_NS_DC, "subject", 0 );
	    XMPMeta::RegisterAlias ( kXMP_NS_XMP, "Locale",       kXMP_NS_DC, "language", 0 );
	    XMPMeta::RegisterAlias ( kXMP_NS_XMP, "Title",        kXMP_NS_DC, "title", 0 );
	    XMPMeta::RegisterAlias ( kXMP_NS_XMP_Rights, "Copyright", kXMP_NS_DC, "rights", 0 );
	}

	if ( doAll || XMP_LitMatch ( schemaNS, kXMP_NS_PDF ) ) {
		// Aliases from PDF to DC and XMP.
	    XMPMeta::RegisterAlias ( kXMP_NS_PDF, "Author",       kXMP_NS_DC, "creator", kXMP_PropArrayIsOrdered );
	    XMPMeta::RegisterAlias ( kXMP_NS_PDF, "BaseURL",      kXMP_NS_XMP, "BaseURL", 0 );
	    XMPMeta::RegisterAlias ( kXMP_NS_PDF, "CreationDate", kXMP_NS_XMP, "CreateDate", 0 );
	    XMPMeta::RegisterAlias ( kXMP_NS_PDF, "Creator",      kXMP_NS_XMP, "CreatorTool", 0 );
	    XMPMeta::RegisterAlias ( kXMP_NS_PDF, "ModDate",      kXMP_NS_XMP, "ModifyDate", 0 );
	    XMPMeta::RegisterAlias ( kXMP_NS_PDF, "Subject",      kXMP_NS_DC,  "description", kXMP_PropArrayIsAltText );
	    XMPMeta::RegisterAlias ( kXMP_NS_PDF, "Title",        kXMP_NS_DC,  "title", kXMP_PropArrayIsAltText );
    }
    
	if ( doAll || XMP_LitMatch ( schemaNS, kXMP_NS_Photoshop ) ) {
	    // Aliases from PHOTOSHOP to DC and XMP.
	    XMPMeta::RegisterAlias ( kXMP_NS_Photoshop, "Author",       kXMP_NS_DC, "creator", kXMP_PropArrayIsOrdered );
	    XMPMeta::RegisterAlias ( kXMP_NS_Photoshop, "Caption",      kXMP_NS_DC, "description", kXMP_PropArrayIsAltText );
	    XMPMeta::RegisterAlias ( kXMP_NS_Photoshop, "Copyright",    kXMP_NS_DC, "rights", kXMP_PropArrayIsAltText );
	    XMPMeta::RegisterAlias ( kXMP_NS_Photoshop, "Keywords",     kXMP_NS_DC, "subject", 0 );
	    XMPMeta::RegisterAlias ( kXMP_NS_Photoshop, "Marked",       kXMP_NS_XMP_Rights, "Marked", 0 );
	    XMPMeta::RegisterAlias ( kXMP_NS_Photoshop, "Title",        kXMP_NS_DC, "title", kXMP_PropArrayIsAltText );
	    XMPMeta::RegisterAlias ( kXMP_NS_Photoshop, "WebStatement", kXMP_NS_XMP_Rights, "WebStatement", 0 );
    }
	    
	if ( doAll || XMP_LitMatch ( schemaNS, kXMP_NS_TIFF ) || XMP_LitMatch ( schemaNS, kXMP_NS_EXIF ) ) {
	    // Aliases from TIFF and EXIF to DC and XMP.
		XMPMeta::RegisterAlias ( kXMP_NS_TIFF, "Artist",           kXMP_NS_DC,  "creator", kXMP_PropArrayIsOrdered);
	    XMPMeta::RegisterAlias ( kXMP_NS_TIFF, "Copyright",        kXMP_NS_DC,  "rights", 0 );
	    XMPMeta::RegisterAlias ( kXMP_NS_TIFF, "DateTime",         kXMP_NS_XMP, "ModifyDate", 0 );
	    XMPMeta::RegisterAlias ( kXMP_NS_TIFF, "ImageDescription", kXMP_NS_DC,  "description", 0 );
	    XMPMeta::RegisterAlias ( kXMP_NS_TIFF, "Software",         kXMP_NS_XMP, "CreatorTool", 0 );
    }
    
	if ( doAll || XMP_LitMatch ( schemaNS, kXMP_NS_PNG ) ) {	// ! From Acrobat ImageCapture:
		XMPMeta::RegisterAlias ( kXMP_NS_PNG, "Author",           kXMP_NS_DC,  "creator", kXMP_PropArrayIsOrdered);
		XMPMeta::RegisterAlias ( kXMP_NS_PNG, "Copyright",        kXMP_NS_DC,  "rights", kXMP_PropArrayIsAltText);
		XMPMeta::RegisterAlias ( kXMP_NS_PNG, "CreationTime",     kXMP_NS_XMP, "CreateDate", 0 );
		XMPMeta::RegisterAlias ( kXMP_NS_PNG, "Description",      kXMP_NS_DC,  "description", kXMP_PropArrayIsAltText);
		XMPMeta::RegisterAlias ( kXMP_NS_PNG, "ModificationTime", kXMP_NS_XMP, "ModDate", 0 );
		XMPMeta::RegisterAlias ( kXMP_NS_PNG, "Software",         kXMP_NS_XMP, "CreatorTool", 0 );
		XMPMeta::RegisterAlias ( kXMP_NS_PNG, "Title",            kXMP_NS_DC,  "title", kXMP_PropArrayIsAltText);
	}

}	// RegisterStandardAliases


// -------------------------------------------------------------------------------------------------
// RegisterPropertyTraits
// ----------------------

/* class-static */ void
XMPMeta::RegisterPropertyTraits	( XMP_StringPtr  schemaNS,
								  XMP_StringPtr	 propName,
								  XMP_OptionBits options )
{

	XMP_Assert ( (schemaNS != 0) && (propName != 0) );	// Enforced by wrapper.
	XMP_Throw ( "Unimplemented method XMPMeta::RegisterPropertyTraits", kXMPErr_Unimplemented );
		void * p; p = &schemaNS; p = &propName; p = &options;	// Avoid unused param warnings.

}	// RegisterPropertyTraits


// =================================================================================================
// Class Methods
// =============
//
//
// =================================================================================================


// -------------------------------------------------------------------------------------------------
// GetProperty
// -----------

bool
XMPMeta::GetProperty ( XMP_StringPtr	schemaNS,
					   XMP_StringPtr	propName,
					   XMP_StringPtr *	propValue,
					   XMP_StringLen *	valueSize,
					   XMP_OptionBits *	options ) const
{
	XMP_Assert ( (schemaNS != 0) && (propName != 0) );	// Enforced by wrapper.
	XMP_Assert ( (propValue != 0) && (valueSize != 0) && (options != 0) );	// Enforced by wrapper.

	XMP_ExpandedXPath expPath;
	ExpandXPath ( schemaNS, propName, &expPath );
	
	XMP_Node * propNode = FindConstNode ( &tree, expPath );
	if ( propNode == 0 ) return false;
	
	*propValue = propNode->value.c_str();
	*valueSize = propNode->value.size();
	*options   = propNode->options;
	
	return true;
	
}	// GetProperty


// -------------------------------------------------------------------------------------------------
// GetArrayItem
// ------------

bool
XMPMeta::GetArrayItem ( XMP_StringPtr	 schemaNS,
						XMP_StringPtr	 arrayName,
						XMP_Index		 itemIndex,
						XMP_StringPtr *	 itemValue,
						XMP_StringLen *	 valueSize,
						XMP_OptionBits * options ) const
{
	XMP_Assert ( (schemaNS != 0) && (arrayName != 0) );	// Enforced by wrapper.
	XMP_Assert ( (itemValue != 0) && (valueSize != 0) && (options != 0) );	// Enforced by wrapper.

	XMP_StringPtr itemPath;
	XMP_StringLen pathLen;

	XMPUtils::ComposeArrayItemPath ( schemaNS, arrayName, itemIndex, &itemPath, &pathLen );
	return GetProperty ( schemaNS, itemPath, itemValue, valueSize, options );

}	// GetArrayItem


// -------------------------------------------------------------------------------------------------
// GetStructField
// --------------

bool
XMPMeta::GetStructField	( XMP_StringPtr	   schemaNS,
						  XMP_StringPtr	   structName,
						  XMP_StringPtr	   fieldNS,
						  XMP_StringPtr	   fieldName,
						  XMP_StringPtr *  fieldValue,
						  XMP_StringLen *  valueSize,
						  XMP_OptionBits * options ) const
{
	XMP_Assert ( (schemaNS != 0) && (structName != 0) && (fieldNS != 0) && (fieldName != 0) );	// Enforced by wrapper.
	XMP_Assert ( (fieldValue != 0) && (valueSize != 0) && (options != 0) );	// Enforced by wrapper.

	XMP_StringPtr fieldPath;
	XMP_StringLen pathLen;

	XMPUtils::ComposeStructFieldPath ( schemaNS, structName, fieldNS, fieldName, &fieldPath, &pathLen );
	return GetProperty ( schemaNS, fieldPath, fieldValue, valueSize, options );

}	// GetStructField


// -------------------------------------------------------------------------------------------------
// GetQualifier
// ------------

bool
XMPMeta::GetQualifier ( XMP_StringPtr	 schemaNS,
						XMP_StringPtr	 propName,
						XMP_StringPtr	 qualNS,
						XMP_StringPtr	 qualName,
						XMP_StringPtr *	 qualValue,
						XMP_StringLen *	 valueSize,
						XMP_OptionBits * options ) const
{
	XMP_Assert ( (schemaNS != 0) && (propName != 0) && (qualNS != 0) && (qualName != 0) );	// Enforced by wrapper.
	XMP_Assert ( (qualValue != 0) && (valueSize != 0) && (options != 0) );	// Enforced by wrapper.

	XMP_StringPtr qualPath;
	XMP_StringLen pathLen;

	XMPUtils::ComposeQualifierPath ( schemaNS, propName, qualNS, qualName, &qualPath, &pathLen );
	return GetProperty ( schemaNS, qualPath, qualValue, valueSize, options );

}	// GetQualifier


// -------------------------------------------------------------------------------------------------
// SetProperty
// -----------

// *** Should handle array items specially, calling SetArrayItem.

void
XMPMeta::SetProperty ( XMP_StringPtr  schemaNS,
					   XMP_StringPtr  propName,
					   XMP_StringPtr  propValue,
					   XMP_OptionBits options )
{
	XMP_Assert ( (schemaNS != 0) && (propName != 0) );	// Enforced by wrapper.

	options = VerifySetOptions ( options, propValue );

	XMP_ExpandedXPath expPath;
	ExpandXPath ( schemaNS, propName, &expPath );

	XMP_Node * propNode = FindNode ( &tree, expPath, kXMP_CreateNodes, options );
	if ( propNode == 0 ) XMP_Throw ( "Specified property does not exist", kXMPErr_BadXPath );
	
	SetNode ( propNode, propValue, options );
	
}	// SetProperty


// -------------------------------------------------------------------------------------------------
// SetArrayItem
// ------------

void
XMPMeta::SetArrayItem ( XMP_StringPtr  schemaNS,
						XMP_StringPtr  arrayName,
						XMP_Index	   itemIndex,
						XMP_StringPtr  itemValue,
						XMP_OptionBits options )
{
	XMP_Assert ( (schemaNS != 0) && (arrayName != 0) );	// Enforced by wrapper.

	XMP_ExpandedXPath arrayPath;
	ExpandXPath ( schemaNS, arrayName, &arrayPath );
	XMP_Node * arrayNode = FindNode ( &tree, arrayPath, kXMP_ExistingOnly );	// Just lookup, don't try to create.
	if ( arrayNode == 0 ) XMP_Throw ( "Specified array does not exist", kXMPErr_BadXPath );
	
	DoSetArrayItem ( arrayNode, itemIndex, itemValue, options );
	
}	// SetArrayItem


// -------------------------------------------------------------------------------------------------
// AppendArrayItem
// ---------------

void
XMPMeta::AppendArrayItem ( XMP_StringPtr  schemaNS,
						   XMP_StringPtr  arrayName,
						   XMP_OptionBits arrayOptions,
						   XMP_StringPtr  itemValue,
						   XMP_OptionBits options )
{
	XMP_Assert ( (schemaNS != 0) && (arrayName != 0) );	// Enforced by wrapper.

	arrayOptions = VerifySetOptions ( arrayOptions, 0 );
	if ( (arrayOptions & ~kXMP_PropArrayFormMask) != 0 ) {
		XMP_Throw ( "Only array form flags allowed for arrayOptions", kXMPErr_BadOptions );
	}
	
	// Locate or create the array. If it already exists, make sure the array form from the options
	// parameter is compatible with the current state.
	
	XMP_ExpandedXPath arrayPath;
	ExpandXPath ( schemaNS, arrayName, &arrayPath );
	XMP_Node * arrayNode = FindNode ( &tree, arrayPath, kXMP_ExistingOnly );	// Just lookup, don't try to create.
	
	if ( arrayNode != 0 ) {
		// The array exists, make sure the form is compatible. Zero arrayForm means take what exists.
		if ( ! (arrayNode->options & kXMP_PropValueIsArray) ) {
			XMP_Throw ( "The named property is not an array", kXMPErr_BadXPath );
		}
		if ( (arrayOptions != 0) && (arrayOptions != (arrayNode->options & kXMP_PropArrayFormMask)) ) {
			XMP_Throw ( "Mismatch of existing and specified array form", kXMPErr_BadOptions );
		}
	} else {
		// The array does not exist, try to create it.
		if ( arrayOptions == 0 ) XMP_Throw ( "Explicit arrayOptions required to create new arry", kXMPErr_BadOptions );
		arrayNode = FindNode ( &tree, arrayPath, kXMP_CreateNodes, arrayOptions );
		if ( arrayNode == 0 ) XMP_Throw ( "Failure creating array node", kXMPErr_BadXPath );
	}
	
	DoSetArrayItem ( arrayNode, kXMP_ArrayLastItem, itemValue, (options | kXMP_InsertAfterItem) );
	
}	// AppendArrayItem


// -------------------------------------------------------------------------------------------------
// SetStructField
// --------------

void
XMPMeta::SetStructField	( XMP_StringPtr	 schemaNS,
						  XMP_StringPtr	 structName,
						  XMP_StringPtr	 fieldNS,
						  XMP_StringPtr	 fieldName,
						  XMP_StringPtr	 fieldValue,
						  XMP_OptionBits options )
{
	XMP_Assert ( (schemaNS != 0) && (structName != 0) && (fieldNS != 0) && (fieldName != 0) );	// Enforced by wrapper.

	XMP_StringPtr	fieldPath;
	XMP_StringLen	pathLen;

	XMPUtils::ComposeStructFieldPath ( schemaNS, structName, fieldNS, fieldName, &fieldPath, &pathLen );
	SetProperty ( schemaNS, fieldPath, fieldValue, options );

}	// SetStructField


// -------------------------------------------------------------------------------------------------
// SetQualifier
// ------------

void
XMPMeta::SetQualifier ( XMP_StringPtr  schemaNS,
						XMP_StringPtr  propName,
						XMP_StringPtr  qualNS,
						XMP_StringPtr  qualName,
						XMP_StringPtr  qualValue,
						XMP_OptionBits options )
{
	XMP_Assert ( (schemaNS != 0) && (propName != 0) && (qualNS != 0) && (qualName != 0) );	// Enforced by wrapper.

	XMP_StringPtr	qualPath;
	XMP_StringLen	pathLen;

	XMP_ExpandedXPath expPath;
	ExpandXPath ( schemaNS, propName, &expPath );
	XMP_Node * propNode = FindNode ( &tree, expPath, kXMP_ExistingOnly );
	if ( propNode == 0 ) XMP_Throw ( "Specified property does not exist", kXMPErr_BadXPath );

	XMPUtils::ComposeQualifierPath ( schemaNS, propName, qualNS, qualName, &qualPath, &pathLen );
	SetProperty ( schemaNS, qualPath, qualValue, options );

}	// SetQualifier


// -------------------------------------------------------------------------------------------------
// DeleteProperty
// --------------

void
XMPMeta::DeleteProperty	( XMP_StringPtr	schemaNS,
						  XMP_StringPtr	propName )
{
	XMP_Assert ( (schemaNS != 0) && (propName != 0) );	// Enforced by wrapper.

	XMP_ExpandedXPath	expPath;
	ExpandXPath ( schemaNS, propName, &expPath );
	
	XMP_NodePtrPos ptrPos;
	XMP_Node * propNode = FindNode ( &tree, expPath, kXMP_ExistingOnly, kXMP_NoOptions, &ptrPos );
	if ( propNode == 0 ) return;
	XMP_Node * parentNode = propNode->parent;
	
	// Erase the pointer from the parent's vector, then delete the node and all below it.
	
	if ( ! (propNode->options & kXMP_PropIsQualifier) ) {

		parentNode->children.erase ( ptrPos );
		DeleteEmptySchema ( parentNode );

	} else {

		if ( propNode->name == "xml:lang" ) {
			XMP_Assert ( parentNode->options & kXMP_PropHasLang );	// *** &= ~flag would be safer
			parentNode->options ^= kXMP_PropHasLang;
		} else if ( propNode->name == "rdf:type" ) {
			XMP_Assert ( parentNode->options & kXMP_PropHasType );
			parentNode->options ^= kXMP_PropHasType;
		}

		parentNode->qualifiers.erase ( ptrPos );
		XMP_Assert ( parentNode->options & kXMP_PropHasQualifiers );
		if ( parentNode->qualifiers.empty() ) parentNode->options ^= kXMP_PropHasQualifiers;

	}
	
	delete propNode;	// ! The destructor takes care of the whole subtree.
	
}	// DeleteProperty


// -------------------------------------------------------------------------------------------------
// DeleteArrayItem
// ---------------

void
XMPMeta::DeleteArrayItem ( XMP_StringPtr schemaNS,
						   XMP_StringPtr arrayName,
						   XMP_Index	 itemIndex )
{
	XMP_Assert ( (schemaNS != 0) && (arrayName != 0) );	// Enforced by wrapper.

	XMP_StringPtr	itemPath;
	XMP_StringLen	pathLen;

	XMPUtils::ComposeArrayItemPath ( schemaNS, arrayName, itemIndex, &itemPath, &pathLen );
	DeleteProperty ( schemaNS, itemPath );

}	// DeleteArrayItem


// -------------------------------------------------------------------------------------------------
// DeleteStructField
// -----------------

void
XMPMeta::DeleteStructField ( XMP_StringPtr schemaNS,
							 XMP_StringPtr structName,
							 XMP_StringPtr fieldNS,
							 XMP_StringPtr fieldName )
{
	XMP_Assert ( (schemaNS != 0) && (structName != 0) && (fieldNS != 0) && (fieldName != 0) );	// Enforced by wrapper.

	XMP_StringPtr	fieldPath;
	XMP_StringLen	pathLen;

	XMPUtils::ComposeStructFieldPath ( schemaNS, structName, fieldNS, fieldName, &fieldPath, &pathLen );
	DeleteProperty ( schemaNS, fieldPath );

}	// DeleteStructField


// -------------------------------------------------------------------------------------------------
// DeleteQualifier
// ---------------

void
XMPMeta::DeleteQualifier ( XMP_StringPtr schemaNS,
						   XMP_StringPtr propName,
						   XMP_StringPtr qualNS,
						   XMP_StringPtr qualName )
{
	XMP_Assert ( (schemaNS != 0) && (propName != 0) && (qualNS != 0) && (qualName != 0) );	// Enforced by wrapper.

	XMP_StringPtr	qualPath;
	XMP_StringLen	pathLen;

	XMPUtils::ComposeQualifierPath ( schemaNS, propName, qualNS, qualName, &qualPath, &pathLen );
	DeleteProperty ( schemaNS, qualPath );

}	// DeleteQualifier


// -------------------------------------------------------------------------------------------------
// DoesPropertyExist
// -----------------

bool
XMPMeta::DoesPropertyExist ( XMP_StringPtr schemaNS,
							 XMP_StringPtr propName ) const
{
	XMP_Assert ( (schemaNS != 0) && (propName != 0) );	// Enforced by wrapper.

	XMP_ExpandedXPath	expPath;
	ExpandXPath ( schemaNS, propName, &expPath );

	XMP_Node * propNode = FindConstNode ( &tree, expPath );
	return (propNode != 0);
	
}	// DoesPropertyExist


// -------------------------------------------------------------------------------------------------
// DoesArrayItemExist
// ------------------

bool
XMPMeta::DoesArrayItemExist	( XMP_StringPtr	schemaNS,
							  XMP_StringPtr	arrayName,
							  XMP_Index		itemIndex ) const
{
	XMP_Assert ( (schemaNS != 0) && (arrayName != 0) );	// Enforced by wrapper.

	XMP_StringPtr	itemPath;
	XMP_StringLen	pathLen;

	XMPUtils::ComposeArrayItemPath ( schemaNS, arrayName, itemIndex, &itemPath, &pathLen );
	return DoesPropertyExist ( schemaNS, itemPath );

}	// DoesArrayItemExist


// -------------------------------------------------------------------------------------------------
// DoesStructFieldExist
// --------------------

bool
XMPMeta::DoesStructFieldExist ( XMP_StringPtr schemaNS,
								XMP_StringPtr structName,
								XMP_StringPtr fieldNS,
								XMP_StringPtr fieldName ) const
{
	XMP_Assert ( (schemaNS != 0) && (structName != 0) && (fieldNS != 0) && (fieldName != 0) );	// Enforced by wrapper.

	XMP_StringPtr	fieldPath;
	XMP_StringLen	pathLen;

	XMPUtils::ComposeStructFieldPath ( schemaNS, structName, fieldNS, fieldName, &fieldPath, &pathLen );
	return DoesPropertyExist ( schemaNS, fieldPath );

}	// DoesStructFieldExist


// -------------------------------------------------------------------------------------------------
// DoesQualifierExist
// ------------------

bool
XMPMeta::DoesQualifierExist	( XMP_StringPtr	schemaNS,
							  XMP_StringPtr	propName,
							  XMP_StringPtr	qualNS,
							  XMP_StringPtr	qualName ) const
{
	XMP_Assert ( (schemaNS != 0) && (propName != 0) && (qualNS != 0) && (qualName != 0) );	// Enforced by wrapper.

	XMP_StringPtr	qualPath;
	XMP_StringLen	pathLen;

	XMPUtils::ComposeQualifierPath ( schemaNS, propName, qualNS, qualName, &qualPath, &pathLen );
	return DoesPropertyExist ( schemaNS, qualPath );

}	// DoesQualifierExist


// -------------------------------------------------------------------------------------------------
// GetLocalizedText
// ----------------

bool
XMPMeta::GetLocalizedText ( XMP_StringPtr	 schemaNS,
							XMP_StringPtr	 arrayName,
							XMP_StringPtr	 _genericLang,
							XMP_StringPtr	 _specificLang,
							XMP_StringPtr *	 actualLang,
							XMP_StringLen *	 langSize,
							XMP_StringPtr *	 itemValue,
							XMP_StringLen *	 valueSize,
							XMP_OptionBits * options ) const
{
	XMP_Assert ( (schemaNS != 0) && (arrayName != 0) && (_genericLang != 0) && (_specificLang != 0) );	// Enforced by wrapper.
	XMP_Assert ( (actualLang != 0) && (langSize != 0) );	// Enforced by wrapper.
	XMP_Assert ( (itemValue != 0) && (valueSize != 0) && (options != 0) );	// Enforced by wrapper.

	XMP_VarString zGenericLang  ( _genericLang );
	XMP_VarString zSpecificLang ( _specificLang );
	NormalizeLangValue ( &zGenericLang );
	NormalizeLangValue ( &zSpecificLang );
	
	XMP_StringPtr genericLang  = zGenericLang.c_str();
	XMP_StringPtr specificLang = zSpecificLang.c_str();
	
	XMP_ExpandedXPath arrayPath;
	ExpandXPath ( schemaNS, arrayName, &arrayPath );
	
	const XMP_Node * arrayNode = FindConstNode ( &tree, arrayPath );	// *** This expand/find idiom is used in 3 Getters.
	if ( arrayNode == 0 ) return false;			// *** Should extract it into a local utility.
	
	XMP_CLTMatch match;
	const XMP_Node * itemNode;
	
	match = ChooseLocalizedText ( arrayNode, genericLang, specificLang, &itemNode );
	if ( match == kXMP_CLT_NoValues ) return false;
	
	*actualLang = itemNode->qualifiers[0]->value.c_str();
	*langSize   = itemNode->qualifiers[0]->value.size();
	*itemValue  = itemNode->value.c_str();
	*valueSize  = itemNode->value.size();
	*options    = itemNode->options;

	return true;
	
}	// GetLocalizedText


// -------------------------------------------------------------------------------------------------
// SetLocalizedText
// ----------------

void
XMPMeta::SetLocalizedText ( XMP_StringPtr  schemaNS,
							XMP_StringPtr  arrayName,
							XMP_StringPtr  _genericLang,
							XMP_StringPtr  _specificLang,
							XMP_StringPtr  itemValue,
							XMP_OptionBits options )
{
	options = options;	// Avoid unused parameter warning.

	XMP_Assert ( (schemaNS != 0) && (arrayName != 0) && (_genericLang != 0) && (_specificLang != 0) );	// Enforced by wrapper.

	XMP_VarString zGenericLang  ( _genericLang );
	XMP_VarString zSpecificLang ( _specificLang );
	NormalizeLangValue ( &zGenericLang );
	NormalizeLangValue ( &zSpecificLang );
	
	XMP_StringPtr genericLang  = zGenericLang.c_str();
	XMP_StringPtr specificLang = zSpecificLang.c_str();
	
	XMP_ExpandedXPath arrayPath;
	ExpandXPath ( schemaNS, arrayName, &arrayPath );
	
	// Find the array node and set the options if it was just created.
	XMP_Node * arrayNode = FindNode ( &tree, arrayPath, kXMP_CreateNodes,
									  (kXMP_PropValueIsArray | kXMP_PropArrayIsOrdered | kXMP_PropArrayIsAlternate) );
	if ( arrayNode == 0 ) XMP_Throw ( "Failed to find or create array node", kXMPErr_BadXPath );
	if ( ! XMP_ArrayIsAltText(arrayNode->options) ) {
		if ( arrayNode->children.empty() && XMP_ArrayIsAlternate(arrayNode->options) ) {
			arrayNode->options |= kXMP_PropArrayIsAltText;
		} else {
			XMP_Throw ( "Localized text array is not alt-text", kXMPErr_BadXPath );
		}
	}
	
	// Make sure the x-default item, if any, is first.
	
	size_t itemNum, itemLim;
	XMP_Node * xdItem = 0;
	bool haveXDefault = false;
	
	for ( itemNum = 0, itemLim = arrayNode->children.size(); itemNum < itemLim; ++itemNum ) {
		XMP_Node * currItem = arrayNode->children[itemNum];
		XMP_Assert ( XMP_PropHasLang(currItem->options) );
		if ( currItem->qualifiers.empty() || (currItem->qualifiers[0]->name != "xml:lang") ) {
			XMP_Throw ( "Language qualifier must be first", kXMPErr_BadXPath );
		}
		if ( currItem->qualifiers[0]->value == "x-default" ) {
			xdItem = currItem;
			haveXDefault = true;
			break;
		}
	}
	
	if ( haveXDefault && (itemNum != 0) ) {
		XMP_Assert ( arrayNode->children[itemNum]->qualifiers[0]->value == "x-default" );
		XMP_Node * temp = arrayNode->children[0];
		arrayNode->children[0] = arrayNode->children[itemNum];
		arrayNode->children[itemNum] = temp;
	}
	
	// Find the appropriate item. ChooseLocalizedText will make sure the array is a language alternative.
		
	const XMP_Node * cItemNode;	// ! ChooseLocalizedText returns a pointer to a const node.
	XMP_CLTMatch match = ChooseLocalizedText ( arrayNode, genericLang, specificLang, &cItemNode );
	XMP_Node * itemNode = const_cast<XMP_Node*> ( cItemNode );

	const bool specificXDefault = XMP_LitMatch ( specificLang, "x-default" );
	
	switch ( match ) {

		case kXMP_CLT_NoValues :

			// Create the array items for the specificLang and x-default, with x-default first.
			AppendLangItem ( arrayNode, "x-default", itemValue );
			haveXDefault = true;
			if ( ! specificXDefault ) AppendLangItem ( arrayNode, specificLang, itemValue );
			break;
			
		case kXMP_CLT_SpecificMatch :
		
			if ( ! specificXDefault ) {
				// Update the specific item, update x-default if it matches the old value.
				if ( haveXDefault && (xdItem != itemNode) && (xdItem->value == itemNode->value) ) {
					SetNodeValue ( xdItem, itemValue );
				}
				SetNodeValue ( itemNode, itemValue );	// ! Do this after the x-default check!
			} else {
				// Update all items whose values match the old x-default value.
				XMP_Assert ( haveXDefault && (xdItem == itemNode) );
				for ( itemNum = 0, itemLim = arrayNode->children.size(); itemNum < itemLim; ++itemNum ) {
					XMP_Node * currItem = arrayNode->children[itemNum];
					if ( (currItem == xdItem) || (currItem->value != xdItem->value) ) continue;
					SetNodeValue ( currItem, itemValue );
				}
				SetNodeValue ( xdItem, itemValue );	// And finally do the x-default item.
			}
			break;

		case kXMP_CLT_SingleGeneric :
		
			// Update the generic item, update x-default if it matches the old value.
			if ( haveXDefault && (xdItem != itemNode) && (xdItem->value == itemNode->value) ) {
				SetNodeValue ( xdItem, itemValue );
			}
			SetNodeValue ( itemNode, itemValue );	// ! Do this after the x-default check!
			break;

		case kXMP_CLT_MultipleGeneric :
		
			// Create the specific language, ignore x-default.
			AppendLangItem ( arrayNode, specificLang, itemValue );
			if ( specificXDefault ) haveXDefault = true;
			break;
			
		case kXMP_CLT_XDefault :

			// Create the specific language, update x-default if it was the only item.
			if ( arrayNode->children.size() == 1 ) SetNodeValue ( xdItem, itemValue );
			AppendLangItem ( arrayNode, specificLang, itemValue );
			break;

		case kXMP_CLT_FirstItem	:

			// Create the specific language, don't add an x-default item.
			AppendLangItem ( arrayNode, specificLang, itemValue );
			if ( specificXDefault ) haveXDefault = true;
			break;
			
		default :
			XMP_Throw ( "Unexpected result from ChooseLocalizedText", kXMPErr_InternalFailure );

	}

	// Add an x-default at the front if needed.
	if ( (! haveXDefault) && (arrayNode->children.size() == 1) ) {
		AppendLangItem ( arrayNode, "x-default", itemValue );
	}

}	// SetLocalizedText


// -------------------------------------------------------------------------------------------------
// GetProperty_Bool
// ----------------

bool
XMPMeta::GetProperty_Bool ( XMP_StringPtr	 schemaNS,
							XMP_StringPtr	 propName,
							bool *			 propValue,
							XMP_OptionBits * options ) const
{
	XMP_Assert ( (schemaNS != 0) && (propName != 0) );	// Enforced by wrapper.
	XMP_Assert ( (propValue != 0) && (options != 0) );	// Enforced by wrapper.

	XMP_StringPtr	valueStr;
	XMP_StringLen	valueLen;
	
	bool found = GetProperty ( schemaNS, propName, &valueStr, &valueLen, options );
	if ( found ) {
		if ( ! XMP_PropIsSimple ( *options ) ) XMP_Throw ( "Property must be simple", kXMPErr_BadXPath );
		*propValue = XMPUtils::ConvertToBool ( valueStr );
	}
	return found;
	
}	// GetProperty_Bool


// -------------------------------------------------------------------------------------------------
// GetProperty_Int
// ---------------

bool
XMPMeta::GetProperty_Int ( XMP_StringPtr	schemaNS,
						   XMP_StringPtr	propName,
						   XMP_Int32 *		propValue,
						   XMP_OptionBits *	options ) const
{
	XMP_Assert ( (schemaNS != 0) && (propName != 0) );	// Enforced by wrapper.
	XMP_Assert ( (propValue != 0) && (options != 0) );	// Enforced by wrapper.

	XMP_StringPtr	valueStr;
	XMP_StringLen	valueLen;
	
	bool found = GetProperty ( schemaNS, propName, &valueStr, &valueLen, options );
	if ( found ) {
		if ( ! XMP_PropIsSimple ( *options ) ) XMP_Throw ( "Property must be simple", kXMPErr_BadXPath );
		*propValue = XMPUtils::ConvertToInt ( valueStr );
	}
	return found;
	
}	// GetProperty_Int


// -------------------------------------------------------------------------------------------------
// GetProperty_Int64
// -----------------

bool
XMPMeta::GetProperty_Int64 ( XMP_StringPtr	  schemaNS,
						     XMP_StringPtr	  propName,
						     XMP_Int64 *	  propValue,
						     XMP_OptionBits * options ) const
{
	XMP_Assert ( (schemaNS != 0) && (propName != 0) );	// Enforced by wrapper.
	XMP_Assert ( (propValue != 0) && (options != 0) );	// Enforced by wrapper.

	XMP_StringPtr	valueStr;
	XMP_StringLen	valueLen;
	
	bool found = GetProperty ( schemaNS, propName, &valueStr, &valueLen, options );
	if ( found ) {
		if ( ! XMP_PropIsSimple ( *options ) ) XMP_Throw ( "Property must be simple", kXMPErr_BadXPath );
		*propValue = XMPUtils::ConvertToInt64 ( valueStr );
	}
	return found;
	
}	// GetProperty_Int64


// -------------------------------------------------------------------------------------------------
// GetProperty_Float
// -----------------

bool
XMPMeta::GetProperty_Float ( XMP_StringPtr	  schemaNS,
							 XMP_StringPtr	  propName,
							 double *		  propValue,
							 XMP_OptionBits * options ) const
{
	XMP_Assert ( (schemaNS != 0) && (propName != 0) );	// Enforced by wrapper.
	XMP_Assert ( (propValue != 0) && (options != 0) );	// Enforced by wrapper.

	XMP_StringPtr	valueStr;
	XMP_StringLen	valueLen;
	
	bool found = GetProperty ( schemaNS, propName, &valueStr, &valueLen, options );
	if ( found ) {
		if ( ! XMP_PropIsSimple ( *options ) ) XMP_Throw ( "Property must be simple", kXMPErr_BadXPath );
		*propValue = XMPUtils::ConvertToFloat ( valueStr );
	}
	return found;
	
}	// GetProperty_Float


// -------------------------------------------------------------------------------------------------
// GetProperty_Date
// ----------------

bool
XMPMeta::GetProperty_Date ( XMP_StringPtr	 schemaNS,
							XMP_StringPtr	 propName,
							XMP_DateTime *	 propValue,
							XMP_OptionBits * options ) const
{
	XMP_Assert ( (schemaNS != 0) && (propName != 0) );	// Enforced by wrapper.
	XMP_Assert ( (propValue != 0) && (options != 0) );	// Enforced by wrapper.

	XMP_StringPtr	valueStr;
	XMP_StringLen	valueLen;
	
	bool found = GetProperty ( schemaNS, propName, &valueStr, &valueLen, options );
	if ( found )  {
		if ( ! XMP_PropIsSimple ( *options ) ) XMP_Throw ( "Property must be simple", kXMPErr_BadXPath );
		XMPUtils::ConvertToDate ( valueStr, propValue );
	}
	return found;
	
}	// GetProperty_Date


// -------------------------------------------------------------------------------------------------
// SetProperty_Bool
// ----------------

void
XMPMeta::SetProperty_Bool ( XMP_StringPtr  schemaNS,
							XMP_StringPtr  propName,
							bool		   propValue,
							XMP_OptionBits options )
{
	XMP_Assert ( (schemaNS != 0) && (propName != 0) );	// Enforced by wrapper.

	XMP_StringPtr	valueStr;
	XMP_StringLen	valueLen;
	
	XMPUtils::ConvertFromBool ( propValue, &valueStr, &valueLen );
	SetProperty ( schemaNS, propName, valueStr, options );
	
}	// SetProperty_Bool


// -------------------------------------------------------------------------------------------------
// SetProperty_Int
// ---------------

void
XMPMeta::SetProperty_Int ( XMP_StringPtr  schemaNS,
						   XMP_StringPtr  propName,
						   XMP_Int32	  propValue,
						   XMP_OptionBits options )
{
	XMP_Assert ( (schemaNS != 0) && (propName != 0) );	// Enforced by wrapper.

	XMP_StringPtr	valueStr;
	XMP_StringLen	valueLen;
	
	XMPUtils::ConvertFromInt ( propValue, "", &valueStr, &valueLen );
	SetProperty ( schemaNS, propName, valueStr, options );
	
}	// SetProperty_Int


// -------------------------------------------------------------------------------------------------
// SetProperty_Int64
// -----------------

void
XMPMeta::SetProperty_Int64 ( XMP_StringPtr  schemaNS,
						     XMP_StringPtr  propName,
						     XMP_Int64	    propValue,
						     XMP_OptionBits options )
{
	XMP_Assert ( (schemaNS != 0) && (propName != 0) );	// Enforced by wrapper.

	XMP_StringPtr	valueStr;
	XMP_StringLen	valueLen;
	
	XMPUtils::ConvertFromInt64 ( propValue, "", &valueStr, &valueLen );
	SetProperty ( schemaNS, propName, valueStr, options );
	
}	// SetProperty_Int64


// -------------------------------------------------------------------------------------------------
// SetProperty_Float
// -----------------

void
XMPMeta::SetProperty_Float ( XMP_StringPtr	schemaNS,
							 XMP_StringPtr	propName,
							 double			propValue,
							 XMP_OptionBits	options )
{
	XMP_Assert ( (schemaNS != 0) && (propName != 0) );	// Enforced by wrapper.

	XMP_StringPtr	valueStr;
	XMP_StringLen	valueLen;
	
	XMPUtils::ConvertFromFloat ( propValue, "", &valueStr, &valueLen );
	SetProperty ( schemaNS, propName, valueStr, options );
	
}	// SetProperty_Float


// -------------------------------------------------------------------------------------------------
// SetProperty_Date
// ----------------

void
XMPMeta::SetProperty_Date ( XMP_StringPtr		   schemaNS,
							XMP_StringPtr		   propName,
							const	XMP_DateTime & propValue,
							XMP_OptionBits		   options )
{
	XMP_Assert ( (schemaNS != 0) && (propName != 0) );	// Enforced by wrapper.

	XMP_StringPtr	valueStr;
	XMP_StringLen	valueLen;
	
	XMPUtils::ConvertFromDate ( propValue, &valueStr, &valueLen );
	SetProperty ( schemaNS, propName, valueStr, options );
	
}	// SetProperty_Date


// -------------------------------------------------------------------------------------------------
// DumpObject
// ----------

XMP_Status
XMPMeta::DumpObject ( XMP_TextOutputProc outProc,
                      void *             refCon ) const
{
	XMP_Assert ( outProc != 0 );	// ! Enforced by wrapper.
	XMP_Status status;
	
	OutProcLiteral ( "Dumping XMPMeta object \"" );
	OutProcString ( tree.name );
	OutProcNChars ( "\"  ", 3 );
	status = DumpNodeOptions ( tree.options, outProc, refCon );
	if ( status != 0 ) goto EXIT;
	#if 0	// *** XMP_DebugBuild
		if ( (tree._namePtr != tree.name.c_str()) ||
			 (tree._valuePtr != tree.value.c_str()) ) OutProcLiteral ( "  ** bad debug string **" );
	#endif
	OutProcNewline();
	
	if ( ! tree.value.empty() ) {
		OutProcLiteral ( "** bad root value **  \"" );
		OutProcString ( tree.value );
		OutProcNChars ( "\"", 1 );
		OutProcNewline();
	}

	if ( ! tree.qualifiers.empty() ) {
		OutProcLiteral ( "** bad root qualifiers **" );
		OutProcNewline();
		for ( size_t qualNum = 0, qualLim = tree.qualifiers.size(); qualNum < qualLim; ++qualNum ) {
			status = DumpPropertyTree ( tree.qualifiers[qualNum], 3, 0, outProc, refCon );
		}
	}

	if ( ! tree.children.empty() ) {

		for ( size_t childNum = 0, childLim = tree.children.size(); childNum < childLim; ++childNum ) {

			const XMP_Node * currSchema = tree.children[childNum];

			OutProcNewline();
			OutProcIndent ( 1 );
			OutProcString ( currSchema->name );
			OutProcNChars ( "  ", 2 );
			OutProcString ( currSchema->value );
			OutProcNChars ( "  ", 2 );
			status = DumpNodeOptions ( currSchema->options, outProc, refCon );
			if ( status != 0 ) goto EXIT;
			#if 0	// *** XMP_DebugBuild
				if ( (currSchema->_namePtr != currSchema->name.c_str()) ||
					 (currSchema->_valuePtr != currSchema->value.c_str()) ) OutProcLiteral ( "  ** bad debug string **" );
			#endif
			OutProcNewline();

			if ( ! (currSchema->options & kXMP_SchemaNode) ) {
				OutProcLiteral ( "** bad schema options **" );
				OutProcNewline();
			}

			if ( ! currSchema->qualifiers.empty() ) {
				OutProcLiteral ( "** bad schema qualifiers **" );
				OutProcNewline();
				for ( size_t qualNum = 0, qualLim = currSchema->qualifiers.size(); qualNum < qualLim; ++qualNum ) {
					DumpPropertyTree ( currSchema->qualifiers[qualNum], 3, 0, outProc, refCon );
				}
			}

			for ( size_t childNum = 0, childLim = currSchema->children.size(); childNum < childLim; ++childNum ) {
				DumpPropertyTree ( currSchema->children[childNum], 2, 0, outProc, refCon );
			}

		}

	}
	
EXIT:
	return status;

}	// DumpObject


// -------------------------------------------------------------------------------------------------
// CountArrayItems
// ---------------

XMP_Index
XMPMeta::CountArrayItems ( XMP_StringPtr schemaNS,
						   XMP_StringPtr arrayName ) const
{
	XMP_Assert ( (schemaNS != 0) && (arrayName != 0) );	// Enforced by wrapper.

	XMP_ExpandedXPath	expPath;
	ExpandXPath ( schemaNS, arrayName, &expPath );
	
	const XMP_Node * arrayNode = FindConstNode ( &tree, expPath );

	if ( arrayNode == 0 ) return 0;
	if ( ! (arrayNode->options & kXMP_PropValueIsArray) ) XMP_Throw ( "The named property is not an array", kXMPErr_BadXPath );
	return arrayNode->children.size();
	
}	// CountArrayItems


// -------------------------------------------------------------------------------------------------
// MarkStaleProperties
// -------------------

void
XMPMeta::MarkStaleProperties ( XMP_OptionBits options )
{

	XMP_Throw ( "Unimplemented method XMPMeta::MarkStaleProperties", kXMPErr_Unimplemented );
		void * p; p = &options;	// Avoid unused param warnings.

}	// MarkStaleProperties


// -------------------------------------------------------------------------------------------------
// GetObjectName
// -------------

void
XMPMeta::GetObjectName ( XMP_StringPtr * namePtr,
						 XMP_StringLen * nameLen ) const
{
	
	*namePtr = tree.name.c_str();
	*nameLen = tree.name.size();
	
}	// GetObjectName


// -------------------------------------------------------------------------------------------------
// SetObjectName
// -------------

void
XMPMeta::SetObjectName ( XMP_StringPtr name )
{
	VerifyUTF8 ( name );	// Throws if the string is not legit UTF-8.
	tree.name = name;

}	// SetObjectName


// -------------------------------------------------------------------------------------------------
// GetObjectOptions
// ----------------

XMP_OptionBits
XMPMeta::GetObjectOptions() const
{
	XMP_OptionBits	options	= 0;
	
	return options;
	
}	// GetObjectOptions


// -------------------------------------------------------------------------------------------------
// SetObjectOptions
// ----------------

void
XMPMeta::SetObjectOptions ( XMP_OptionBits options )
{

	XMP_Throw ( "Unimplemented method XMPMeta::SetObjectOptions", kXMPErr_Unimplemented );
		void * p; p = &options;	// Avoid unused param warnings.

}	// SetObjectOptions


// -------------------------------------------------------------------------------------------------
// Clone
// -----

XMPMeta *
XMPMeta::Clone ( XMP_OptionBits options ) const
{
	if ( options != 0 ) XMP_Throw ( "No options are defined yet", kXMPErr_BadOptions );
	
	XMPMeta * clone = new XMPMeta;
	
	clone->tree.options = this->tree.options;
	clone->tree.name    = this->tree.name;
	clone->tree.value   = this->tree.value;
	XMP_Assert ( this->tree.parent == 0 );

	#if 0	// *** XMP_DebugBuild
		clone->tree._namePtr = clone->tree.name.c_str();
		clone->tree._valuePtr = clone->tree.value.c_str();
	#endif
	
	CloneOffspring ( &this->tree, &clone->tree );
	
	XMP_Assert ( clone->clientRefs == 0 );	// Gets incremneted later.
	return clone;
	
}	// Clone


// -------------------------------------------------------------------------------------------------
// DetermineInputEncoding
// ----------------------
//
// We don't know the character encoding yet. This is complicated by the nasty case of single byte
// input. Try to accumulate at least 4 bytes of input. Throw an exception if the total input is
// less than 4 bytes. We need an outermost XML element, the shortest possible UTF-8 sequence is
// "<x/>", which still isn't even legit XMP. The result is the number of bytes absorbed.

static XMP_StringLen
DetermineInputEncoding ( XMLParserAdapter * xmlParser,
						 XMP_StringPtr	    buffer,
						 XMP_StringLen      xmpSize,
						 bool               lastClientCall )
{
	size_t copySize = 0;
	
	if ( (xmlParser->partialUniSize + xmpSize) < 4 ) {

		// Accumulate the input in the partialUniChar buffer.
		
		copySize = xmpSize;
		if ( lastClientCall ) XMP_Throw ( "Insufficient parsing input", kXMPErr_BadXML );
		for ( size_t i = 0; i < copySize; ++i ) {
			xmlParser->partialUniChar[xmlParser->partialUniSize+i] = buffer[i];
		}
		xmlParser->partialUniSize += copySize;

	} else {
	
		// Fill out the partialUniChar buffer and determine the encoding.
		
		copySize = 4 - xmlParser->partialUniSize;
		for ( size_t i = 0; i < copySize; ++i ) {
			xmlParser->partialUniChar[xmlParser->partialUniSize+i] = buffer[i];
		}
		xmlParser->partialUniSize += copySize;
		
		// The first character must be U+FEFF or ASCII, typically '<' for an outermost element,
		// initial processing instruction, or XML declaration. The second character can't be U+0000.
		// The possible input sequences are:
		//   Cases with U+FEFF
		//      EF BB BF -- - UTF-8
		//      FE FF -- -- - Big endian UTF-16
		//      00 00 FE FF - Big endian UTF 32
		//      FF FE 00 00 - Little endian UTF-32
		//      FF FE -- -- - Little endian UTF-16
		//   Cases with ASCII
		//      nn mm -- -- - UTF-8 -
		//      00 00 00 nn - Big endian UTF-32
		//      00 nn -- -- - Big endian UTF-16
		//      nn 00 00 00 - Little endian UTF-32
		//      nn 00 -- -- - Little endian UTF-16
		
		XMP_Uns8 * uniChar = (XMP_Uns8*)xmlParser->partialUniChar;	// ! Make sure comparisons are unsigned.
		
		if ( uniChar[0] == 0 ) {
		
			// These cases are:
			//   00 nn -- -- - Big endian UTF-16
			//   00 00 00 nn - Big endian UTF-32
			//   00 00 FE FF - Big endian UTF 32
			
			if ( uniChar[1] != 0 ) {
				xmlParser->charEncoding = kXMP_EncodeUTF16Big;			// 00 nn
			} else {
				if ( (uniChar[2] == 0) && (uniChar[3] != 0) ) {
					xmlParser->charEncoding = kXMP_EncodeUTF32Big;		// 00 00 00 nn
				} else if ( (uniChar[2] == 0xFE) && (uniChar[3] == 0xFF) ) {
					xmlParser->charEncoding = kXMP_EncodeUTF32Big;		// 00 00 FE FF
				}
			}
			
		} else {
		
			// These cases are:
			//   FE FF -- -- - Big endian UTF-16, FE isn't valid UTF-8
			//   FF FE 00 00 - Little endian UTF-32, FF isn't valid UTF-8
			//   FF FE -- -- - Little endian UTF-16
			//   nn mm -- -- - UTF-8, includes EF BB BF case
			//   nn 00 00 00 - Little endian UTF-32
			//   nn 00 -- -- - Little endian UTF-16
			
			if ( uniChar[0] == 0xFE ) {
				if ( uniChar[1] == 0xFF ) xmlParser->charEncoding = kXMP_EncodeUTF16Big;	// FE FF
			} else if ( uniChar[0] == 0xFF ) {
				if ( uniChar[1] == 0xFE ) {
					if ( (uniChar[2] == 0) && (uniChar[3] == 0) ) {
						xmlParser->charEncoding = kXMP_EncodeUTF32Little;	// FF FE 00 00
					} else {
						xmlParser->charEncoding = kXMP_EncodeUTF16Little;	// FF FE
					}
				}
			} else if ( uniChar[1] != 0 ) {
				xmlParser->charEncoding = kXMP_EncodeUTF8;					// nn mm
			} else {
				if ( (uniChar[2] == 0) && (uniChar[3] == 0) ) {
					xmlParser->charEncoding = kXMP_EncodeUTF32Little;		// nn 00 00 00
				} else {
					xmlParser->charEncoding = kXMP_EncodeUTF16Little;		// nn 00
				}
			}

		}
		
		switch ( xmlParser->charEncoding ) {
			case kXMP_EncodeUTF16Big :
				xmlParser->UniConverter = (void*)UTF16BE_to_UTF8;
				break;
			case kXMP_EncodeUTF16Little :
				xmlParser->UniConverter = (void*)UTF16LE_to_UTF8;
				break;
			case kXMP_EncodeUTF32Big :
				xmlParser->UniConverter = (void*)UTF32BE_to_UTF8;
				break;
			case kXMP_EncodeUTF32Little :
				xmlParser->UniConverter = (void*)UTF32LE_to_UTF8;
				break;
			case kXMP_EncodeUTF8 :
				// Shift the buffered UTF-8 input so that it gets processed in the normal flow.
				for ( size_t i = 0; i < xmlParser->partialUniSize; ++i ) xmlParser->prevTail[i] = xmlParser->partialUniChar[i];
				xmlParser->prevTail[xmlParser->partialUniSize] = 0;
				xmlParser->partialUniSize = 0;
				break;
			default :
				XMP_Throw ( "Can't determine character encoding", kXMPErr_BadXMP );
		}
	
	}
	
	return copySize;
		
}	// DetermineInputEncoding


// -------------------------------------------------------------------------------------------------
// ProcessUTF8Portion
// ------------------
//
// The code to handle ASCII controls is vastly easier if the input is UTF-8. Higher level code
// converts the input buffers to UTF-8 if necessary.

static void
ProcessUTF8Portion ( XMLParserAdapter * xmlParser,
					 XMP_StringPtr	    buffer,
					 XMP_StringLen	    xmpSize,
					 XMP_OptionBits     options )
{
	// Old versions of the XMP toolkit wrote unescaped or singly escaped ASCII controls.
	// These get rejected by expat (or any other correct XML parser). Scan the input looking
	// for these and convert them to doubly escaped controls. We don't handle all possible
	// cases of escaping, just ones written by older versions of XMP: "&#xnn;" where "nn" is
	// one or two hex digits with uppercase letters. An incomplete escape at the end of a
	// buffer is saved as a string and finished from the start of the next buffer.
	
	// Ignore unescaped and singly escaped tab (0x9), LF (0xA), and CR (0xD), they are OK.

	size_t prevLen = strlen ( (char*) xmlParser->prevTail );
	XMP_Assert ( prevLen < 6 );
	size_t currLen = 0;	// The amount absorbed from the current buffer.

	while ( prevLen > 0 ) {
	
		// Have a partial escape left from the previous buffer. Gather 6 bytes and make a recursive
		// call to process it. Keep going until done or any leftover is fully in the current buffer.
		
		size_t moreLen = 6 - prevLen;
		if ( moreLen > xmpSize ) moreLen = xmpSize;
		for ( size_t i = 0; i < moreLen; ++i ) xmlParser->prevTail[prevLen+i] = buffer[i];
		xmlParser->prevTail[prevLen+moreLen] = 0;
		// printf ( "Leftover partial escape, had %d bytes, added %d bytes\n", prevLen, moreLen );

		currLen += moreLen;	// Accumulate the total amount taken from the current buffer;
		buffer  += moreLen;
		xmpSize -= moreLen;
		prevLen += moreLen;
		
		if ( prevLen < 6 ) {
			XMP_Assert ( xmpSize == 0 );
			break;	// ! Don't return, this might be the last buffer.
		}
		
		unsigned char  thisTail[kXMLParseTailSize];	// Capture the tail in a local buffer.
		XMP_Assert ( sizeof(thisTail) == sizeof(xmlParser->prevTail) );
		memcpy ( thisTail, xmlParser->prevTail, sizeof(thisTail) );	// AUDIT: Using dest sizeof for length is safe.
		xmlParser->prevTail[0] = 0;	// ! The recursive call must not think it has a tail.
		ProcessUTF8Portion ( xmlParser, (XMP_StringPtr)thisTail, prevLen, (options | kXMP_ParseMoreBuffers) );
		
		prevLen = strlen ( (char*) xmlParser->prevTail );	// Determine what was left from the recursive call.
		XMP_Assert ( prevLen < 6 );	// Need to make forward progress.
		if ( prevLen < currLen ) {	// Whatever is left is fully within this buffer.
			buffer  -= prevLen;
			xmpSize += prevLen;
			prevLen = 0;
			xmlParser->prevTail[0] = 0;
		}

	}
	
	XMP_Uns8 * spanStart = (XMP_Uns8*)buffer;
	XMP_Uns8 * bufferEnd = (XMP_Uns8*)buffer + xmpSize;
	XMP_Uns8 * spanEnd;
	XMP_Uns8 * refEnd;
	unsigned char controlValue;
	
	while ( spanStart < bufferEnd ) {
	
		// Find the next unescaped or singly escaped ASCII control.
		for ( spanEnd = spanStart; spanEnd < bufferEnd; ++spanEnd ) {

			if ( *((XMP_Uns8*)spanEnd) < 0x20 ) {

				refEnd = spanEnd + 1;
				if ( (*spanEnd != 0x9) && (*spanEnd != 0xA) && (*spanEnd != 0xD) ) break;

			} else if ( *spanEnd == '&' ) {

				refEnd = spanEnd + 1;
				controlValue = 0;

				if ( refEnd == bufferEnd ) break;	// A rather ugly sequence here, but hard to find better.
				if ( *refEnd == '#' ) {				// This is essentially a linearized state machine.
					++refEnd;
				} else {
					spanEnd = refEnd;
					continue;
				}

				if ( refEnd == bufferEnd ) break;
				if ( *refEnd == 'x' ) {
					++refEnd;
				} else {
					spanEnd = refEnd;
					continue;
				}

				if ( refEnd == bufferEnd ) break;
				if ( IsHexDigit ( *refEnd ) ) {
					controlValue = HexDigitValue ( *refEnd );
					++refEnd;
				} else {
					spanEnd = refEnd;
					continue;
				}

				if ( refEnd == bufferEnd ) break;
				if ( IsHexDigit ( *refEnd ) ) {	// ! The second hex digit is optional.
					controlValue = (controlValue << 4) + HexDigitValue ( *refEnd );
					if ( controlValue < 0x20 ) {
						++refEnd;
					} else {
						spanEnd = refEnd;
						continue;
					}
				}

				if ( refEnd == bufferEnd ) break;
				if ( (*refEnd == ';') && (controlValue != 0x9) && (controlValue != 0xA) && (controlValue != 0xD) ) {
					++refEnd;
					break;
				} else {
					spanEnd = refEnd;
					continue;
				}

			}

		}
		
		// Parse the span before the control, and the control if we found all of it. We get
		// here in one of three situations:
		//	1. The normal span extends to the end of the buffer.
		//	2. We have a normal span and a full control.
		//	3. We have a normal span and a partial control at the end of the buffer.
		
		// ! It is possible that refEnd == bufferEnd when we have a full control.

		if ( spanStart < spanEnd ) {
			// printf ( "Normal span: \"%.*s\"\n", (spanEnd - spanStart), spanStart ); fflush ( stdout );
			xmlParser->ParseBuffer ( spanStart, (spanEnd - spanStart), false );
			spanStart = spanEnd;
		}

		if ( spanEnd < bufferEnd ) {

			char escBuffer[32];
			XMP_Assert ( (*((XMP_Uns8*)spanEnd) < 0x20) || (*spanEnd == '&') );

			if ( *((XMP_Uns8*)spanEnd) < 0x20 ) {

				// printf ( "Raw control: 0x%X\n", *spanEnd ); fflush ( stdout );
				XMP_Assert ( refEnd == spanEnd+1 );
				snprintf ( escBuffer, sizeof(escBuffer), "&amp;#x%X;", *spanEnd );	// AUDIT: Using sizeof for snprintf length is safe.
				xmlParser->ParseBuffer ( escBuffer, strlen(escBuffer), false );

			} else {
			
				size_t refLen = refEnd - spanEnd;
				
				if ( *(refEnd-1) == ';' ) {
					++spanEnd;	// Skip the '&' in the singly escaped control.
					snprintf ( escBuffer, sizeof(escBuffer), "&amp;%.*s", refLen-1, spanEnd );	// AUDIT: Using sizeof for snprintf length is safe.
					// printf ( "Escaped control: %.*s -> %s\n", refLen, spanEnd-1, escBuffer ); fflush ( stdout );
					xmlParser->ParseBuffer ( escBuffer, strlen(escBuffer), false );
				} else {
					// printf ( "Partial escape: %.*s\n", refLen, spanEnd ); fflush ( stdout );
					XMP_Assert ( (refEnd == bufferEnd) && (xmlParser->prevTail[0] == 0) );
					XMP_Enforce ( refLen < sizeof(xmlParser->prevTail) );
					memcpy ( xmlParser->prevTail, spanEnd, refLen );	// AUDIT: Checked refLen above.
					xmlParser->prevTail[refLen] = 0;
				}

			}

			spanStart = refEnd;

		}
					
	}			
	
}	// ProcessUTF8Portion


// -------------------------------------------------------------------------------------------------
// ParseFromBuffer
// ---------------

void
XMPMeta::ParseFromBuffer ( XMP_StringPtr  buffer,
						   XMP_StringLen  xmpSize,
						   XMP_OptionBits options )
{
	if ( ! tree.children.empty() ) XMP_Throw ( "Can only parse into an empty object", kXMPErr_BadParse );
	if ( (buffer == 0) && (xmpSize != 0) ) XMP_Throw ( "Null parse buffer", kXMPErr_BadParam );
	if ( xmpSize == kXMP_UseNullTermination ) xmpSize = strlen ( buffer );
	
	const bool lastClientCall = ((options & kXMP_ParseMoreBuffers) == 0);	// *** Could use FlagIsSet & FlagIsClear macros.

	if ( xmlParser == 0 ) {
		if ( (xmpSize == 0) && lastClientCall ) return;	// Tolerate empty parse. Expat complains if there are no XML elements.
		xmlParser = new ExpatAdapter;
	}
	
	#if 0	// XMP_DebugBuild
		if ( xmlParser->parseLog != 0 ) {
			char message [200];
			snprintf ( message, sizeof(message), "<!-- ParseFromBuffer, length = %d, options = %X%s -->",
					   xmpSize, options, (lastClientCall ? " (last)" : "") );	// AUDIT: Using sizeof for snprintf length is safe.
			fwrite ( message, 1, strlen(message), xmlParser->parseLog );
			fflush ( xmlParser->parseLog );
		}
	#endif
	
	try {	// Cleanup the tree and xmlParser if anything fails.
	
		if ( xmlParser->charEncoding == XMP_OptionBits(-1) ) {
			size_t byteCount = DetermineInputEncoding ( xmlParser, buffer, xmpSize, lastClientCall );
			if ( xmlParser->charEncoding == XMP_OptionBits(-1) ) return;	// Still don't know the encoding.
			buffer  += byteCount;
			xmpSize -= byteCount;
		}
		
		if ( xmlParser->charEncoding == kXMP_EncodeUTF8 ) {

			// UTF-8 input buffers get directly processed.
			if ( xmpSize != 0 ) {	// The expat parser doesn't tolerate empty input.
				ProcessUTF8Portion ( xmlParser, buffer, xmpSize, options );
			}

		} else {
		
			// UTF-16 and UTF-32 input must be converted to UTF-8 then processed.
						
			enum { u8BufferSize = 4*1024 };
			XMP_Uns8 u8Buffer[u8BufferSize];
			size_t inCount, outCount;
			
			if ( xmlParser->partialUniSize > 0 ) {

				// We've got 1 to 3 bytes of a partial UTF-16 or UTF-32 character. Try to fill to 4
				// bytes, convert and process, resume below with the current input buffer. If we can
				// fill to 4 bytes then we're sure to get into the current buffer. UTF-32 input will
				// consume all 4 bytes. A UTF-16 surrogate pair will consume all 4 bytes. If the next
				// character is UTF-16 in the BMP we will consume at least 2 bytes. This happens only
				// if we enter with 1 byte, otherwise the BMP character would have been converted in
				// the prior call. If we can't fill to 4 bytes then we've consumed all of this buffer
				// and can leave any excess for the next call. There is also an edge case at the end
				// of UTF-6 input - come in with 1 byte, accumulate 1 more byte, have a BMP character.
				// So always try to convert and parse, even if we don't have 4 bytes.

				const XMP_StringPtr origBuffer = buffer;
				for ( ; (xmlParser->partialUniSize < 4) && (xmpSize > 0); ++xmlParser->partialUniSize, ++buffer, --xmpSize ) {
					xmlParser->partialUniChar[xmlParser->partialUniSize] = *buffer;
				}
				const size_t takenFromBuffer = buffer - origBuffer;
				
				if ( xmlParser->charEncoding & _XMP_UTF16_Bit ) {
					UTF16_to_UTF8_Proc UTF16_to_UTF8 = UTF16_to_UTF8_Proc ( xmlParser->UniConverter );
					UTF16_to_UTF8 ( (UTF16Unit*)xmlParser->partialUniChar, xmlParser->partialUniSize/2, u8Buffer, sizeof(u8Buffer), &inCount, &outCount );
					ProcessUTF8Portion ( xmlParser, XMP_StringPtr(u8Buffer), outCount, options );
					inCount *= 2;	// Convert it to a byte count.
				} else {
					XMP_Assert ( xmlParser->charEncoding & _XMP_UTF32_Bit );
					UTF32_to_UTF8_Proc UTF32_to_UTF8 = UTF32_to_UTF8_Proc ( xmlParser->UniConverter );
					UTF32_to_UTF8 ( (UTF32Unit*)xmlParser->partialUniChar, xmlParser->partialUniSize/4, u8Buffer, sizeof(u8Buffer), &inCount, &outCount );
					ProcessUTF8Portion ( xmlParser, XMP_StringPtr(u8Buffer), outCount, options );
					inCount *= 4;	// Convert it to a byte count.
				}
				
				const size_t partialLeft = xmlParser->partialUniSize - inCount;
				if ( partialLeft <= takenFromBuffer ) {
					buffer  -= partialLeft;
					xmpSize += partialLeft;
					xmlParser->partialUniSize = 0;
				} else {
					XMP_Assert ( xmpSize == 0 );
					for ( size_t i = 0; i < partialLeft; ++i ) xmlParser->partialUniChar[i] = xmlParser->partialUniChar[inCount+i];
					xmlParser->partialUniSize = partialLeft;
				}

			}
			
			// Convert and process the current buffer.
			
			bool doneLastPortion = false;
			while ( (xmpSize > 0) && (! doneLastPortion) ) {
			
				doneLastPortion = (xmpSize <= u8BufferSize/2);	// Worst case expansion is 2 bytes -> 3 bytes.
				
				if ( xmlParser->charEncoding & _XMP_UTF16_Bit ) {
					UTF16_to_UTF8_Proc UTF16_to_UTF8 = UTF16_to_UTF8_Proc ( xmlParser->UniConverter );
					UTF16_to_UTF8 ( (UTF16Unit*)buffer, xmpSize/2, u8Buffer, sizeof(u8Buffer), &inCount, &outCount );
					ProcessUTF8Portion ( xmlParser, XMP_StringPtr(u8Buffer), outCount, options );
					inCount *= 2;	// Convert it to a byte count.
				} else {
					XMP_Assert ( xmlParser->charEncoding & _XMP_UTF32_Bit );
					UTF32_to_UTF8_Proc UTF32_to_UTF8 = UTF32_to_UTF8_Proc ( xmlParser->UniConverter );
					UTF32_to_UTF8 ( (UTF32Unit*)buffer, xmpSize/4, u8Buffer, sizeof(u8Buffer), &inCount, &outCount );
					ProcessUTF8Portion ( xmlParser, XMP_StringPtr(u8Buffer), outCount, options );
					inCount *= 4;	// Convert it to a byte count.
				}
				
				buffer  += inCount;
				xmpSize -= inCount;
			
			}
			
			if ( xmpSize > 0 ) {
				XMP_Assert ( (xmpSize < 4) && (xmlParser->partialUniSize == 0) );
				for ( size_t i = 0; i < xmpSize; ++i ) xmlParser->partialUniChar[i] = buffer[i];
				xmlParser->partialUniSize = xmpSize;
				buffer += xmpSize;
				xmpSize = 0;
			}

		}
		
		if ( lastClientCall ) {
		
			if ( xmlParser->partialUniSize > 0 ) XMP_Throw ( "Input ends with a partial character", kXMPErr_BadXML );

			size_t remainder = strlen ( (char*)xmlParser->prevTail );
			// printf ( "Final remainder = %d\n", remainder ); fflush ( stdout );
			if ( remainder == 0 ) {
				xmlParser->prevTail[0] = ' ';	// The expat parser can't tolerate empty buffers.
				remainder = 1;
			}
			xmlParser->ParseBuffer ( (char*)xmlParser->prevTail, remainder, true );
			xmlParser->prevTail[0] = 0;
			
			#if XMP_DebugBuild && DumpXMLParseTree
				if ( xmlParser->parseLog == 0 ) xmlParser->parseLog = stdout;
				DumpXMLTree ( xmlParser->parseLog, xmlParser->tree, 0 );
			#endif

			const XML_Node * xmlRoot = FindRootNode ( this, *xmlParser, options );

			if ( xmlRoot != 0 ) {

				ProcessRDF ( &tree, *xmlRoot, options );
				NormalizeDCArrays ( &tree );
				if ( tree.options & kXMP_PropHasAliases ) CleanupParsedAliases ( &tree, options );
				TouchUpParsedNodes ( &tree );
				FixGPSTimeStamp ( &tree );
				
				// Delete empty schema nodes. Do this last, other cleanup can make empty schema.
				size_t schemaNum = 0;
				while ( schemaNum < tree.children.size() ) {
					XMP_Node * currSchema = tree.children[schemaNum];
					if ( currSchema->children.size() > 0 ) {
						++schemaNum;
					} else {
						delete tree.children[schemaNum];	// ! Delete the schema node itself.
						tree.children.erase ( tree.children.begin() + schemaNum );
					}
				}
				
			}

			delete xmlParser;
			xmlParser = 0;

		}
		
	} catch ( ... ) {

		delete xmlParser;
		xmlParser = 0;
		prevTkVer = 0;
		tree.ClearNode();
		throw;

	}
	
}	// ParseFromBuffer

// -------------------------------------------------------------------------------------------------
// SerializeToBuffer
// -----------------

void
XMPMeta::SerializeToBuffer ( XMP_StringPtr * rdfString,
							 XMP_StringLen * rdfSize,
							 XMP_OptionBits	 options,
							 XMP_StringLen	 padding,
							 XMP_StringPtr	 newline,
							 XMP_StringPtr	 indentStr,
							 XMP_Index		 baseIndent ) const
{
	XMP_Assert ( (rdfString != 0) && (rdfSize != 0) && (newline != 0) && (indentStr != 0) );

	// Fix up some default parameters.
	
	enum { kDefaultPad = 2048 };
	size_t unicodeUnitSize = 1;
	XMP_OptionBits charEncoding = options & kXMP_EncodingMask;

	if ( charEncoding != kXMP_EncodeUTF8 ) {
		if ( options & _XMP_UTF16_Bit ) {
			if ( options & _XMP_UTF32_Bit ) XMP_Throw ( "Can't use both _XMP_UTF16_Bit and _XMP_UTF32_Bit", kXMPErr_BadOptions );
			unicodeUnitSize = 2;
		} else if ( options & _XMP_UTF32_Bit ) {
			unicodeUnitSize = 4;
		} else {
			XMP_Throw ( "Can't use _XMP_LittleEndian_Bit by itself", kXMPErr_BadOptions );
		}
	}
	
	if ( *newline == 0 ) newline = "\xA";	// Linefeed
	if ( *indentStr == 0 ) {
		indentStr = " ";
		if ( ! (options & kXMP_UseCompactFormat) ) indentStr  = "   ";
		if ( options & kXMP_UsePlainXMP ) indentStr  = "  ";
	}
	
	if ( options & kXMP_ExactPacketLength ) {
		if ( options & (kXMP_OmitPacketWrapper | kXMP_IncludeThumbnailPad) ) {
			XMP_Throw ( "Inconsistent options for exact size serialize", kXMPErr_BadOptions );
		}
		if ( (padding & (unicodeUnitSize-1)) != 0 ) {
			XMP_Throw ( "Exact size must be a multiple of the Unicode element", kXMPErr_BadOptions );
		}
	} else if ( options & kXMP_ReadOnlyPacket ) {
		if ( options & (kXMP_OmitPacketWrapper | kXMP_IncludeThumbnailPad) ) {
			XMP_Throw ( "Inconsistent options for read-only packet", kXMPErr_BadOptions );
		}
		padding = 0;
	} else if ( options & kXMP_OmitPacketWrapper ) {
		if ( options & kXMP_IncludeThumbnailPad ) {
			XMP_Throw ( "Inconsistent options for non-packet serialize", kXMPErr_BadOptions );
		}
		padding = 0;
	} else {
		if ( padding == 0 ) padding = kDefaultPad * unicodeUnitSize;
		if ( options & kXMP_IncludeThumbnailPad ) {
			if ( ! this->DoesPropertyExist ( kXMP_NS_XMP, "Thumbnails" ) ) padding += (10000 * unicodeUnitSize);	// *** Need a better estimate.
		}
	}

	// Serialize as UTF-8, then convert to UTF-16 or UTF-32 if necessary, and assemble with the padding and tail.
	
	std::string tailStr;

	SerializeAsRDF ( *this, *sOutputStr, tailStr, options, newline, indentStr, baseIndent );
	if ( charEncoding == kXMP_EncodeUTF8 ) {

		if ( options & kXMP_ExactPacketLength ) {
			size_t minSize = sOutputStr->size() + tailStr.size();
			if ( minSize > padding ) XMP_Throw ( "Can't fit into specified packet size", kXMPErr_BadSerialize );
			padding -= minSize;	// Now the actual amount of padding to add.
		}
		
		size_t newlineLen = strlen ( newline );
	
		if ( padding < newlineLen ) {
			sOutputStr->append ( padding, ' ' );
		} else {
			padding -= newlineLen;	// Write this newline last.
			while ( padding >= (100 + newlineLen) ) {
				sOutputStr->append ( 100, ' ' );
				*sOutputStr += newline;
				padding -= (100 + newlineLen);
			}
			sOutputStr->append ( padding, ' ' );
			*sOutputStr += newline;
		}

		*sOutputStr += tailStr;
	
	} else {
	
		// Need to convert the encoding. Swap the UTF-8 into a local string and convert back. Assemble everything.
		
		XMP_VarString utf8Str, newlineStr;
		
		if ( charEncoding & _XMP_UTF16_Bit ) {

			std::string padStr ( "  " );  padStr[0] = 0;	// Assume big endian.
			UTF8_to_UTF16_Proc Converter = UTF8_to_UTF16BE;

			if ( charEncoding & _XMP_LittleEndian_Bit ) {
				padStr[0] = ' '; padStr[1] = 0;
				Converter = UTF8_to_UTF16LE;
			}
			
			utf8Str.swap ( *sOutputStr );
			UTF8_to_UTF16 ( utf8Str, sOutputStr, Converter );
			utf8Str.swap ( tailStr );
			UTF8_to_UTF16 ( utf8Str, &tailStr, Converter );

			if ( options & kXMP_ExactPacketLength ) {
				size_t minSize = sOutputStr->size() + tailStr.size();
				if ( minSize > padding ) XMP_Throw ( "Can't fit into specified packet size", kXMPErr_BadSerialize );
				padding -= minSize;	// Now the actual amount of padding to add (in bytes).
			}

			utf8Str.assign ( newline );
			UTF8_to_UTF16 ( utf8Str, &newlineStr, Converter );
			size_t newlineLen = newlineStr.size();
	
			if ( padding < newlineLen ) {
				for ( int i = padding/2; i > 0; --i ) *sOutputStr += padStr;
			} else {
				padding -= newlineLen;	// Write this newline last.
				while ( padding >= (200 + newlineLen) ) {
					for ( int i = 100; i > 0; --i ) *sOutputStr += padStr;
					*sOutputStr += newlineStr;
					padding -= (200 + newlineLen);
				}
				for ( int i = padding/2; i > 0; --i ) *sOutputStr += padStr;
				*sOutputStr += newlineStr;
			}

			*sOutputStr += tailStr;

		} else {

			std::string padStr ( "    " );  padStr[0] = padStr[1] = padStr[2] = 0;	// Assume big endian.
			UTF8_to_UTF32_Proc Converter = UTF8_to_UTF32BE;

			if ( charEncoding & _XMP_LittleEndian_Bit ) {
				padStr[0] = ' '; padStr[1] = padStr[2] = padStr[3] = 0;
				Converter = UTF8_to_UTF32LE;
			}
			
			utf8Str.swap ( *sOutputStr );
			UTF8_to_UTF32 ( utf8Str, sOutputStr, Converter );
			utf8Str.swap ( tailStr );
			UTF8_to_UTF32 ( utf8Str, &tailStr, Converter );

			if ( options & kXMP_ExactPacketLength ) {
				size_t minSize = sOutputStr->size() + tailStr.size();
				if ( minSize > padding ) XMP_Throw ( "Can't fit into specified packet size", kXMPErr_BadSerialize );
				padding -= minSize;	// Now the actual amount of padding to add (in bytes).
			}

			utf8Str.assign ( newline );
			UTF8_to_UTF32 ( utf8Str, &newlineStr, Converter );
			size_t newlineLen = newlineStr.size();
	
			if ( padding < newlineLen ) {
				for ( int i = padding/4; i > 0; --i ) *sOutputStr += padStr;
			} else {
				padding -= newlineLen;	// Write this newline last.
				while ( padding >= (400 + newlineLen) ) {
					for ( int i = 100; i > 0; --i ) *sOutputStr += padStr;
					*sOutputStr += newlineStr;
					padding -= (400 + newlineLen);
				}
				for ( int i = padding/4; i > 0; --i ) *sOutputStr += padStr;
				*sOutputStr += newlineStr;
			}

			*sOutputStr += tailStr;

		}
	
	}

	// Return the finished string.
	
	*rdfString = sOutputStr->c_str();
	*rdfSize   = sOutputStr->size();

}	// SerializeToBuffer

// =================================================================================================
